/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/*
 * The Original implementation kept data in a queue in case teh TCP connection
 * is not up anymore. This should be held transparent to this layer of
 * communication. In addition the design does NOT 
 *
 * The original idea of the sending queue was for the mode of send ad forget
 * about it to allow the router to dump a table to the SRx without having to
 * wait until each packet is send. The question here is:
 *
 * Do we really need that or can we just wait. I am not sure how long it would
 * take to send. The queue is anyhow max 100 entries wide and if it is for
 * performance issues than 100 are not enough anyhow, the queue then will slowly
 * fill up. If we need more than we hava a memory issue.
 *
 * I guess we can wait at least for each one to be send at a time. If we can not
 * wait for the receipt then the proxy needs to generate the id and the receipt
 * will be ignored. (maybe a list with id and results and if they match, then
 * don't notify the router).
 *
 * Maybe the Verify Request  can use the zero field for indication if the
 * receipt is needed. (A dangerous game though)
 *
 * GET RID OFF SEND QUEUE ??
 *
 *
 * @file client_connection_handler.c
 * @date Created: 04/15/2010
 */
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include "client_connection_handler.h"
#include "util/client_socket.h"
#include "util/log.h"
#include "shared/srx_packets.h"
#include "util/mutex.h"

/** Seconds between reconnect attempts */
#define RECONNECT_DELAY 30

/** THis value is used to prevent a deadlock during initialization and data
 * receiving. The data receiving waits in 1 seconds intervals until the client
 * connection handler is initialized. This wait could create a deadlock,
 * therefore the wait is reduced to a maximum of DEADLOCK_INDICATOR in seconds
 */
#define DEADLOCK_INDICATOR 30

/** Maximum attempts to connect to server during a connection request */
#define MAX_RECONNECT_ATTEMPT 10

/** Max. entries in the send queue */
#define MAX_SEND_QUEUE  100

#define HDR "([0x%08X] Client Connection Handler): "

static void* receivePacketsFromServer(void* clConnectionHandler);

////////////////////////////////////////////////////////////////////////////////
// Implementatinon of header file
////////////////////////////////////////////////////////////////////////////////

/*
 * Initialized the client connection handler used by the SRx Proxy to connect 
 * to the SRx server. The instance must be created elsewhere. This method 
 * requires the attribute initialized to be set to false.
 * Once set to true the instance must never be initialized again. the attribute
 * initialized will be set to true once this method exits with true.
 *
 * @param self             Instance that should be be initialized.
 * @param host             Server host name.
 * @param port             Server port address.
 * @param packetHandler    Received packets are passed to this function.
 * @param handshakeTimeout The time in seconds a handshake MUST be completed.
 * @param srxProxy         The proxy that uses this connection handler.
 *
 * @return true = connection established and handshake performed,
 *         false = failed to connect.
 *
 * @see releaseClientConnectionHandler
 */
bool initializeClientConnectionHandler(ClientConnectionHandler* self,
                                       const char* host, int port,
                                       SRxPacketHandler packetHandler,
                                       uint32_t handshakeTimeout,
                                       void* srxProxy)
{
  char* errPrefix = "[ClientConnectionHandler]";
  int iSemState =0;

  if (self->initialized)
  {
    // already initialized
    RAISE_ERROR("%s instance is already initialized!", errPrefix);

    return false;
  }

  // Try to establish a TCP connection with the server
  if (!createClientSocket(&self->clSock, host, port, true))
  {
    LOG(LEVEL_INFO, "%s Could not establish a connection to the server %s:%u.",
                errPrefix, host, port);

    return false;
  }

  // Set up the send queue -> it is used in case the TCP connection is currently
  // not available to queue the requests. Might not be needed though.

  if (!createRWLock(&self->queueLock))
  {

    closeClientSocket(&self->clSock);
    RAISE_ERROR("%s Coult not aquire read write lock!", errPrefix);

    return false;
  }

  initSList(&self->sendQueue);

  // Set misc. variables
  self->packetHandler     = packetHandler;
  self->stop              = false;
  self->handshake_timeout = handshakeTimeout;
  self->srxProxy          = srxProxy;

  // binary counter semaphore: inter-process shared: no,  inital value: 0 
  iSemState = sem_init(&self->sem_transx, 0, 1); 
  if(iSemState != 0) 
  {
    RAISE_ERROR("%s Failed to create the semaphore", errPrefix);
    return false;
  }

  // binary counter semaphore: inter-process shared: no,  inital value: 0 
  iSemState = sem_init(&self->sem_register, 0, 0); 
  if(iSemState != 0) 
  {
    RAISE_ERROR("%s Failed to create the semaphore", errPrefix);
    return false;
  }
 
  // binary counter semaphore: inter-process shared: no,  inital value: 1 
  iSemState = sem_init(&self->sem_crit_sec, 0, 1); 
  if(iSemState != 0) 
  {
    RAISE_ERROR("%s Failed to create the semaphore", errPrefix);
    return false;
  }
 

  // Thread to receive packets from the server
  if (pthread_create(&self->recvThread, NULL, receivePacketsFromServer, self)
      != 0)
  {
    closeClientSocket(&self->clSock);
    RAISE_ERROR("%s Failed to create the receiving thread", errPrefix);

    return false;
  }


  // Create the receive lock.
  self->cond       = (pthread_cond_t*)malloc(sizeof(pthread_cond_t));
  self->rcvMonitor = (pthread_mutex_t*)malloc(sizeof(pthread_mutex_t));
  
  // Initialize the receive lock.
  pthread_cond_init(self->cond, NULL);
  initMutex(self->rcvMonitor);

  // Instance is initialized and connected on TCP layer...
  self->initialized = true;
  // ... but not on application layer yet! This is to be done in the proxy.
  self->established = false;

  return true;
}

/**
 * Empty signal handler to prevent SIGINT from closing the program. This method
 * does absolutely nothing.
 *
 * @param _unused Not used here. (Just for Debug printout
 */
static void _suppressSIGINT(int _unused)
{
  // Do nothing - just suppress the signal
  LOG(LEVEL_DEBUG, HDR "SIGNAL RECEIVED [%u]\n", pthread_self(), _unused);
}

/**
 * Closes the connection to the server.
 *
 * @param self Instance that should be used
 */
void releaseClientConnectionHandler(ClientConnectionHandler* self)
{
  if (self != NULL)
  {
    // It is possible to receive a Goodbye during handshake in this case the 
    // connection handler is NOT initialized yet. The main process is still in 
    // init process and the init process has to cleanup.
    if (self->initialized)
    {
      // Do not receive or try to connect anymore
      self->stop = true;
      
      // Make sure the SIGINT does not terminate the program
//      signal(SIGINT, _suppressSIGINT);
      signal(SIGINT, SIG_IGN); // Ignore the signals

      if (self->established)
      {
        // disconnect on application layer.
        sendGoodbye(self, self->keepWindow);
      }

      pthread_kill(self->recvThread, SIGINT);
      closeClientSocket(&self->clSock);

      // Reinstall the default signal handler
      signal(SIGINT, SIG_DFL);

      // Deallocate the send queue and lock
      acquireWriteLock(&self->queueLock);
        releaseSList(&self->sendQueue);
      releaseRWLock(&self->queueLock);

      // Deallocate The packet receive momnitor
      if (self->rcvMonitor != NULL)
      {
        pthread_mutex_destroy(self->rcvMonitor);
        free(self->rcvMonitor);
        self->rcvMonitor = NULL;
      }
      
      if (self->cond != NULL)
      {
        pthread_cond_destroy(self->cond);
        free(self->cond);
        self->cond       = NULL;
      }
    }
  }
}


////////////////////////////////////////////////////////////////////////////////
// METHODS FOR WAIT AND NOTIFY
////////////////////////////////////////////////////////////////////////////////


/**
 * This method waits until data is received. It is expected that the lock is
 * already set.
 *
 * @param self The connection handler to wait on.
 *
 * @return true if the wait could be executed, otherwise false.
 */
inline bool connHandlerWait(ClientConnectionHandler* self)
{
  bool retVal = false;
  if (self->cond != NULL)
  {
    LOG(LEVEL_DEBUG, HDR "connHandlerWait enter", pthread_self());
    pthread_cond_wait(self->cond, self->rcvMonitor);
    LOG(LEVEL_DEBUG, HDR "connHandlerWait exit", pthread_self());
    retVal = true;
  }
  return retVal;
}

/**
 * This method waits until data is received. It is expected that the lock is
 * already set.
 *
 * @param self The Client connection handler.
 *
 * @return true if the notify could be send, otherwise false.
 */
inline bool connHandlerNotify(ClientConnectionHandler* self)
{
  bool retVal = false;
  if (self->cond != NULL)
  {
    LOG(LEVEL_DEBUG, HDR "connHandlerNotify", pthread_self());
    pthread_cond_signal(self->cond);
    retVal = true;
  }
  return retVal;
}

/**
 * This method locks the mutex for wait and notification
 *
 * @param self The Client connection handler.
 *
 * @return true if the lock could be taken.
 */
inline bool connHandlerLock(ClientConnectionHandler* self)
{
  bool retVal = false;
  if (self->rcvMonitor != NULL)
  {
    LOG(LEVEL_DEBUG, HDR "connHandlerLock", pthread_self());
    lockMutex(self->rcvMonitor);
    retVal = true;
  }
  return retVal;
}

/**
 * This method unlocks the mutex for wait and notification
 *
 * @param self The Client connection handler.
 *
 * @return true if the lock could be removed.
 */
inline bool connHandlerUnlock(ClientConnectionHandler* self)
{
  bool retVal = false;
  if (self->rcvMonitor != NULL)
  {
    LOG(LEVEL_DEBUG, HDR "connHandlerUnlock", pthread_self());
    pthread_mutex_unlock(self->rcvMonitor);
    retVal = true;
  }
  return retVal;
}

////////////////////////////////////////////////////////////////////////////////
// METHODS USED TO SEND
////////////////////////////////////////////////////////////////////////////////

/**
 * Sends a packet to the server. If the TCP connection is down but the
 * application layer connection is established, the packet might be queued.
 *
 * @param self Instance that should be used
 * @param data The proxySRX PDU to be send.
 * @param length Data size in bytes. This method does not read the header length
 *               field, it goes with the length provided. This can allow to
 *               send a stream of multiple pdu's at once.
 * @return true = data sent successfully, false = sending failed
 *         (e.g. no connection)
 */
bool sendPacketToServer(ClientConnectionHandler* self, void* data,
                        uint32_t length)
{
  if (isConnectedToServer(&self->clSock))
  {
    // This can contain more than one packet depending on the length and content
    // of the header.
    return sendData(&self->clSock, data, length);
    // Not online - store a copy and send later
  }
  else if (self->established) // Application layer connected.
  {
    void* dataCopy;

    // Check if the queue is not already full
    acquireReadLock(&self->queueLock);

    if (sizeOfSList(&self->sendQueue) == MAX_SEND_QUEUE)
    {
        unlockReadLock(&self->queueLock);

        return false;
    }

    // Store
    changeReadToWriteLock(&self->queueLock);

    dataCopy = appendToSList(&self->sendQueue, (size_t)length);

    unlockWriteLock(&self->queueLock);

    if (dataCopy == NULL)
    {
      return false;
    }

    memcpy(dataCopy, data, length);
  }

  return true;
}

/*
 * Create the connection of application layer between srx and proxy. This method
 * uses active waiting (1 sec. at a time) for max 30 seconds
 *
 * @param clSock The socket to be used to talk to the server
 * @param noPeers The number of peers this proxy contains
 * @param peerAS the peer numbers
 * @return true if the handshake was successful.
 */
bool handshakeWithServer(ClientConnectionHandler* self, SRXPROXY_HELLO* pdu)
{
  uint8_t                  responsePdu[sizeof(SRXPROXY_HELLO_RESPONSE)];
  SRXPROXY_HELLO_RESPONSE* responseHdr;
  uint32_t                 bytesReceived;
  int                      timeout = 0;

  // Send the 'Welcome' to the server
  if (!sendData(&self->clSock, (void*)pdu, ntohl(pdu->length)))
  {
    return false;
  }

  LOG(LEVEL_DEBUG, HDR "Wait for Handshake to complete...", pthread_self());
  // Wait until the server response
  // The Dispatcher will receive the notigication. Then we can proceed.
  while (   !self->established && !self->stop
         && (timeout < self->handshake_timeout))
  {
    // Sleep for a second and see if we received the hello response for
    // the handshake
    timeout++;
    sleep(1);
  }
  if (timeout >= self->handshake_timeout)
  {
    RAISE_ERROR("...Handshake timed out!");
  }
  else
  {
    LOG(LEVEL_DEBUG, HDR"...Handshake %s!", pthread_self(), (self->established 
                                                            ? "suceeded"
                                                            : "failed"));
  }
    
  return self->established;
}

/**
 * Send the goodbye message to the server. This message is done by directly
 * calling the sendData method without using the sendPacketToServer method.
 * Most likely after a goodbye the socket will be closed right away and all
 * buffers are emptied!
 *
 * The attribute "established" will be set to false due to the fact that this
 * message the application layer connection closes. (only if the function
 * returns true)
 *
 * @param self The pointer to the client connection handler.
 * @param keepWindow the keep window value.
 *
 * @return true if the message could be send, otherwise false.
 */
bool sendGoodbye(ClientConnectionHandler* self, uint16_t keepWindow)
{
  uint32_t length = sizeof(SRXPROXY_GOODBYE);
  uint8_t pdu[length];
  SRXPROXY_GOODBYE* hdr = (SRXPROXY_GOODBYE*)pdu;
  memset(pdu, 0, length);

  hdr->type       = PDU_SRXPROXY_GOODBYE;
  hdr->keepWindow = htons(keepWindow);
  hdr->length     = htonl(length);

  if (isConnectedToServer(&self->clSock))
  {
    if (sendData(&self->clSock, &pdu, length))
    {
      self->established = false;
      return true;
    }
  }

  return false;
}

////////////////////////////////////////////////////////////////////////////////
// Local helper functions
////////////////////////////////////////////////////////////////////////////////

/**
 * Send all messages that are held back in the queue.
 *
 * @param self The connection handler.
 */
static void sendWholeQueue(ClientConnectionHandler * self)
{
  SListNode* node;
  size_t     len;
  void*      edata;
  bool       succ;

  acquireWriteLock(&self->queueLock);

  while (!self->stop)
  {
    node = getRootNodeOfSList(&self->sendQueue);

    if (node == NULL)
    {
      break;
    }

    // Get entry
    len   = getDataSizeOfSListNode(node);
    edata = shiftFromSList(&self->sendQueue);

    // Send
    succ = sendData(&self->clSock, edata, len);

    free(edata);

    if (!succ)
    {
      break;
    }
  }

  unlockWriteLock(&self->queueLock);
}

/**
 * This method is called by the system thread that is created to manage the
 * package receiving. The method receivePackets is implemented in
 * packet.h/c and used for both proxy as well as SRx. The method
 * dispatchPackets then will handle all received packets.
 *
 * @param clConnectionHandler The pointer to the connection handler.
 * @return
 */
static void* receivePacketsFromServer(void* clConnectionHandler)
{
  ClientConnectionHandler* self = (ClientConnectionHandler*)clConnectionHandler;
  int deadLockPreventer = 0;

  LOG (LEVEL_DEBUG, "([0x%08X]) > Client Connection Handler Thread started!", pthread_self());
  
  do
  {
    if (self->initialized)
    {
      deadLockPreventer = 0;
      // Receive all packets - until an error occurs or the connection is lost
      receivePackets(getClientFDPtr(&self->clSock), self->packetHandler, 
                     self->srxProxy, PHT_PROXY);

      // Internal error or stop
      if (isConnectedToServer(&self->clSock) || self->stop)
      {
        break;
      }

      // Try to reconnect - on TCP layer.
      if (!reconnectToServer(&self->clSock, RECONNECT_DELAY,
                             MAX_RECONNECT_ATTEMPT))
      {
        break;
      }

      // Send all Accumulated data
      sendWholeQueue(self);
    }
    else
    {
      // Give it some time to initialize the client connection handler.
      sleep(1);
      deadLockPreventer++;
    }
  }
  while (!self->stop && (deadLockPreventer < DEADLOCK_INDICATOR ));
  
  LOG (LEVEL_DEBUG, "([0x%08X]) < Client Connection Handler Thread stopped!", pthread_self());
  
  pthread_exit(0);
}
