/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 *
 * Provides the API for the router implementation.
 *
 * Changelog:
 * -----------------------------------------------------------------------------
 *   0.2.0 - 2011/01/07 - oborchert
 *           * Changelog added with version 0.2.0 and date 2011/01/07
 *           * Version tag added
 *   0.1.0 - 2010/04/15 - pgleichm
 *           * Code Created
 * -----------------------------------------------------------------------------
 */

#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <stdint.h>
#include <errno.h>
#include "shared/srx_packets.h"
#include "srx_api.h"
#include "client_connection_handler.h"
#include "util/mutex.h"
#include "util/log.h"

// Constants
#define ECHO_BUFFER_INCREASE  1024
#define MAX_ECHO_MESSAGE_LEN  (ECHO_BUFFER_INCREASE * 8)

#define HDR "([0x%08X] SRX API): "

////////////////////////////////////////////////////////////////////////////////
// Forward declaration
////////////////////////////////////////////////////////////////////////////////
static void dispatchPackets(SRXPROXY_BasicHeader* packet, void* proxyPtr);

////////////////////////////////////////////////////////////////////////////////
// Implementation of header file
////////////////////////////////////////////////////////////////////////////////

/**
 * Creates a proxy instance and registers the callback methods for signature
 * generation and validation notification.
 *
 * @param vallidationReadyCallback The function the proxy calls to communicate
 *                   the validation result or changes in prior validation
 *                   results to the proxy user. This can be for instance the
 *                   router or policy module, etc.
 * @param signatureReadyCallback The function, the SRx calls once the requested
 *                   signature is generated.
 * @param requestSynchronizationCallback This function is used to request a
 *                   synchronization. A call of this method should initiate a
 *                   complete validation request from the routers side to
 *                   guarantee that SRx has a complete view on the validated
 *                   data within the user.
 * @param proxyID    The id of the proxy used during the handshake with SRx. If
 *                   SRx is requested to generate the ID this value MUST be "0"
 *                   zero.
 * @param proxyAS    The AS of the user that uses this proxy.
 *
 * @param userPtr    This parameter is used or not by the router or policy
 *                   module or other that accesses the proxy. It does not have
 *                   any meaning for SRx or the proxy itself but might be
 *                   meaningful for the software (router, policy module, etc)
 *                   that uses this API.
 *
 * @return The SRx proxy instance or null if not enough memory was available
 *         or the chosen operational mode is not supported.
 */
SRxProxy* createSRxProxy(ValidationReady  validationReadyCallback,
                         SignaturesReady  signatureReadyCallback,
                         SyncNotification requestSynchronizationCallback,
                         uint32_t proxyID, uint32_t proxyAS, void* userPtr)
{
  SRxProxy* proxy;
  // The client connection handler
  ClientConnectionHandler* connHandler;

  // Allocate the internal structure
  proxy = malloc(sizeof(SRxProxy));
  if (proxy == NULL)
  {
    RAISE_SYS_ERROR("Not enough memory for an SRx instance");
    return NULL;
  }
  // Initialize it to zero
  memset(proxy, 0, sizeof(SRxProxy));
  
  // Initialize the member variables
  proxy->resCallback      = validationReadyCallback;
  proxy->sigCallback      = signatureReadyCallback;
  proxy->syncNotification = requestSynchronizationCallback;

  // initialize the connection handler
  connHandler = malloc(sizeof(ClientConnectionHandler));
  
  proxy->connHandler = connHandler;
  connHandler->initialized = false;
  connHandler->established = false;
  connHandler->keepWindow  = SRX_DEFAULT_KEEP_WINDOW;
  // The connHandler->cond and connHandler->rcvMonitor are created, initialized 
  // and released in the connection handlers init and release method
  connHandler->cond        = NULL;
  connHandler->rcvMonitor  = NULL;

  proxy->proxyAS = proxyAS;
  proxy->proxyID = proxyID;

  initSList(&proxy->peerAS);

  proxy->userPtr = userPtr;

  return proxy;
}

/**
 * Disconnect proxy SRx server if necessary and frees up the memory again.
 *
 * @param proxy the proxy instance.
 */
void releaseSRxProxy(SRxProxy* proxy)
{
  if (proxy != NULL)
  {
    disconnectFromSRx(proxy, SRX_DEFAULT_KEEP_WINDOW);
    releaseSList(&proxy->peerAS);
    free(proxy->connHandler);
    free(proxy);
  }
}

/**
 * Add peers to the configuration of the proxy. In case the proxy is already
 * connected, the peer change will be communicated to SRx.
 *
 * @param proxy The proxy instance
 * @param noPeers The number of peers.
 * @param peerAS The array of peers.
 */
void addPeers(SRxProxy* proxy, uint32_t noPeers, uint32_t* peerAS)
{
  uint32_t  i;
  uint32_t* dataVal = NULL;
  // The client connection handler
  ClientConnectionHandler* connHandler = 
                                   (ClientConnectionHandler*)proxy->connHandler;

  //if (connHandler->initialized)
  if (isConnected(proxy))
  {
    uint32_t pduSize = sizeof(SRXPROXY_PEER_CHANGE);
    uint32_t dataSize = noPeers * pduSize;
    uint8_t  data[dataSize];
    SRXPROXY_PEER_CHANGE* hdr = (SRXPROXY_PEER_CHANGE*)data;

    memset(data, 0, sizeof(dataSize));

    for (i = 0;i < noPeers; i++, hdr++, peerAS++)
    {
      // Create header element
      hdr->type       = PDU_SRXPROXY_PEER_CHANGE;
      hdr->changeType = SRX_PROXY_PEER_CHANGE_TYPE_ADD;
      hdr->length     = htonl(pduSize);
      hdr->peerAS     = htonl(*peerAS);

      // Add peerAS to list
      dataVal = appendToSList(&proxy->peerAS, sizeof(uint32_t));
      memcpy(dataVal, peerAS, sizeof(uint32_t));
    }
    
    // Send peerAS changes to SRx
    sendData(&connHandler->clSock, &data, dataSize);
  }
  else
  {
    for (i = 0;i < noPeers; i++, peerAS++)
    {
      // Add peerAS to list
      dataVal  = appendToSList(&proxy->peerAS, sizeof(uint32_t));
      *dataVal = *peerAS;
    }
  }
}

/**
 * Remove peers from the proxy configuration. In case the proxy is already
 * connected, the peer change will be communicated to SRx.
 *
 * @param proxy The proxy instance
 * @param noPeers The number of peers.
 * @param peerAS The array of peers.
 */
void removePeers(SRxProxy* proxy, uint32_t noPeers, uint32_t* peerAS)
{
  uint32_t  i;
  // The client connection handler
  ClientConnectionHandler* connHandler = 
                                   (ClientConnectionHandler*)proxy->connHandler;

  //if (connHandler->initialized)
  if (isConnected(proxy))
  {
    uint32_t pduSize = sizeof(SRXPROXY_PEER_CHANGE);
    uint32_t dataSize = noPeers * pduSize;
    uint8_t  data[dataSize];
    SRXPROXY_PEER_CHANGE* hdr = (SRXPROXY_PEER_CHANGE*)data;
    
    memset(data, 0, sizeof(dataSize));

    hdr->type       = PDU_SRXPROXY_PEER_CHANGE;
    hdr->changeType = SRX_PROXY_PEER_CHANGE_TYPE_REMOVE;
    hdr->length     = htonl(pduSize);
    for (i = 0;i < noPeers; i++, hdr++, peerAS++)
    {
      // Create header element
      hdr->peerAS     = htonl(*peerAS);
      // Remove peerAS from list
      deleteFromSList(&proxy->peerAS, peerAS);
      sendData(&connHandler->clSock, &data, dataSize);
    }
  }
  else
  {
    for (i = 0;i < noPeers; i++, peerAS++)
    {
      // Remove peerAS from list
      deleteFromSList(&proxy->peerAS, peerAS);
    }
  }
}

/**
 * Remove peers from the proxy configuration. In case the proxy is already
 * connected, the peer change will be communicated to SRx.
 *
 * @param proxy The proxy instance
 * @param keep_window The number of seconds requested to keep the data on the
 *        server side.
 * @param updateID The ID of the update to deleted.
 */
void deleteUpdate(SRxProxy* proxy, uint16_t keep_window, SRxUpdateID* updateID)
{
  uint32_t  i;
  // The client connection handler
  ClientConnectionHandler* connHandler = 
                                   (ClientConnectionHandler*)proxy->connHandler;

  //if (connHandler->initialized && connHandler->established)
  if (isConnected(proxy))
  {
    uint32_t length = sizeof(SRXPROXY_DELETE_UPDATE);
    SRXPROXY_DELETE_UPDATE* hdr = malloc(length);
    
    memset(hdr, 0, length);

    hdr->type             = PDU_SRXPROXY_DELTE_UPDATE;
    hdr->keepWindow       = htons(keep_window);
    hdr->length           = htonl(length);
    hdr->updateIdentifier = htonl(*updateID);

    sendData(&connHandler->clSock, hdr, length);

    free(hdr);
  }
  else
  {
    if (!connHandler->initialized)
    {
      RAISE_ERROR("Delete update [0x%08X] could not be send, Connection handler"
                  " is not initialized!", *updateID);
    }
    else
    {
      RAISE_ERROR("Delete update [0x%08X] could not be send, connection to SRx"
                  " is not established!", *updateID);
    }
  }
}

/**
 * Connect to the SRx server instance. This method performs the complete
 * session negotiation with the SRx server. See 6.1 of srx-proxy protocol
 * specification.
 *
 * The given proxy instance is expected to be completely configured. Any 
 * previously established connection will be ended.
 *
 * In case the connection handshake fails the connection is closed and the
 * connection handler is released.
 *
 * @param proxy The proxy instance
 * @param host  Server host name
 * @param port  Server port address
 * @param handshakeTimeout The time in seconds before a handshake is timed out.
 *
 * @return \c true = connected, \c false = failed to connect
 */
bool connectToSRx(SRxProxy* proxy, const char* host, int port,
                  int handshakeTimeout)
{
  // The client connection handler
  ClientConnectionHandler* connHandler = 
                                   (ClientConnectionHandler*)proxy->connHandler;
  // Check if the connectionHandler is initialized.
  if (connHandler->initialized)
  {
    if (connHandler->stop)
    {
      RAISE_ERROR("Proxy [ID:%u] is already initialized and stopped!",
                  proxy->proxyID);
      return false;
    }
    else if (connHandler->established)
    {
      LOG(LEVEL_INFO, "Proxy [ID:%u] is already connected!");
      return true;
    }
  }
  else if (!initializeClientConnectionHandler(connHandler, host, port,
                                              dispatchPackets, handshakeTimeout,
                                              (void*)proxy))
  {
    LOG(LEVEL_DEBUG, "Proxy [ID:%u] could not be initialized - Check if server"
                      "%s:%u is accessible!", proxy->proxyID, host, port);
    return false;
  }

  if (connHandler->established)
  {
    // Maybe upgrade to RAISE_SYS_ERROR
    // If it ever reaches here the connection is unexpectedly established.
    RAISE_ERROR("BUG: Proxy [ID:%u] is unexpectedly connected to SRx, "
                "disconnect, stop proxy and return connection failed!");
    disconnectFromSRx(proxy, SRX_DEFAULT_KEEP_WINDOW);
    return false;
  }

  LOG(LEVEL_INFO, "Establish connection with proxy [%u]...", proxy->proxyID);
  uint32_t noPeers    = proxy->peerAS.size;
  uint32_t length     = 20 + (noPeers * 4);
  uint8_t  pdu[length];
  SRXPROXY_HELLO* hdr = (SRXPROXY_HELLO*)pdu;
  uint32_t peerASN    = 0;
  uint32_t* peerAS    = NULL;

  memset(pdu, 0, length);

  hdr->type            = PDU_SRXPROXY_HELLO;
  hdr->version         = htons(SRX_PROTOCOL_VER);
  hdr->length          = htonl(length);
  hdr->proxyIdentifier = htonl(proxy->proxyID);
  hdr->asn             = htonl(proxy->proxyAS);
  hdr->noPeers         = htonl(noPeers);

  peerAS = (uint32_t*)&hdr->peerAS;
  SListNode* node = getRootNodeOfSList(&proxy->peerAS);
  while (!(node == NULL))
  {
    peerASN = *((uint32_t*)node->data);
    *peerAS = htonl(peerASN);
    peerAS++;
    node = node->next;
  }

  if (handshakeWithServer(proxy->connHandler, hdr))
  {
    connHandler->established = true;
    LOG(LEVEL_INFO, "Connection with proxy [%u] established!",proxy->proxyID);
  }
  else
  {
    if (!connHandler->stop) // No Goodbye received
    {
      releaseClientConnectionHandler(connHandler);
    }

    RAISE_ERROR("Handshake with server (%s:%u) failed!", host, port);

    return false;
  }

  // IF WE GOT HERE ALL WENT WELL !
  return connHandler->established;
}

/**
 * Disconnects the proxy from the SRx Server instance on both, application and
 * transport layer.
 *
 * @param proxy The proxy instance.
 * @param keepWindow Sends a request to keep the proxy data with SRx for given
 *                   amount of seconds.
 *
 * @return false if the given proxy is of unknown implementation, otherwise
 *         true;
 */
bool disconnectFromSRx(SRxProxy* proxy, uint16_t keepWindow)
{
  // The client connection handler
  ClientConnectionHandler* connHandler = 
                                   (ClientConnectionHandler*)proxy->connHandler;
  // Macro for type casting back to the proxy.
  if (isConnected(proxy))
  {
    sendGoodbye(connHandler, keepWindow);
    connHandler->established = false;    
    releaseClientConnectionHandler(connHandler);
  }

  return true;
}

/**
 * Returns if the proxy is connected or not.
 * 
 * @param proxy The proxy itself.
 * 
 * @return true if the proxy is connected, false otherwise.
 */
bool isConnected(SRxProxy* proxy)
{
  bool retVal = false;
  if (proxy != NULL)
  {
    if (proxy->connHandler != NULL)
    {
      ClientConnectionHandler* hdl=(ClientConnectionHandler*)proxy->connHandler;
      retVal = hdl->initialized && hdl->established;
    }
  }
  return retVal;
}


////////////////////////////////////////////////////////////////////////////////
// Local helper functions
////////////////////////////////////////////////////////////////////////////////


void processDeleteUpdate(SRxProxy* proxy, SRXPROXY_DELETE_UPDATE* hdr);


/**
 * Create an IPv4 update validation request
 * @param pdu The allocated space for this PDU
 * @param method The verification method
 * @param defaultResult the provided result information
 * @param prefix The IPv4 prefix
 * @param as32 The origin as
 * @param bgpsec The bgpsec blob
 * @return The filled PDU
 */
uint8_t* createV4Request(uint8_t* pdu, SRxVerifyFlag method,
                      SRxDefaultResult* defaultResult, IPPrefix* prefix, 
                      uint32_t as32, BGPSecData* bgpsec)
{
  int i;
  SRXPROXY_VERIFY_V4_REQUEST* hdr = (SRXPROXY_VERIFY_V4_REQUEST*)pdu;
  hdr->type          = PDU_SRXPROXY_VERIFY_V4_REQUEST;
  hdr->flags         = method;
  hdr->roaResSrc     = defaultResult->resSourceROA;
  hdr->bgpsecResSrc  = defaultResult->resSourceBGPSEC;
  hdr->length        = htonl(24 + bgpsec->length);
  hdr->roaDefRes     = defaultResult->result.roaResult;
  hdr->bgpsecDefRes  = defaultResult->result.bgpsecResult;
  hdr->prefixLen     = prefix->length;
  hdr->prefixAddress = prefix->ip.addr.v4;
  hdr->originAS      = htonl(as32);
  hdr->bgpsecLength  = htonl(bgpsec->length);

  for (i = 0; i < bgpsec->length; i++)
  {
    pdu[i+24] = bgpsec->data[i];
  }

  return pdu;
}

/**
 * Create an IPv6 update validation request
 * @param pdu The allocated space for this PDU
 * @param method The verification method
 * @param defaultResult the provided result information
 * @param prefix The ipv6 prefix
 * @param as32 The origin as
 * @param bgpsec The bgpsec blob
 * @return The filled PDU
 */
uint8_t* createV6Request(uint8_t* pdu, SRxVerifyFlag method,
                     SRxDefaultResult* defaultResult, IPPrefix* prefix, 
                     uint32_t as32, BGPSecData* bgpsec)
{
  int i;
  SRXPROXY_VERIFY_V6_REQUEST* hdr = (SRXPROXY_VERIFY_V6_REQUEST*)pdu;
  hdr->type          = PDU_SRXPROXY_VERIFY_V6_REQUEST;
  hdr->flags         = method;
  hdr->roaResSrc     = defaultResult->resSourceROA;
  hdr->bgpsecResSrc  = defaultResult->resSourceBGPSEC;
  hdr->length        = htonl(36 + bgpsec->length);
  hdr->roaDefRes     = defaultResult->result.roaResult;
  hdr->bgpsecDefRes  = defaultResult->result.bgpsecResult;
  hdr->prefixLen     = prefix->length;
  hdr->prefixAddress = prefix->ip.addr.v6;
  hdr->originAS      = htonl(as32);
  hdr->bgpsecLength  = htonl(bgpsec->length);
  for (i = 0; i < bgpsec->length; i++)
  {
    pdu[i+37] = bgpsec->data[i];
  }

  return pdu;
}

/**
 * Verifies the given update data. All parameters except the result parameter
 * are IN parameters, result is an OUT parameter that will be filled within this
 * function. The memory MUST be allocated outside of this function.
 *
 * @param proxy The proxy instance
 * @param result This is the pointer to the result object that will be returned.
 *               This is an OUT parameter.
 * @param method Method(s) to use for the verification (ROA, BGPSEC, BOTH). The 
 *               receipt is requested is encoded in the method as well.
 * @param defaultResult The parameter contains the default information to be 
 *           used in case the validation result is not readily available.
 * @param prefix The prefix of the request. (both v4/v6 possible)
 * @param as32 Origin AS (32-bit)
 * @param bgpsec the bgpsec information.
 *
 * @return The pointer to the provided validation result or NULL if validation
 *         encountered an error. Most likely the error is that the proxy is not
 *         connected.
 */
SRxValidationResult* verifyUpdate(SRxProxy* proxy, SRxValidationResult* result,
                                  SRxVerifyFlag method, 
                                  SRxDefaultResult* defaultResult,
                                  IPPrefix* prefix, uint32_t as32,
                                  BGPSecData* bgpsec)

{
  int s=0;
  if (!isConnected(proxy))
  {
    RAISE_ERROR(HDR "Abort verify, not connected to SRx server!" , 
                pthread_self());
    return NULL;
  }  
  
  int i_semval;
  bool isV4 = prefix->ip.version == 4;
  // The client connection handler
  ClientConnectionHandler* connHandler = 
                                   (ClientConnectionHandler*)proxy->connHandler;  
  // create data packet.
  uint32_t length = (isV4 ? sizeof(SRXPROXY_VERIFY_V4_REQUEST)
                          : sizeof(SRXPROXY_VERIFY_V6_REQUEST)) +bgpsec->length;
  uint8_t pdu[length];
  memset(pdu, 0, length);

  // PREPARE MUTEX - This is used to synchronize this block!
  //connHandlerLock(connHandler);

  //sem_wait(&connHandler->sem_transx); // P operation
  // Prepare the result value
  
  memset(result, 0, sizeof(SRxValidationResult));
  
  if (   (pthread_self() == connHandler->recvThread)
      && ((method & SRX_FLAG_REQUEST_RECEIPT) != 0))
  {
    // The receipt flag is set for a request that is processed by the proxies 
    // receiving process. To prevent a deadlock within the receiver process, 
    // the receipt flag will be removed. The proxy itself will generate the id
    // and uses the default values as result. In case the result differs from 
    // the SRx database, SRx will send a notification to communicate the 
    // correct value.
    LOG(LEVEL_WARNING, "Remove receipt flag to prevent deadlock. See srx_api.c"
                       ":verifyUpdate(...) in-line doc for information!");
    // To prevent a deadlock in the receiver process a possible 
    method &= ~SRX_FLAG_REQUEST_RECEIPT;
  }

  
  if ((method & SRX_FLAG_REQUEST_RECEIPT) != 0)
  {
    if (connHandler->syncValData == NULL)
    {
      connHandler->syncValData = result;
    }
    else
    {
      RAISE_ERROR("A validation request with receipt is requested but a "
                  "previous data structure for results is still instantiated! "
                  "Abort verification request!");
      // UNLOCK MUTEX
      //pthread_mutex_unlock(connHandler->rcvMonitor);
      //sem_post(&connHandler->sem_transx); // V operation
      return false;
    }
  }
  else
  {
    // Fill the result value
    // This should normally be done by SRx
    SRxUpdateID updateID = (SRxUpdateID)generateIdentifier(as32, prefix,
                                                           bgpsec->length,
                                                           bgpsec->data);

    result->valType = method;
    result->updateID = updateID;

    result->valResult.roaResult = ((method & SRX_FLAG_ROA) != 0)      
                                     ? defaultResult->result.roaResult
                                     : SRx_RESULT_UNDEFINED;
    
    result->valResult.bgpsecResult = ((method & SRX_FLAG_BGPSEC) > 0)
                                     ? defaultResult->result.bgpsecResult
                                     : SRx_RESULT_UNDEFINED;
  }

  // Generate VERIFY PACKET
  if (isV4)
  {
    createV4Request(pdu, method, defaultResult, prefix, as32, bgpsec);
  }
  else
  {
    createV6Request(pdu, method, defaultResult, prefix, as32, bgpsec);
  }

  if (!sendPacketToServer(connHandler, (SRXPROXY_PDU*)pdu, length))
  {
      RAISE_ERROR("Failure during sending update request!");
      //pthread_mutex_unlock(connHandler->rcvMonitor);
      //sem_post(&connHandler->sem_transx); // V operation : atomically increase
      return NULL;
  }
#if 0
  do {
      usleep(2000);
      sem_getvalue(&connHandler->sem_transx, &i_semval);
  } while( i_semval != 0);
#endif

#if 0
  sem_getvalue(&connHandler->sem_transx, &i_semval);
  if( i_semval == 0 && connHandler->sem_transx.__size[0]==0)
      sem_post(&connHandler->sem_transx); // V operation
#endif


  // NOW GOTO SLEEP AND WAIT FOR THE RECEIPT IF NECESSARY
  if ((method & SRX_FLAG_REQUEST_RECEIPT) != 0)
  {
    LOG(LEVEL_DEBUG, HDR "verifyUpdate: Go into passive wait mode until the "
                     "verify notification containing the receipt is received!", 
                     pthread_self());
    // Wait until SRx sends the receipt with the new data
    LOG(LEVEL_DEBUG, HDR "WAIT FOR RESULT of 0x%08X\n", result->updateID);

#if 0
    struct timespec to;
    memset(&to, 0, sizeof(to));
    to.tv_sec = time(0) +5;
    to.tv_nsec = 0;
    return(pthread_cond_timedwait(cond, self, &to));
#endif
    
    SRxProxy *old_proxy;
    s = sem_wait_wrapper(&connHandler->sem_register); // P operation : atomically decrease
    if (s == -1) 
    {
	if(errno == ETIMEDOUT) {
	    LOG(LEVEL_DEBUG, HDR "sem_timedwait() time out\n");

	    /* call disconnect */
	    LOG(LEVEL_DEBUG, HDR "Disconnect !");      
#if 0
	    old_proxy = proxy;
	    proxy = createSRxProxy(proxy->resCallback, \
                                   proxy->sigCallback, proxy->syncNotification, \
                                   proxy->proxyID, proxy->proxyAS, proxy->userPtr);

	    disconnectFromSRx (old_proxy, 0);      
	    releaseSRxProxy(old_proxy);
#endif
	}
	else if (errno == EINTR)
	    ;//continue;
    }
    //connHandlerWait(proxy->connHandler);
    LOG(LEVEL_DEBUG, HDR "BACK FROM WAIT FOR RESULT of 0x%08X\n", result->updateID);
  }

  // Data was filled by receiver.
  LOG(LEVEL_DEBUG,HDR "verifyUpdate: Result for update [ID:0x%08X], "
	  "Type[%02X], ROA[%02X], BGPSEC[%02X]", pthread_self(), 
	  result->updateID, result->valType, 
	  result->valResult.roaResult, result->valResult.bgpsecResult);


  // Free data for next request.
  connHandler->syncValData = NULL;

  //connHandlerUnlock(proxy->connHandler);
  
  return result;
}

/**
 * This method generates a signature request. The signature will be returned
 * using the signature notification callback.
 *
 * @param proxy Pointer to the proxy instance
 * @param updId The update id the signature is requested for
 * @param onlyOwnSignature Return only the latest signature block generated for
 *                         this BGP router, otherwise the own signatures will be
 *                         concatenated to the previous given BGPSEC data block.
 * @param algorithm        The algorithm to use.
 * @param prependCounter   Number of times the own AS will be prepended into the
 *                         path. This allows traffic engineering.
 * @param peerAS           The peer AS the update will be send to.
 *
 * @return true if sending the signature request was successful.
 */
bool signUpdate(SRxProxy* proxy, SRxUpdateID* updateId, bool onlyOwnSignature,
                uint16_t algorithm, uint32_t prependCounter, uint32_t peerAS)
{
  bool retVal = true;

  uint32_t length = sizeof(SRXPROXY_SIGN_REQUEST);
  uint8_t pdu[length];
  SRXPROXY_SIGN_REQUEST* hdr = (SRXPROXY_SIGN_REQUEST*)pdu;
  memset(pdu, 0, length);
  // Assemble the PDU
  hdr->type      = PDU_SRXPROXY_SIGN_REQUEST;
  hdr->algorithm = htons(algorithm);
  hdr->blockType = onlyOwnSignature ? SRX_PROXY_BLOCK_TYPE_LATEST_SIGNATURE
                                    : 0;
  hdr->length    = htonl(length);
  hdr->updateIdentifier = htonl(*updateId);
  hdr->prependCounter   = htonl(prependCounter);
  hdr->peerAS           = htonl(peerAS);

  if (!sendPacketToServer(proxy->connHandler, (SRXPROXY_PDU*)pdu, length))
  {
    RAISE_ERROR("Failure during sending signature request for update "
                "[ID:0x%08X]!", updateId);
    retVal = false;
  }
  
  return retVal;
}

////////////////////////////////////////////////////////////////////////////////
// RECEIVE AND HANDLE PACKETS
////////////////////////////////////////////////////////////////////////////////

/**
 * Finish the handshake or raise an error.
 *
 * @param hdr The "Hello Response" Header.
 * @param self The instance of the connection handler.
 */
void processHelloResponse(SRXPROXY_HELLO_RESPONSE* hdr, SRxProxy* proxy)
{
  // The client connection handler
  ClientConnectionHandler* connHandler = 
                                   (ClientConnectionHandler*)proxy->connHandler;
  if (ntohs(hdr->version) == SRX_PROTOCOL_VER)
  {
    connHandler->established = true;
  }
  else
  {
    RAISE_ERROR("Protocol version between SRx [%u] and proxy [%u] differ.",
                ntohs(hdr->version), SRX_PROTOCOL_VER);
  }
}

/**
 * Server send a connection closing.
 *
 * @param hdr The "Goodbye" Header
 * @param self The instance of the connection handler.
 */
void processGoodbye(SRXPROXY_GOODBYE* hdr, SRxProxy* proxy)
{
  // The client connection handler
  ClientConnectionHandler* connHandler = 
                                   (ClientConnectionHandler*)proxy->connHandler;
  LOG(LEVEL_DEBUG, HDR "Received Goodbye", pthread_self());
  // SERVER CLOSES THE CONNECTION. END EVERYTHING.
  connHandler->established = false;
  connHandler->stop = true;
  // First send a signal just in case someone still waiting
  // (should not be)
  // Now release connection handler.
  releaseClientConnectionHandler(connHandler);
}

/**
 * The SRx server send a verification notification. This method uses the proxy
 * callback.
 *
 * @param hdr The "Verify Notification" Header
 * @param self The instance of the connection handler.
 */
void processVerifyNotify(SRXPROXY_VERIFY_NOTIFICATION* hdr, SRxProxy* proxy)
{
  int i_semval;
  bool receipt = (hdr->resultType & SRX_FLAG_REQUEST_RECEIPT) > 0;
  SRxValidationResult* result = NULL;

  //printf("receipt:%02x %02x KH_type:%02x\n", hdr->type, hdr->resultType, g_KH_type);
  // The client connection handler
  ClientConnectionHandler* connHandler = 
                                   (ClientConnectionHandler*)proxy->connHandler;

  // GET MUTEX AND LOCK
  //connHandlerLock(proxy->connHandler);

  sem_wait_wrapper(&connHandler->sem_transx); // P operation: atomically decrease

  if (receipt)
  {
    // Check if the receipt can be processed!
    if (connHandler->syncValData == NULL)
    {
      RAISE_SYS_ERROR("Received a notification request receipt but no result "
                      "Object is registered. Continue processing it as regular "
                      "notification and remove receipt flag!");
      // turn off receipt mode and continue processing data!.
      receipt = false;
      // Remove receipt flag, to not confuse the router instance.
      hdr->resultType = hdr->resultType & (~SRX_FLAG_REQUEST_RECEIPT);
    }
    else
    {
      result = (SRxValidationResult*)connHandler->syncValData;
    }
  }

  if (!receipt) // can not be else block because receipt can change (see above)
  {
    result = malloc(sizeof(SRxValidationResult));
  }

  // Set result values - UNDEFINED if flag is not set! 
  result->updateID = ntohl(hdr->updateID);
  result->valType = hdr->resultType;
  result->valResult.roaResult = ((result->valType & SRX_FLAG_ROA) != 0)
                                ? hdr->roaResult
                                : SRx_RESULT_UNDEFINED;
  result->valResult.bgpsecResult = ((result->valType & VRT_BGPSEC) != 0)
                                   ? hdr->bgpsecResult
                                   : SRx_RESULT_UNDEFINED;

  // Now if it is not a receipt use the callback function and free the allocated
  // memory
  if (!receipt)
  {
    LOG(LEVEL_DEBUG, HDR "NO RECEIPT! 0x%08X\n", result->updateID);
    if (proxy->resCallback != NULL)
    {

	// Call the routers notification method
	proxy->resCallback(result, proxy->userPtr);
	sem_post(&connHandler->sem_transx); // V operation 
    }
    // Free the memory
    free(result);
  }
  else
  {
    // notify the thread waiting for the receipt
    LOG(LEVEL_DEBUG, HDR "RECEIPT! 0x%08X\n", result->updateID);
    //connHandlerNotify(proxy->connHandler);
    sem_post(&connHandler->sem_register); // V operation 
    
  }

  // Release the lock!!
  //connHandlerUnlock(proxy->connHandler);
  if(receipt){
    //connHandlerNotify(proxy->connHandler);
    LOG(LEVEL_DEBUG, HDR "HANDLER NOTIFIED FOR RECEIPT OF 0x%08X\n", result->updateID);
  }

   usleep(2000);
}

/**
 *
 * @param hdr The "Signature Notification" Header
 * @param self The instance of the connection handler.
 */
void processSignNotify(SRXPROXY_SIGNATURE_NOTIFICATION* hdr, SRxProxy* proxy)
{
  LOG(LEVEL_INFO, "processSignNotify: NOT IMPLEMENTED IN THIS PROTOTYPE!!!\n");
  // TODO: BGPSEC SIGNATURES
  // Check the API itself, signatures and data types 
  // BGPSECValData vs. BGPSECData etc.
}

/**
 * If the user of this API registered a synchNotification handler it will be 
 * called now, otherwise the request will just be logged.
 * 
 * @param hdr The "Synchronization Request" Header
 * @param self The instance of the connection handler.
 */
void processSyncRequest(SRXPROXY_SYNCH_REQUEST* hdr, SRxProxy* proxy)
{
  if (proxy->syncNotification != NULL)
  {
    proxy->syncNotification(proxy->userPtr);
  }
  else
  {
    LOG(LEVEL_INFO, "processSyncRequest: NO IMPLEMENTATION SPECIFIED FOR "
                    "syncNotification!!!\n");
  }
}

/**
 * An error is received from SRx server. 
 * 
 * @param hdr The "Synchronization Request" Header
 * @param self The instance of the connection handler.
 */
void processError(SRXPROXY_ERROR* hdr, SRxProxy* proxy)
{
  uint32_t errCode = ntohs(hdr->errorCode);
  switch (errCode)
  {
    case SRXERR_ALGO_NOT_SUPPORTED:
      RAISE_ERROR("SRx Reports the requested signature algorithm is not "
                  "supported!");
      break;
    case SRXERR_UPDATE_NOT_FOUND:
      RAISE_ERROR("SRx Reports the last delete/signature request was aborted, "
                  "the update could not be found!");
      break;
    default:
      RAISE_ERROR("SRx Reports a fatal Error(%u)!", errCode);
  }
}

/**
 * This method is called by the thread that reads the socket. This dipatcher is
 * called for each fully received packet. Here is moment where each packet will
 * be decoded.

 * @param packet   The SRXPROXY packet header.
 * @param proxyPtr The proxy that deals with this packet.
 *
 */
static void dispatchPackets(SRXPROXY_BasicHeader* packet, void* proxyPtr)
{
  SRxProxy* proxy = (SRxProxy*)proxyPtr;

  switch (packet->type)
  {
    case PDU_SRXPROXY_HELLO_RESPONSE:
      processHelloResponse((SRXPROXY_HELLO_RESPONSE*)packet, proxy);
      break;

    case PDU_SRXPROXY_GOODBYE:
      processGoodbye((SRXPROXY_GOODBYE*)packet, proxy);
      break;

    case PDU_SRXPROXY_VERI_NOTIFICATION:
      processVerifyNotify((SRXPROXY_VERIFY_NOTIFICATION*)packet, proxy);
      break;

    case PDU_SRXPROXY_SIGN_NOTIFICATION:
      processSignNotify((SRXPROXY_SIGNATURE_NOTIFICATION*)packet, proxy);
      break;

    case PDU_SRXPROXY_SYNC_REQUEST:
      processSyncRequest((SRXPROXY_SYNCH_REQUEST*)packet, proxy);
      break;
      
    case PDU_SRXPROXY_ERROR:
      processError((SRXPROXY_ERROR*)packet, proxy);
      break;
    default:
      RAISE_ERROR("Dispatcher received an unknown packet type (%u)",
                  packet->type);
  }
}


int srx_api_sem_post(SRxProxy* proxy, SRxVerifyFlag method)
{
    int i_semval;
    ClientConnectionHandler* connHandler = 
	(ClientConnectionHandler*)proxy->connHandler;  

    if ((pthread_self() == connHandler->recvThread)
	    && ((method & SRX_FLAG_REQUEST_RECEIPT) != 0))
	method &= ~SRX_FLAG_REQUEST_RECEIPT;

    if ((method & SRX_FLAG_REQUEST_RECEIPT) != 0)
    {
	  sem_getvalue(&connHandler->sem_transx, &i_semval);
	  if( i_semval == 0 && connHandler->sem_transx.__size[0]==0)
      {
	    sem_post(&connHandler->sem_transx); // V operation
      }

	  //sem_post(&connHandler->sem_crit_sec); // V operation 
	  return 0;
    }
    else
    {

    }

    return -1;

}


