/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 * @file srx_api.h 
 * @date Created: 03/10/2010
 *
 * Secure Routing extension (SRx) client API.
 */

#ifndef __SRX_API_H__
#define __SRX_API_H__

#ifdef __cplusplus
extern "C" {
#endif

#include "shared/srx_defs.h"
#include "util/prefix.h"
#include "util/types.h"
#include "util/slist.h"

typedef enum {
  IN_LINE,     // wait until the validation is done. The validation method will.
  OUT_OF_LINE  // return right away and don't wait for the validation to return.
} SRxOpMode;

////////////////////////////////////////////////////////////////////////////////
// Callback notification functions for proxy user
////////////////////////////////////////////////////////////////////////////////

/**
 * Used to report the validation result for an update ID.
 * Called as a result of an explicit execution of verifyUpdate, or triggered
 * by a withdrawal (prefix, origin).
 *
 * @param result The SRx validation result, contains the update ID.
 * @param usrPtr Pointer to SRxProxy.userPtr provided by router / user of the
 *               API.
 */
typedef void (*ValidationReady)(SRxValidationResult* result, void* userPtr);

/**
 * Used to return the calculated signatures.
 *
 * @todo Parameters have yet to be defined
 *
 * @updId The update id the signature is calculated for
 * @param data the data containing the signature.
 * @param usrPtr Pointer to SRxProxy.userPtr provided by router / user of the
 *               API.
 */
typedef void (*SignaturesReady)(SRxUpdateID* updId, BGPSecData* data,
                                void* userPtr);

/**
 * This callback function allows the SRx to send a synchronization request to
 * the proxy.
 * 
 * @param usrPtr Pointer to SRxProxy.userPtr provided by router / user of the
 *               API.
 */
typedef void (*SyncNotification)(void* userPtr);


////////////////////////////////////////////////////////////////////////////////
// Proxy type and functions
////////////////////////////////////////////////////////////////////////////////

/** The data structure of the proxy. DO NOT change the settings, this is done
 * within the proxy implementation.
 */
typedef struct {
  ValidationReady  resCallback;
  SignaturesReady  sigCallback;
  SyncNotification syncNotification;

  uint32_t         proxyID;
  SRxOpMode        mode;

  void*            connHandler;  // MUST be of type ClientConnectionHandler

  uint32_t         proxyAS;
  SList            peerAS;          // Maybe change to a pointer

  // Something meaningful for the software that uses this API This pointer will
  // be used only for the callback
  void*            userPtr;
} SRxProxy;


/**
 * Creates a proxy instance and registers the callback methods for signature
 * generation and validation notification..
 *
 * @param vallidationReadyCallback The function the proxy calles to communicatre
 *                   the validation result or changes in prior validation
 *                   results to the proxy user. This can be for instance the
 *                   router or polivy module, etc.
 * @param signatureReadyCallback The function, the SRx calles once the requested
 *                   signature is generated.
 * @param requestSynchronizationCallback This function is used to request a
 *                   synchronization. A call of this method should initiate a
 *                   complete calidation request from the router's side to
 *                   guarantee that SRx has a complete view on the validated
 *                   data within the user.
 * @param proxyID    The id of the proxy used during the handshake with SRx. If
 *                   SRx is requested to generate the ID this value MUST be "0"
 *                   zero.
 * @param proxyAS    The AS of the user that uses this proxy.
 *
 * @param userPtr    This parameter is used or not by the router or policy
 *                   module or other that accesses the proxy. It does not have
 *                   any meaning for SRx or the proxy itself but might be
 *                   meaningful for the software (router, policy module, etc)
 *                   that uses this API.
 *
 * @return the instance of the SRx proxy, or NULL in case of an error
 */
SRxProxy* createSRxProxy(ValidationReady  validationReadyCallback,
                         SignaturesReady  signatureReadyCallback,
                         SyncNotification requestSynchronizationCallback,
                         uint32_t proxyID, uint32_t proxyAS, void* userPtr);
/**
 * Releases the instance. This also closes the connection if necessary.
 *
 * @param proxy The proxy instance
 */
void releaseSRxProxy(SRxProxy* proxy);

/** 
 * Add peers to the configuration of the proxy. In case the proxy is already 
 * connected, the peer change will be communicated to SRx.
 * 
 * @param proxy The proxy instance
 * @param noPeers The number of peers.
 * @param peerAS The array of peers.
 */
void addPeers(SRxProxy* proxy, uint32_t noPeers, uint32_t* peerAS);

/**
 * Remove peers from the proxy configuration. In case the proxy is already
 * connected, the peer change will be communicated to SRx.
 *
 * @param proxy The proxy instance
 * @param noPeers The number of peers.
 * @param peerAS The array of peers.
 */
void removePeers(SRxProxy* proxy, uint32_t noPeers, uint32_t* peerAS);

/**
 * Remove peers from the proxy configuration. In case the proxy is already
 * connected, the peer change will be communicated to SRx.
 *
 * @param proxy The proxy instance
 * @param keep_window The number of seconds requested to keep the data on the
 *        server side.
 * @param updateID The ID of the update to deleted.
 */
void deleteUpdate(SRxProxy* proxy, uint16_t keep_window, SRxUpdateID* updateID);

/** 
 * Tries to establish a connection to an SRx server. The given proxy instance
 * is expected to be completely configured. Any previously established
 * connection will be ended.
 *
 * In case the connection handshake fails the connection is closed and the
 * connection handler is released.
 *
 * @param proxy The proxy instance
 * @param host Server host name
 * @param port Server port address
 * @param handshakeTimeout The time in seconds before a handshake is timed out.
 *
 * @return true if connected, otherwise false if the connection failed.
 */
bool connectToSRx(SRxProxy* proxy, const char* host, int port,
                  int handshakeTimeout);

/**
 * Disconnects the proxy from the SRx Server instance on both, application and
 * transport layer.
 *
 * @param proxy The proxy instance
 * @param keepWindow Sends a request to keep the proxy data with SRx for given
 *                   amount of seconds.
 *
 * @return true if the instance is disconnected, false in case of on invalid 
 *         instance
 */
bool disconnectFromSRx(SRxProxy* proxy, uint16_t keepWindow);

/**
 * Returns if the proxy is connected or not.
 * 
 * @param proxy The proxy itself.
 * 
 * @return true if the proxy is connected, false otherwise.
 */
bool isConnected(SRxProxy* proxy);

/**
 * Verifies the given update data. All parameters except the result parameter
 * are IN parameters, result is an OUT parameter that will be filled within this
 * function. The memory MUST be allocated outside of this function.
 *
 * @param proxy The proxy instance
 * @param result This is the pointer to the result object that will be returned.
 *               This is an OUT parameter.
 * @param method Method(s) to use for the verification (ROA, BGPSEC, BOTH). The 
 *               receipt is requested is encoded in the method as well.
 * @param defaultResult The parameter contains the default information to be 
 *           used in case the validation result is not readily available.
 * @param prefix The prefix of the request. (both v4/v6 possible)
 * @param as32 Origin AS (32-bit)
 * @param bgpsec the bgpsec information.
 *
 * @return The pointer to the provided validation result or NULL if validation
 *         encountered an error. Most likely the error is that the proxy is not
 *         connected.
 */
SRxValidationResult* verifyUpdate(SRxProxy* proxy, SRxValidationResult* result,
                                  SRxVerifyFlag method, 
                                  SRxDefaultResult* defaultResult,
                                  IPPrefix* prefix, uint32_t as32,
                                  BGPSecData* bgpsec);

/**
 * This method generates a signature request. The signature will be returned
 * using the signature notification callback.
 *
 * @param proxy Pointer to the proxy instance
 * @param updId The update id the signature is requested for
 * @param onlyOwnSignature Return only the latest signature block generated for
 *                         this BGP router, otherwise the own signatures will be
 *                         concatenated to the previous given BGPSEC data block.
 * @param algorithm        The algorithm to use.
 * @param prependCounter   Number of times the own AS will be prepended into the
 *                         path. This allows traffic engineering.
 * @param peerAS           The peer AS the update will be send to.
 *
 * @return true if sending the signature request was successful.
 */
bool signUpdate(SRxProxy* proxy, SRxUpdateID* updateId, bool onlyOwnSignature,
                uint16_t algorithm, uint32_t prependCounter, uint32_t peerAS);

/* KH define */
int srx_api_sem_post(SRxProxy* proxy, SRxVerifyFlag flag);

#if 0
extern void handleSRxValidationResult (SRxValidationResult* result,
	                                       void* bgpRouter);
extern void handleSRxSignatures(SRxUpdateID* updateID, BGPSecData* data,
	                                void* bgpRouter);
extern void handleSRxSynchRequest(void* bgpRouter);
#endif

#ifdef __cplusplus
}
#endif

#endif // !__SRX_API_H__
