/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
//// OB





/**
 * Methods in this file are called by the command handler. The command handler
 * runs in its own thread and gets fed using the command queue. The command
 * queue is fed by the srx-proxy communication thread.
 *
 * @file command_handler.c
 * @date Created: 4/15/2010
 */

#include "command_handler.h"
#include "shared/srx_defs.h"
#include "shared/srx_packets.h"
#include "shared/srx_identifier.h"
#include "util/log.h"
#include "util/math.h"
#include "util/prefix.h"
#include "util/slist.h"
#include "srx_packet_sender.h"

#define HDR "([0x%08X] Command Handler): "

// Forward declaration
static void* handleCommands(void* arg);

/**
 * Registers a BGPSec Handler, RPKI Handler and Update Cache.
 *
 * @param self Variable that should be initialized
 * @param cfg The system configuration.
 * @param bgpsecHandler Existing BGPSec Handler instance
 * @param rpkiHandler Existing RPKI Handler instance
 * @param updCache Existing Update Cache instance
 * 
 * @return true if the command handler was initialized successful
 */
bool initializeCommandHandler(CommandHandler* self, Configuration* cfg,
                              ServerConnectionHandler* svrConnHandler,
                              BGPSecHandler* bgpsecHandler, 
                              RPKIHandler* rpkiHandler, UpdateCache* updCache)
{
  self->sysConfig = cfg;
  self->svrConnHandler = svrConnHandler;
  self->bgpsecHandler = bgpsecHandler;
  self->rpkiHandler = rpkiHandler;
  self->updCache = updCache;

  // Queue can be changed every time 'start' is called
  self->queue = NULL;

  // 'start' has not been called
  self->numThreads = 0;

  return true; // Does not fail
}

/**
 * Frees all allocated resources.
 *
 * @param self Instance
 */
void releaseCommandHandler(CommandHandler* self)
{  
  LOG(LEVEL_DEBUG, HDR "Send goodbye to all proxies!", pthread_self());
  ServerSocket* srvSoc = &self->svrConnHandler->svrSock;  
  if (sizeOfSList(&self->svrConnHandler->clients) > 0)
  {
    SListNode*      cnode;
    ServerClient**  clientPtr;

    // Walk through the list of proxies
    FOREACH_SLIST(&self->svrConnHandler->clients, cnode) 
    {
      clientPtr = (ServerClient**)getDataOfSListNode(cnode);
      if (clientPtr != NULL) 
      {
        if (!sendGoodbye(srvSoc, *clientPtr))
        {
          LOG(LEVEL_DEBUG, HDR "Could not send packet to proxy!", 
                           pthread_self());
        }
      }
    }
  }
  
  LOG(LEVEL_DEBUG, HDR "Command Handler released!", pthread_self());
  // Do nothing
}

bool startProcessingCommands(CommandHandler* self, CommandQueue* cmdQueue)
{
  int idx;

  self->queue = cmdQueue;
  LOG(LEVEL_DEBUG, HDR "Start Processing Commands...", pthread_self());

  for (idx = 0; idx < NUM_COMMAND_HANDLER_THREADS; idx++)
  {
    LOG (LEVEL_DEBUG, HDR "Create command handler Thread No %u", pthread_self(),
                      idx);
    if (pthread_create(&self->threads[idx], NULL, handleCommands, self) > 0)
    {
      // Continue with less threads
      if (idx > 0)
      {
        RAISE_ERROR("Failed to initiate a command handler thread "
                    "- continuing with %d threads",
                    self->numThreads);
        break;

        // No thread
      }
      else
      {
        RAISE_ERROR("Failed to initiate a command handler thread - stopping");
        return false;
      }
    }

    self->numThreads++;
  }

  return true;
}

/**
 * This method stops all command handler thread loops.
 *
 * @param self The command handler
 */
void stopProcessingCommands(CommandHandler* self)
{
  if (self->queue)
  {
    int idx;

    // First remove all pending commands
    removeAllCommands(self->queue);

    // Send SHUTDOWN to terminate the thread
    for (idx = 0; idx < self->numThreads; idx++)
    {
      queueCommand(self->queue, COMMAND_TYPE_SHUTDOWN, 
                   NULL, NULL, 0, 0, NULL);
    }

    // Wait until each thread terminated
    for (idx = 0; idx < self->numThreads; idx++)
    {
      pthread_join(self->threads[idx], NULL);
    }
  }
}

/**
 * This method is called when a HELLO packet is received. Here the handshake
 * will be performed. This message will send either a Hello Response or a
 * Goodbye. The TCP session itself will NOT be closed.
 *
 * @param cmdHandler The command handler instance.
 * @param item The command item to process. It contains the packet.
 *
 * @return true if the handshake was successful, otherwise false.
 */
static bool processHandshake(CommandHandler* cmdHandler, CommandQueueItem* item)
{
  SRXPROXY_HELLO* hdr  = (SRXPROXY_HELLO*)item->data;
  bool retVal          = true;
  bool sendRes         = false;

  if (ntohs(hdr->version) != SRX_PROTOCOL_VER)
  {
    RAISE_ERROR("Received Hello packet is of protocol version %u but expected "
                "is a Hello packet of protocol version %u", 
                ntohs(hdr->version), SRX_PROTOCOL_VER);
    sendGoodbye(item->serverSocket, item->client);
    retVal = false;
  }
  else
  {
    // Lookup the proxy identifier, accept if not existent or send an error is
    // it exists and is active.
    uint32_t usedID, proposedID = ntohl(hdr->proxyIdentifier);
    usedID = setServerClientID(item->serverSocket, item->client, proposedID);   
    if ((usedID != proposedID) && (proposedID != 0))
    {
      LOG(LEVEL_INFO, "The proposed Proxy ID [0x%08X] is already in use!");      
      sendError(SRXERR_DUPLICATE_PROXY_ID, item->serverSocket, item->client);
      sendGoodbye(item->serverSocket, item->client);
      retVal = false;      
    }
    else
    {
      sendRes = sendHelloResponse(item->serverSocket, item->client, usedID);    
      
      if (cmdHandler->sysConfig->syncAfterConnEstablished)
      {
        LOG(LEVEL_DEBUG, HDR "The configuration requires a sync request to be "
                             "send after establishing an SRx/proxy connection!",
                             pthread_self());
        if (!sendSynchRequest(item->serverSocket, item->client))
        {
          RAISE_SYS_ERROR("Could not send the synchronization request!");
        }
      }
    }
  }
  LOG(LEVEL_DEBUG, HDR "Sending \'%s\' back: \'%s\'", pthread_self(), 
                                                   (retVal ? "Hello Response"
                                                           : "Goodbye"), 
                                                   (sendRes ? "OK" : "Failed"));
  
  return retVal;
}

/**
 * The update did not have any result stored. This means that the update was
 * not yet validates using RPKI. This will be started here.
 * 
 * @param rpkiHandler The RPKI validation handler
 * @param updId The update ID
 * @param prefix The prefix of the update
 * @param originAS The origin AS of the update
 * @param srxRes The result container.
 * 
 * @return true if the validation could be performed, otherwise false.
 */
static bool verifyUpdateViaRPKI(RPKIHandler* rpkiHandler, SRxUpdateID* updateID,
                                IPPrefix* prefix, uint32_t originAS, 
                                SRxResult* srxRes)
{
  RAISE_ERROR("\n\nCALL DIRECTLY requestPrefixOriginValidation IN RPKI_CACHE!!! ");
  return true;
}

/**
 * The update did not have any result stored. This means that the update was
 * not yet validated using BGPSEC. This will be started here.
 *
 * @param rpkiHandler The RPKI validation handler
 * @param updId The update ID
 * @param srxRes The result container.
 *
 * @return true if the validation could be performed.
 */
static bool verifyViaBGPSEC(void* bgpsecHandler, SRxUpdateID updateID,
                            SRxResult* srxRes, SRxDefaultResult* defResult)
{
  LOG(LEVEL_DEBUG, HDR " " FILE_LINE_INFO "BGPSEC Validation requested for "
                 "update [0x%08X] can not be performed, implementation for this"
                 " does not exist yet!", pthread_self(), updateID);  
  srxRes->bgpsecResult = defResult->result.bgpsecResult;
  return true;
}

/**
 * This method retrieved the update id of 0 in case of an error.
 *
 * @param bhdr The srx-proxy packet containing the verification request header.
 * @param bgpsecLength Points to the position of the bgpsec blob length within
 *                   the provided SRXPROXY data header. This is an OUT parameter
 *                   and any value will be overwritten.
 * @param bgpsecBlob Points to the bgpsed data blob within the SRXCPROXT data
 *                   header. This parameter is an OUT parameter and will be set
 *                   within the function.
 *
 * @return The update id or 0 in case of an error. In case of an error the
 *         OUT parameters bgpsecLength and bgpsecBlob will not be touched at
 *         all.
 */
static SRxUpdateID getUpdateID(SRXRPOXY_BasicHeader_VerifyRequest* bhdr,
                               uint32_t* bgpsecLength, uint8_t** bgpsecBlob)
{
  uint32_t originAS;
  SRxUpdateID updateID = 0;

  uint32_t blobLength;
  uint8_t* blob;

  // Gather all data necessary to generate the update ID
  if (bhdr->type == PDU_SRXPROXY_VERIFY_V4_REQUEST)
  {
    SRXPROXY_VERIFY_V4_REQUEST* v4Hdr = (SRXPROXY_VERIFY_V4_REQUEST*)bhdr;
    blob = (uint8_t*)bhdr;
    // Move ahead to the blob
    blob += sizeof(SRXPROXY_VERIFY_V4_REQUEST);
    blobLength = ntohl(v4Hdr->bgpsecLength);

    originAS = ntohl(v4Hdr->originAS);

    IPPrefix* prefix = malloc(sizeof(IPPrefix));
    prefix->length = v4Hdr->prefixLen;
    prefix->ip.version = 4;
    prefix->ip.addr.v4 = v4Hdr->prefixAddress;

    // Generate the ID
    updateID = generateIdentifier(originAS, prefix, blobLength, blob);
    LOG(LEVEL_DEBUG, "(a) CommandHandler generated update id = 0x%08X (%u)", 
                    updateID, updateID);
    free(prefix);
  }
  else if (bhdr->type == PDU_SRXPROXY_VERIFY_V6_REQUEST)
  {
    SRXPROXY_VERIFY_V6_REQUEST* v6Hdr = (SRXPROXY_VERIFY_V6_REQUEST*)bhdr;
    blob = (uint8_t*)bhdr;
    blob += sizeof(SRXPROXY_VERIFY_V6_REQUEST);
    blobLength = ntohl(v6Hdr->bgpsecLength);

    originAS = ntohl(v6Hdr->originAS);

    IPPrefix* prefix = malloc(sizeof(IPPrefix));
    prefix->length = v6Hdr->prefixLen;
    prefix->ip.version = 6;
    prefix->ip.addr.v6 = v6Hdr->prefixAddress;

    // Generate the ID
    updateID = generateIdentifier(originAS, prefix, blobLength, blob);
    LOG(LEVEL_DEBUG, "(b) CommandHandler generated update id = 0x%08X (%u)", 
                    updateID, updateID);
    free(prefix);
  }
  else
  {
    RAISE_ERROR("Validation cannot be processed for PDU_TYPE[%u]. Required "
                "PDU types are: PDU_SRXPROXY_VERIFY_V4_REQUEST[%u] and "
                "PDU_SRXPROXY_VERIFY_V6_REQUEST[%u]", bhdr->type,
                PDU_SRXPROXY_VERIFY_V4_REQUEST,
                PDU_SRXPROXY_VERIFY_V6_REQUEST);
    return 0;
  }
  LOG(LEVEL_DEBUG,"Update ID: 0x%08X (%u)", updateID, updateID);

  // Set the out parameters.
  bgpsecLength = &blobLength;
  bgpsecBlob   = &blob;

  return updateID;
}

/**
 * This method is used to verify an update it is called by the command handlers
 * loop method that works through the command queue!
 * 
 * It might be that the very first time only once of the validations is 
 * requested. In such a case the ROA value will be SRx_RESULT_DONOTUSE.
 * If now the validation is requested for this type the command the update 
 * validation must be started. For instance ROA validation means an update is 
 * stored in the prefix cache.
 *
 * @param cmdHandler The command handler itself
 * @param item The command item that contains the information.
 *
 * @return false if the packet could not be processed.
 */
static bool processUpdateValidation(CommandHandler* cmdHandler,
                                    CommandQueueItem* item)
{
  SRXRPOXY_BasicHeader_VerifyRequest* bhdr;

  SRxDefaultResult defRes;
  SRxResult        srxRes;

  uint32_t bgpsecLength = 0;
  uint8_t* bgpsecBlob = NULL;

  uint32_t* originAS;
  IPPrefix* prefix;

  bhdr = (SRXRPOXY_BasicHeader_VerifyRequest*)item->data;
  // Here I know already that the packet has the correct size -> no need to
  // check it again

  // Check if updateID does already exist, otherwise calculate the update ID.
  // It only should exist if this verification request requested a receipt.
  if (item->dataID == 0)
  {
    item->dataID = getUpdateID(bhdr, &bgpsecLength, &bgpsecBlob);
    if (item->dataID == 0)
    {
      // ERROR IS ALREADY LOGGED!!
      return;
    }
  }

  // Set the Result values:
  defRes.resSourceROA    = bhdr->roaResSrc;
  defRes.resSourceBGPSEC = bhdr->bgpsecResSrc;
  defRes.result.roaResult    = bhdr->roaDefRes;
  defRes.result.bgpsecResult = bhdr->bgpsecDefRes;

  // update ID is known , query if the current result is already known
  bool verifyROA    = false;
  bool verifyBGPSEC = false;
  bool updateExist  = getUpdateResult(cmdHandler->updCache, &item->dataID, 
                                      &srxRes, &defRes);
  // If the update was found in the update cache and the default value differs
  // from the srx roa result determine if a result notification must be send to 
  // the router.
  if (updateExist)
  {        
    // Check if the default results should be kept as they are in the proxy
    bool useROADefRes =    (srxRes.roaResult == SRx_RESULT_UNDEFINED)
                        || (srxRes.roaResult == SRx_RESULT_DONOTUSE);
    bool useBGPSECDefRes =    (srxRes.bgpsecResult == SRx_RESULT_UNDEFINED)
                           || (srxRes.bgpsecResult == SRx_RESULT_DONOTUSE);
    bool sendNotification = false;
    
    // if the srx roa result differs from the default result a notification 
    // might be send.
    if (srxRes.roaResult != defRes.result.roaResult) 
    {      
      if (useROADefRes)
      {
        srxRes.roaResult = defRes.result.roaResult;
      }      
      else
      {
        sendNotification = true;
      }
    }
    if (srxRes.bgpsecResult != defRes.result.bgpsecResult)
    {
      if (!useBGPSECDefRes)
      {
        srxRes.bgpsecResult = defRes.result.bgpsecResult;
      }
      else
      {
        sendNotification = true;
      }
    }

    // In case the default value differs from result value the correct value
    // must be returned.
    if (sendNotification)
    {
      LOG(LEVEL_INFO,"The proxy provided default result of update[0x%08X] does"
                     " not match the srx result, a notification must be send!",
                     item->dataID);      

      if (!sendVerifyNotification(item->serverSocket, item->client,
                                  item->dataID, bhdr->flags, srxRes.roaResult,
                                  srxRes.bgpsecResult))
      {
        RAISE_ERROR("The Verification notification for update [0x%08X] could"
                    " not be send!");
      }
    }
    
    // It could be that the update was known but is not validated because it was
    // initially stored using validate BGPSEC only.
    if (srxRes.roaResult == SRx_RESULT_DONOTUSE)
    {
      verifyROA = ((bhdr->flags & SRX_FLAG_ROA) != 0);
    }
    // It could be that the update was known but is not validated because it was
    // initially stored using validate ROA only.    
    if (srxRes.bgpsecResult == SRx_RESULT_DONOTUSE)
    {
      verifyBGPSEC = ((bhdr->flags & SRX_FLAG_BGPSEC) != 0);
    }
  }
  else
  {    
    LOG(LEVEL_DEBUG, HDR "No Result found for Update [0x%08X]", pthread_self(),
                     item->dataID);
    
    verifyROA    = ((bhdr->flags & SRX_FLAG_ROA) != 0);
    verifyBGPSEC = ((bhdr->flags & SRX_FLAG_BGPSEC) != 0);
    // No such update in the cache yet. The method getUpdateResult by default
    // fills the srxRes value with the default value. 
    // Here we reassign it to UNDEFINED to indicate that SRx itself did not 
    // have any result value available.
    srxRes.roaResult    = verifyROA ? SRx_RESULT_UNDEFINED 
                                    : SRx_RESULT_DONOTUSE;
    srxRes.bgpsecResult = verifyBGPSEC ? SRx_RESULT_UNDEFINED
                                       : SRx_RESULT_DONOTUSE;
  }
  
  BGPSecData bgpsec;
  
  // Now check if ROA or BGPSEC must be validated using rpki-cacke or 
  // bgpsec cache
  if (verifyBGPSEC || verifyROA)
  {    
    // If the update did not exist in the update cache, the bgpsec data needs to
    // be extracted. the update will be stored in the update cache further down
    // in the code.
    if (!updateExist)
    {
      if (bhdr->type == PDU_SRXPROXY_VERIFY_V4_REQUEST)
      {
        SRXPROXY_VERIFY_V4_REQUEST* hdr = (SRXPROXY_VERIFY_V4_REQUEST*)item->data;
        bgpsec.length = ntohl(hdr->bgpsecLength);      
        bgpsec.data = (uint8_t*)&hdr->bgpsecValData;
      }
      else if (bhdr->type == PDU_SRXPROXY_VERIFY_V6_REQUEST)
      {
        SRXPROXY_VERIFY_V6_REQUEST* hdr = (SRXPROXY_VERIFY_V6_REQUEST*)item->data;
        bgpsec.length = ntohl(hdr->bgpsecLength);
        bgpsec.data = (uint8_t*)&hdr->bgpsecValData;      
      }
      else 
      {
        // BUG SHOULD NEVER HAPPEN
        RAISE_SYS_ERROR("A validation request with invalid packet [%u]received.!",
                        bhdr->type);
        return false;
      }
    }
       
    // also if either the update did not exist in the update cache of a ROA 
    // validation must be performed using the prefix cache, the prefix 
    // information must be retrieved.
    if (!updateExist || verifyROA)
    {
      IPPrefix*  prefix  = malloc(sizeof(IPPrefix));
      uint32_t   asn;
      memset(prefix, 0, sizeof(IPPrefix));

      if (bhdr->type == PDU_SRXPROXY_VERIFY_V4_REQUEST)
      {
        SRXPROXY_VERIFY_V4_REQUEST* v4 = (SRXPROXY_VERIFY_V4_REQUEST*)item->data;
        prefix->ip.version = 4;
        prefix->length = v4->prefixLen;
        cpyIPv4Address(&prefix->ip.addr.v4, &v4->prefixAddress);
        asn = ntohl(v4->originAS);
      }
      else
      {
        SRXPROXY_VERIFY_V6_REQUEST* v6 = (SRXPROXY_VERIFY_V6_REQUEST*)item->data;
        prefix->length = v6->prefixLen;
        prefix->ip.version = 6;
        cpyIPv6Address(&prefix->ip.addr.v6, &v6->prefixAddress);
        asn = ntohl(v6->originAS);
      }
      
      // Now store the update in the update cache if it did not exist already!
      if (!updateExist)
      {
        // Now store the update so it is available once the validation algorithm 
        // returns a value.
        storeUpdate(cmdHandler->updCache, &item->dataID, prefix, asn, &srxRes, 
                    &defRes, &bgpsec);
      }
    
      // Perform ROA Validation if necessary
      if (verifyROA)
      {
        if (!requestUpdateValidation(cmdHandler->rpkiHandler->prefixCache, 
                                     &item->dataID, prefix, asn))
        {
          RAISE_SYS_ERROR( HDR "An error occurred during the validation for "
                               "update [0x%08X] within the prefix cache!", 
                          pthread_self(), item->dataID);
        }
      }
    }

    // perform the bgpsec validation if requested.
    if (verifyBGPSEC)
    {
      // VerifyViaBGPSEC will notify the update cache with the newest result.
      if (!verifyViaBGPSEC(cmdHandler->rpkiHandler, item->dataID, &srxRes,
                           &defRes))
      {
        RAISE_SYS_ERROR("Update could not be validated using BGPSEC");
        return false;
      }
    }
  }
}

/**
 * This method performs the signing of updates.
 *
 * @param cmdHandler The command handler
 * @param item The item containing all data needed to sign.
 */
static void processUpdateSigning(CommandHandler* cmdHandler,
                                 CommandQueueItem* item)
{
  // TODO Sign the data
  LOG(LEVEL_INFO, "Signing of updates is currently not supported!");
}

/**
 * This method handles the delete request for updates.
 * 
 * @param cmdHandler The command handler.
 * @param item The item containing all information needed.
 */
static void processDeleteUpdate(CommandHandler* cmdHandler,
                                CommandQueueItem* item)
{
  // TODO delete data
  LOG(LEVEL_INFO, "Delete Updates is currently not supported!");
}

/**
 * This method handles the peer change.
 * 
 * @param cmdHandler The command handler.
 * @param item The item containing all information needed.
 */
static void processPeerChange(CommandHandler* cmdHandler,
                              CommandQueueItem* item)
{
  // TODO delete data
  LOG(LEVEL_INFO, "Peer Changes are currently not supported!");
}

/**
 * This method implements the command handler loop. Once commands are added into
 * the command queue this loop will receive them and process them. Commands
 * can be added by receiving a white list entry, BGPSEC entry, as well as a
 * request or action received from the SRx proxy.
 *
 * @param arg The Command Handler
 *
 */
static void* handleCommands(void* arg)
{
  CommandHandler* cmdHandler = (CommandHandler*)arg;
  CommandQueueItem* item;
  bool keepGoing = true;
  
  generalSignalProcess();

  LOG (LEVEL_DEBUG, "([0x%08X]) > Command Handler Thread started!", pthread_self());  
  
  while (keepGoing)
  {
    LOG(LEVEL_DEBUG, HDR "Fetch Command ...", pthread_self());
    // Block until the next command is available for this thread
    LOG(LEVEL_DEBUG, HDR "recvLock request ...%s", pthread_self(),__FUNCTION__);

    item = fetchNextCommand(cmdHandler->queue);
    
    switch (item->cmdType)
    {
      case COMMAND_TYPE_SHUTDOWN:
        LOG(LEVEL_DEBUG, HDR "Received shutdown!", pthread_self());
        LOG(LEVEL_INFO, "SRx server shutdown...");
        keepGoing = false;
        break;
      case COMMAND_TYPE_SRX_PROXY:        
        if ((item->dataLength == 0) && (item->data == NULL))
        {
          RAISE_ERROR("SRX-PROXY command but no data! [%u].");
          // Don't stop, just pass on this command.
        }
        else
        {
          SRXPROXY_BasicHeader* bhdr = (SRXPROXY_BasicHeader*)item->data;
          // Logging
          LOG(LEVEL_DEBUG, HDR "SRXPROXY PDU type [%u] (%s) fetched!", 
            pthread_self(), bhdr->type, 
            packetTypeToStr(bhdr->type));
          // Depending on the type
          switch (bhdr->type)
          {
            case PDU_SRXPROXY_HELLO:
              if (!processHandshake(cmdHandler, item))
              {
                RAISE_ERROR("Handshake between SRx and proxy failed. Shutdown "
                            "TCP connection!");

                closeClientConnection(&cmdHandler->svrConnHandler->svrSock, 
                                      item->client);
              }
              break;
            case PDU_SRXPROXY_VERIFY_V4_REQUEST:
            case PDU_SRXPROXY_VERIFY_V6_REQUEST:
              processUpdateValidation(cmdHandler, item);
              break;
            case PDU_SRXPROXY_SIGN_REQUEST:
              processUpdateSigning(cmdHandler, item);
              break;
            case PDU_SRXPROXY_GOODBYE:
              closeClientConnection(&cmdHandler->svrConnHandler->svrSock,
                                    item->client);
              printf("GoodBye!\n");
              break;
            case PDU_SRXPROXY_DELTE_UPDATE:
              processDeleteUpdate(cmdHandler, item);
              break;
            case PDU_SRXPROXY_PEER_CHANGE:
              processPeerChange(cmdHandler, item);
              break;
            default:
              RAISE_ERROR("Unknown/unsupported pdu type: %d",
                          item->dataID);
              sendError(SRXERR_INVALID_PACKET, item->serverSocket,
                        item->client);
              sendGoodbye(item->serverSocket, item->client);
              closeClientConnection(&cmdHandler->svrConnHandler->svrSock,
                                    item->client);
          }
        }
        break;
      default:
        RAISE_ERROR("Unknown Command Handler Command! [%u].", item->cmdType);
      // Still keep going.
    }

    if (cmdHandler->queue->nextItemNode != NULL)
    {
      LOG(LEVEL_DEBUG, HDR "recv Unlock request ...%s", pthread_self(), 
                       __FUNCTION__);
      //unlockMutex(&cmdHandler->queue->recvMutex);
      //signalCond(&cmdHandler->queue->recvCond);
    }
    //}
/*
    else
    {
	   
	RAISE_ERROR("receiving new packet has overwritten Command Handler item!.");
	RAISE_ERROR("Handshake between SRx and proxy failed. Shutdown "
		"TCP connection!");
     
	//unlockMutex(&cmdHandler->queue->recvMutex);
	//signalCond(&cmdHandler->queue->recvCond);

	//sendError(SRXERR_INVALID_PACKET, item->serverSocket, item->client);
	//sendGoodbye(item->serverSocket, item->client);

	//closeClientConnection(&cmdHandler->svrConnHandler->svrSock, item->client);
    }
*/
    //unlockMutex(&cmdHandler->queue->recvMutex);
    //signalCond(&cmdHandler->queue->recvCond);

    // Now remove the item from command handler. it is processed.
    deleteCommand(cmdHandler->queue, item);
  }

  LOG (LEVEL_DEBUG, "([0x%08X]) < Command Handler Thread stopped!", pthread_self());  

  pthread_exit(0);
}


/**
 * Sends a (new) result to all connected clients.
 *
 * @param self Instance
 * @param updId Update identifier
 * @param res Result
 * @return true if the broadcast could be successfully send.
 */
bool broadcastResult(CommandHandler* self, SRxValidationResult* valResult)
{
  SRxUpdateID updateId;
  
  SRXPROXY_VERIFY_NOTIFICATION* pdu;
  uint32_t pduLength = sizeof(SRXPROXY_VERIFY_NOTIFICATION);
  bool retVal = true;
  
  
  pdu = malloc(pduLength);
  pdu->type         = PDU_SRXPROXY_VERI_NOTIFICATION;
  pdu->resultType   = (valResult->valType & SRX_FLAG_ROA_AND_BGPSEC);
  pdu->roaResult    = valResult->valResult.roaResult;
  pdu->bgpsecResult = valResult->valResult.bgpsecResult;
  
  pdu->length           = htonl(pduLength);
  pdu->updateID = htonl(valResult->updateID);

  retVal = broadcastPacket(self->svrConnHandler, pdu, pduLength);
  free(pdu);
  
  return retVal;
}

