/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/*
 * @file command_queue.c
 * @date Created: 05/05/2010
 */

#include "command_queue.h"
#include "util/log.h"
#include "util/math.h"
#include "shared/srx_defs.h"
#include "shared/srx_packets.h"

#define HDR "([0x%08X] Command Queue): "

/** 
 * Initializes and setup the command queue.
 *
 * @param self Variable that should be initialized
 * 
 * @return true if the queue could be initialized.
 */
bool initializeCommandQueue(CommandQueue* self)
{
  // Create Mutex
  if (!initMutex(&self->waitMutex))
  {
    return false;
  }
  if (!initMutex(&self->writeMutex))
  {
    releaseMutex(&self->waitMutex);
    return false;
  }
  if (!initMutex(&self->recvMutex))
  {
    releaseMutex(&self->waitMutex);
    releaseMutex(&self->writeMutex);
    return false;
  }

  if(!initCond(&self->recvCond))
  {
    pthread_cond_destroy(&self->recvCond);
    return false;
  }

  // An empty list
  self->totalItems = 0;
  self->unprocessedItems = 0;
  initSList(&self->queue);

  // No item is available, i.e. block fetch
  self->nextItemNode = NULL;
  lockMutex(&self->waitMutex);

  return true;
}

void releaseCommandQueue(CommandQueue* self)
{
  if (self != NULL)
  {
    removeAllCommands(self);
   
    // Release all items and the mutexes
    releaseSList(&self->queue); 
    releaseMutex(&self->writeMutex);
    releaseMutex(&self->waitMutex);
    releaseMutex(&self->recvMutex);
  }
}

/**
 * Add a given command into the command queue. THe type of command is stored in 
 * the parameter cmdType.
 *
 * @param self The command queue where the command has to be added to
 * @param cmdType The type of the command.
 * @param svrSock The server socket
 * @param client The server client
 * @param dataID An identifier related to the data block. In case of SRX_PROXY
 *               this identifier contains either 0 or the update ID.
 * @param dataLength The length of the data attached to this command queue.
 * @param data The data package attached.
 *
 * @return true if the command could be added to the queue.
 */
bool queueCommand(CommandQueue* self, CommandQueueType cmdType,
                  ServerSocket* svrSock, ServerClient* client, uint32_t dataID,
                  uint32_t dataLength, uint8_t* data)
{
  LOG(LEVEL_INFO, HDR "queueComamnd type (%u)", pthread_self(), cmdType);
  CommandQueueItem* newItem;  
  bool              notify = false;
  //ClientThread* cthread = (ClientThread*)client;

  // Try to add a new item - make sure no one modifies the queue
  lockMutex(&self->writeMutex);
  newItem = (CommandQueueItem*)appendToSList(&self->queue, 
                                             sizeof(CommandQueueItem));
  
  // Set the new next node inside this mutex to make sure we -
  // - get the new item
  // - unlock only once
  if (self->nextItemNode == NULL)
  {
    self->nextItemNode = getLastNodeOfSList(&self->queue);
    notify = true;
  }
  unlockMutex(&self->writeMutex);

  // Failed to add an item
  if (newItem == NULL)
  {
    return false;
  }

  // 'NULL' packet
  if (data == NULL)
  {
    newItem->data = NULL;
  } 
  else
  {
    // Try to copy the 'packet' into the command item
    newItem->data = malloc(dataLength);
    if (newItem->data == NULL)
    {
      RAISE_SYS_ERROR("Not enough memory to copy the data into the queue");
      deleteFromSList(&self->queue, newItem);
      
      return false;
    }
    if (dataLength < 1000) // test
	memcpy(newItem->data, data, dataLength); // SEGV due to dataLength : 50529027 (0x03030303)
    else{
      RAISE_SYS_ERROR("too big length due to transmission error --KH--");
      deleteFromSList(&self->queue, newItem);
    }

  }

  // Set the other item members
  newItem->serverSocket = svrSock;
  newItem->client       = client;
  newItem->cmdType      = cmdType;
  newItem->dataID       = dataID;
  newItem->dataLength   = dataLength;
    
  self->totalItems++;
  self->unprocessedItems++;
    
  // Now that there is data - unlock eventually waiting clients
  if (notify)
  {
    LOG(LEVEL_DEBUG, HDR "recvLock request ...%s", pthread_self(),__FUNCTION__);
    unlockMutex(&self->waitMutex);
    //lockMutex(&self->recvMutex);

    //wait until 'fetch --> doing some jobs'
//    waitCond(&self->recvCond, &self->recvMutex);

	    
    LOG(LEVEL_DEBUG, HDR "recv UNLock request ...%s", pthread_self(),__FUNCTION__);
    //unlockMutex(&self->recvMutex);
  }

  return true;
}

/**
 * REtrieves the next command. This method DOES NOT clear the memory. 
 * After a command is processed the method 'deleteCommand' will remove it from 
 * the queue and free up all associated memory.
 * 
 * @param self The command queue
 * 
 * @return The command queue command.
 *  */
CommandQueueItem* fetchNextCommand(CommandQueue* self)
{
  CommandQueueItem* item;
 
  // Block until a command is available
  LOG(LEVEL_DEBUG, HDR "Fetch Command: Wait for command to arrive in "
                        "command queue...", pthread_self());
  lockMutex(&self->waitMutex);
  LOG(LEVEL_DEBUG, HDR "Fetch Command: ... command arrived in command queue!",
                       pthread_self());
  //lockMutex(&self->recvMutex);

  // Retrieve the item and set the fetcher to the next one.
  item = (CommandQueueItem*)self->nextItemNode->data;  // SUSPICIOUS PART --KH-- 
  self->unprocessedItems--;				// SEE also ./util/slist.c:106
  self->nextItemNode = getNextNodeOfSListNode(self->nextItemNode);

  // Only let the others wait if there is no other command
  if (self->nextItemNode != NULL)
  {
    unlockMutex(&self->waitMutex);
  }

  return item;
}

/**
 * Remove the queue element from the list and frees up all allocated memory 
 * associated with this element.
 * 
 * @param self The command queue
 * @param item The item. It also will be freed!
 */
void deleteCommand(CommandQueue* self, CommandQueueItem* item)
{
  LOG(LEVEL_DEBUG, HDR "Dequeue the given command queue item.", pthread_self());
  lockMutex(&self->writeMutex);
  // Free The packet data within the item
  if(item!=NULL) 
      free(item->data);
  self->totalItems--;
  // The delete from list also frees the item.
  deleteFromSList(&self->queue, item);
  unlockMutex(&self->writeMutex);
}

/**
 * Clears the complete queue.
 * 
 * @param self The command queue.
 */
void removeAllCommands(CommandQueue* self)
{
  LOG(LEVEL_DEBUG, HDR "Remove all commands from the command queue.",
                   pthread_self());
  SListNode* currNode;

  // No adding or single removing allowed
  lockMutex(&self->writeMutex);

  // Release all packets stored in the items
  FOREACH_SLIST(&self->queue, currNode)
  {
    void* p = ((CommandQueueItem*)getDataOfSListNode(currNode))->data;
    if (p)
    {
      free(p);
    }
  }
  // Now delete the items.
  emptySList(&self->queue);

  if (self->nextItemNode != NULL)
  {
    RAISE_SYS_ERROR("Queue[0x%08X] Removing of all commands without success!", 
                    self);
  };
  
  self->totalItems       = self->queue.size;
  self->unprocessedItems = self->nextItemNode == NULL ? 0 : self->totalItems;
  
  // Grant write access again
  unlockMutex(&self->writeMutex); 
}

/**
 * Return the maximum number of commands in the queue
 * 
 * @param self The command queue
 * 
 * @return the maximum number of items in the queue.
 */
inline int getTotalQueueSize(CommandQueue* self)
{
  return self->totalItems;
}

/**
 * Return the number of unprocessed commands in the queue
 * 
 * @param self The command queue
 * 
 * @return the number of unprocessed items in the queue.
 */
inline int getUnprocessedQueueSize(CommandQueue* self)
{
  return self->unprocessedItems;
}
