/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 * @file configuration.c
 * @date Created: 12/23/2009
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <getopt.h>
#include <libconfig.h>
#include "configuration.h"
#include "util/log.h"

#define CFG_PARAM_LOGLEVEL  1
#define CFG_PARAM_SYSLOG    2
#define CFG_PARAM_RPKI_HOST 3
#define CFG_PARAM_RPKI_PORT 4

#define CFG_PARAM_BGPSEC_HOST 5
#define CFG_PARAM_BGPSEC_PORT 6

#define HDR "([0x%08X] Configuration): "

/** Supported short options */
static const char* _SHORT_OPTIONS = "hvls:pcP::::";

/** The commandline options (\c getopt_long format) */
static const struct option _LONG_OPTIONS[] = {
  { "help",         no_argument, NULL, 'h'},
  
  { "verbose",      no_argument, NULL, 'v'},
  { "loglevel",     required_argument, NULL, CFG_PARAM_LOGLEVEL},
  { "sync",         no_argument, NULL, 's'},
  { "log",          required_argument, NULL, 'l'},
  { "syslog",       no_argument, NULL, CFG_PARAM_SYSLOG},
  
  { "port",             required_argument, NULL, 'p'},
  { "console.port",     required_argument, NULL, 'c'},
  { "console.password", required_argument, NULL, 'P'},
  
  { "rpki.host",    required_argument, NULL, CFG_PARAM_RPKI_HOST},
  { "rpki.port",    required_argument, NULL, CFG_PARAM_RPKI_PORT},
  
  { "bgpsec.host",  required_argument, NULL, CFG_PARAM_BGPSEC_HOST},
  { "bgpsec.port",  required_argument, NULL, CFG_PARAM_BGPSEC_PORT},
  { NULL, 0, NULL, 0}
};

/** Commandline help text */
static const char* _USAGE_TEXT =
  "[--options|-hvlspcP]\n\n"
  "Options:\n"
  "  -h, --help                   Display this help and exit\n"
  "  -v, --verbose                Enable verbode output\n"
  "      --loglevel               The log level for the verbose output.\n"
  "                               (3)=ERROR, (4)=WANRING, (5)=NOTICE,\n"
  "                               (6)=INFO, (7)=DEBUG\n"
  "  -l, --log <file>             Write all messages to a file\n"
  "      --syslog                 Send all messages to syslog\n"
  "  -s  --sync                   Send synchronization request each time a\n"
  "                               proxy connection is established!\n"
  "  -p, --port <no>              Use a different listening port (def.: 17900)\n"
  "  -c, --console.port <no>      Use a different console port (def.: 17901)\n"
  "  -P, --console.password <pwd> Password for remote shutdown.\n"
  "      --rpki.host <name>       RPKI/Router protocol server host name\n"
  "      --rpki.port <no>         RPKI/Router protocol server port number\n"
  "      --bgpsec.host <name>     BGPSec/Router protocol server host name\n"
  "      --bgpsec.port <no>       BGPSec/Router protocol server port number\n";

void initConfiguration(Configuration* self)
{
  self->verbose  = false;
  self->loglevel = LEVEL_ERROR;
  self->syncAfterConnEstablished = false;
  self->msgDest = MSG_DEST_STDERR;
  self->msgDestFilename = NULL;
  
  self->server_port  = 17900;
  self->console_port = 17901;
  self->console_password = NULL;
  
  self->rpki_host = NULL;
  self->rpki_port = -1;
  
  self->bgpsec_host = NULL;
  self->bgpsec_port = -1;
}

void releaseConfiguration(Configuration* self)
{
  LOG(LEVEL_DEBUG, HDR "Release configuration object", pthread_self());
  if (self != NULL)
  {
    if (self->msgDestFilename != NULL)
    {
      free(self->msgDestFilename);
    }
    if (self->rpki_host != NULL)
    {
      free(self->rpki_host);
    }
    if (self->bgpsec_host != NULL)
    {
      free(self->bgpsec_host);
    }
  }
  LOG(LEVEL_DEBUG, HDR "Configuration objects released", pthread_self());
}

static bool _copyString(char** dest, char* var, const char* desc)
{
  if (*dest != NULL)
  {
    free(*dest);
  }
  *dest = strdup(var);
  if (*dest == NULL)
  {
    RAISE_SYS_ERROR("Not enough memory for %s", desc);
    return false;
  }
  return true;
}

bool parseProgramArgs(Configuration* self, int argc, const char** argv)
{
  int optc;

  optind = 0; // Reset
  while ((optc = getopt_long(argc, (char* const*)argv, 
                             _SHORT_OPTIONS, _LONG_OPTIONS, NULL)) != -1)
  {
    switch (optc)
    {
      case 'v':
        self->verbose = true;
        break;
      case 's' :
        self->syncAfterConnEstablished = true;
        break;
      case 'l':
        self->msgDest = MSG_DEST_FILENAME;
        if (optarg == NULL)
        {
          RAISE_ERROR("Log filename missing!");
          return false;
        }
        if (!_copyString(&self->msgDestFilename, optarg, "Log filename"))
        {
          return false;
        }
        break;
      case CFG_PARAM_LOGLEVEL:
        if (optarg == NULL)
        {
          RAISE_ERROR("Log level number missing!");
          return false;
        }
        self->loglevel = strtol(optarg, NULL, 10);
        if ((self->loglevel < LEVEL_ERROR) || (self->loglevel > LEVEL_DEBUG))
        {
          RAISE_SYS_ERROR("Invalid log level ('%s')", optarg);
          return false;
        }
        break;
      case CFG_PARAM_SYSLOG:
        self->msgDest = MSG_DEST_SYSLOG;
        break;
      case 'p':
        if (optarg == NULL)
        {
          RAISE_ERROR("Server port number missing!");
          return false;
        }
        self->server_port = strtol(optarg, NULL, 10);
        if (self->server_port == 0)
        {
          RAISE_SYS_ERROR("Invalid server port number ('%s')", optarg);
          return false;
        }
        break;
      case 'c':
        if (optarg == NULL)
        {
          RAISE_ERROR("Console port number missing!");
          return false;
        }
        self->console_port = strtol(optarg, NULL, 10);
        if (self->console_port == 0)
        {
          RAISE_SYS_ERROR("Invalid SRx console port ('%s')", optarg);
          return false;
        }
        break;                
      case 'P':
        if (optarg == NULL)
        {
          RAISE_ERROR("Console password missing!");
          return false;
        }
        if (!_copyString(&self->console_password, optarg, 
                        "Console remote shutdown password"))
        {
          return false;
        }
        break;
      case CFG_PARAM_RPKI_HOST:
        if (optarg == NULL)
        {
          RAISE_ERROR("Validation cache host name missing !");
          return false;
        }
        if (!_copyString(&self->rpki_host, optarg, "Validation cache host "
                                                   "name"))
        {
          return false;
        }
        break;
      case CFG_PARAM_RPKI_PORT:
        if (optarg == NULL)
        {
          RAISE_ERROR("Validation cache port number missing!");
          return false;
        }
        self->rpki_port = strtol(optarg, NULL, 10);
        if (self->rpki_port == 0)
        {
          RAISE_SYS_ERROR("Invalid validation cache server port ('%s')",
                          optarg);
          return false;
        }
        break;
      case CFG_PARAM_BGPSEC_HOST:
        if (optarg == NULL)
        {
          RAISE_ERROR("BGPSEC certificate cache host name missing!");
          return false;
        }
        if (!_copyString(&self->bgpsec_host, optarg,
                         "BGPSec certificate cache host name"))
        {
          return false;
        }
        break;
      case CFG_PARAM_BGPSEC_PORT:
        if (optarg == NULL)
        {
          RAISE_ERROR("BGPSEC certificate cache port number missing!");
          return false;
        }
        self->bgpsec_port = strtol(optarg, NULL, 10);
        if (self->bgpsec_port == 0)
        {
          RAISE_SYS_ERROR("Invalid BGPSec certificate cache server port "
                          "(\'%s\')", optarg);
          return false;
        }
        break;
      default:
        RAISE_ERROR("Usage: %s %s", argv[0], _USAGE_TEXT);
          return false;
    }
  }

  return true;
}

bool readConfigFile(Configuration* self, const char* filename)
{
  bool ret = false; // By default something went wrong
  config_t cfg;
  config_setting_t* sett;
  const char* strtmp;
  bool useSyslog;

  // Initialize libconfig
  config_init(&cfg);

  // Try to parse the configuration file
  if (!config_read_file(&cfg, filename))
  {
    RAISE_ERROR("Unknown or invalid configuration file: %s (line %d) - %s",
                config_error_file(&cfg), config_error_line(&cfg),
                config_error_text(&cfg));
    goto free_config;
  }

  // Global & server settings
  (void)config_lookup_bool(&cfg, "verbose", (int*)&self->verbose);
  (void)config_lookup_int(&cfg, "port", &self->server_port);
  (void)config_lookup_bool(&cfg, "sync", (int*)&self->syncAfterConnEstablished);

  // Global - message destination
  (void)config_lookup_bool(&cfg, "syslog", (int*)&useSyslog);
  (void)config_lookup_int(&cfg, "loglevel", &self->loglevel);
  if (config_lookup_string(&cfg, "log", &strtmp) == CONFIG_TRUE)
  {
    if (useSyslog)
    {
      RAISE_ERROR("Conflicting 'log' and 'syslog' specified");
      goto free_config;
    }
    self->msgDest = MSG_DEST_FILENAME;
    if (!_copyString(&self->msgDestFilename, (char*)strtmp, "log. filename"))
    {
      goto free_config;
    }
  }
  else if (useSyslog)
  {
    self->msgDest = MSG_DEST_SYSLOG;
  }

  // Console
  sett = config_lookup(&cfg, "console");
  if (sett != NULL)
  {
    if (config_setting_lookup_string(sett, "password", &strtmp))
    {
      if (!_copyString(&self->console_password, (char*)strtmp,
                       "Console password"))
      {
        goto free_config;
      }
    }
    (void)config_setting_lookup_int(sett, "port", &self->console_port);
  }
  
  
  // RPKI
  sett = config_lookup(&cfg, "rpki");
  if (sett != NULL)
  {
    if (config_setting_lookup_string(sett, "host", &strtmp))
    {
      if (!_copyString(&self->rpki_host, (char*)strtmp,
                       "RPKI/Router host name"))
      {
        goto free_config;
      }
    }
    (void)config_setting_lookup_int(sett, "port", &self->rpki_port);
  }

  // BGPSec
  sett = config_lookup(&cfg, "bgpsec");
  if (sett != NULL)
  {
    if (config_setting_lookup_string(sett, "host", &strtmp))
    {
      if (!_copyString(&self->bgpsec_host, (char*)strtmp,
                       "BGPSec/Router host name"))
      {
        goto free_config;
      }
    }
    (void)config_setting_lookup_int(sett, "port", &self->bgpsec_port);
  }

  // No errors
  ret = true;

  // Release the config structure
  free_config:
  config_destroy(&cfg);

  return ret;
}

bool isCompleteConfiguration(Configuration* self)
{
#define ERROR_IF_TRUE(COND, FMT, ...) \
    if (COND) { \
      RAISE_ERROR(FMT, ## __VA_ARGS__); \
      return false; \
    }

  ERROR_IF_TRUE(self->msgDest == MSG_DEST_FILENAME
                && self->msgDestFilename == NULL,
                "Logfile not specified given!");
  ERROR_IF_TRUE(self->console_port <= 0,
                "Invalid console port '%d'!", self->console_port);
  ERROR_IF_TRUE(self->console_password == NULL,
                "Console password is not set!");
  ERROR_IF_TRUE(self->server_port <= 0,
                "Invalid server port '%d'!", self->server_port);
  ERROR_IF_TRUE(self->rpki_host == NULL,
                "Host name of validation cache is not set!");
  ERROR_IF_TRUE(self->rpki_port <= 0,
                "Port number of validation cache is not set or invalid!");
  ERROR_IF_TRUE(self->bgpsec_host == NULL,
                "Host name of BGPSec certificate cache is not set!");
  ERROR_IF_TRUE(self->bgpsec_port <= 0,
                "Port number of BGPSec certificate cache is not set or "
                "invalid!");

  return true;
}

