/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 * @file configuration.h
 * @date Created: 12/23/2009
 *
 * Configuration.
 *
 * Uses log.h to report errors.
 */

#ifndef __CONFIGURATION_H__
#define __CONFIGURATION_H__

#include "util/types.h"

static char* DEFAULT_CONSOLE_PASSWORD = "SRxSERVER";

/**
 * Destination for messages (errors, information).
 */
typedef enum {
  MSG_DEST_STDERR = 0,  ///< Write to Stderr
  MSG_DEST_FILENAME,    ///< Write to a log. file
  MSG_DEST_SYSLOG,      ///< Write to Syslog

  NUM_MSG_DEST          ///< Number of different message destinations
} MessagesDestination;

/**
 * Current configuration.
 */
typedef struct {
  /** Verbose-mode (default: \c  false) */
  bool                  verbose;  
  /** The level of verbose output (see log levels)*/
  int                   loglevel;
  /** Where should all messages go (default: MSG_DEST_STDERR) */
  MessagesDestination   msgDest;
  /** Send a synchronization request each time after a proxy connection is 
   * established. This allows to process validation requests for updates 
   * received by a router before a connection to SRx could be established. */
  bool                  syncAfterConnEstablished;
    
  /** Set only if \c msgDest is MSG_DEST_FILENAME */
  char*                 msgDestFilename;

  // SRx Server
  /** Port the SRx server should run on (default: 17900) */
  int                   server_port;
  /** Port for the server console (default: 17901) */
  int                   console_port;
  /** The console password */
  char*                 console_password;
  
  // RPKI
  /** Host name of the RPKI/Router protocol server */
  char*                 rpki_host;    
  /** Port number of the RPKI/Router protocol server */
  int                   rpki_port;    
 
  // BGPSec
  /** Host name of the BGPSec protocol server */
  char*                 bgpsec_host;  
  /** Port number of the BGPSec protocol server */
  int                   bgpsec_port;  
} Configuration;

/**
 * Sets all values to their defaults.
 *
 * @param self Configuration instance
 */
void initConfiguration(Configuration* self);

/**
 * Frees all allocated resources.
 *
 * @param self Config. instance
 */
void releaseConfiguration(Configuration* self);

/**
 * Parses the command-line.
 *
 * @param self Config. instance that should be modified
 * @param argc Number of arguments
 * @param argv Argument vector (= values)
 * @return \c true = successful, \c false = invalid options found
 */
bool parseProgramArgs(Configuration* self, int argc, const char** argv);

/**
 * Reads a configuration file.
 *
 * @param self Config. instance that should be modified
 * @param filename Path and filename of the configuration file
 * @return \c true = read and applied successfully, \c invalid or no file
 */
bool readConfigFile(Configuration* self, const char* filename);

/**
 * Checks if all necessary values are set.
 *
 * @param self Config. instance that should be checked
 * @return \c true = yes, all necessary values are set, \c false = incomplete
 */
bool isCompleteConfiguration(Configuration* self);

#endif // !__CONFIGURATION_H__


