/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 * @file rpki_router_client.h
 * @date Created: 03/08/2010
 *
 * RPKI/Router protocol client-side implementation.
 *
 * Uses log.h for error reporting
 */

#ifndef __RPKI_ROUTER_CLIENT_H__
#define __RPKI_ROUTER_CLIENT_H__

#include <pthread.h>
#include "shared/rpki_router.h"
#include "util/client_socket.h"
#include "util/mutex.h"
#include "util/prefix.h"
#include "util/types.h"

/** Maximum allowable RPKI header. After that the rest will be dropped. */
#define RPKI_MAX_HEADER_LENGTH 102400
/** The maximum number of reconnect attempts within one connection request. */
#define MAX_RECONNECTION_ATTEMPTS 10

/**
 * Client parameter settings.
 */
typedef struct {
  // Handlers
  /**
   * This function is called for each prefix announcement / withdrawl received
   * from the RPKI validation cache.
   * 
   * @param valCacheID  This Id represents the cache. It is used to be able to 
   *                    later on identify the whitelist / roa entry in case the 
   *                    cache state changes.
   * @param cache_nonce The cache nonce entry for this data. It is be usefull
   *                    for chache nonce changes in case SRx is implementing a 
   *                    performance driven aproach.
   * @param isAnn       Indicates if this in an announcement or not.
   * @param prefix      The prefix itself. Contains the information of v4/v6
   * @param maxLen      the maximum length this whitelist/roa entry covers.
   * @param oas         The origin AS for this entry.
   * @param user        Some user data. (might be deleted later on)             // THIS MIGHT BE DELETED LATER ON
   */
  void (*prefixCallback)(uint32_t valCacheID, uint16_t cache_nonce,
                         bool isAnn, IPPrefix* prefix, uint16_t maxLen,
                         uint32_t oas, void* user);


  /**
   * The cache/server sent a reset response. Usually, the client should reset 
   * his own cache.
   * 
   * @note There is no need to send a reset query - this was done already
   * The cache though can reestablish the connection similar to a cache nonce
   * change.
   *
   * @param user User data
   */
  void (*resetCallback)(uint32_t valCacheID,  void* user);

  /**
   * This method is called in case of a cache nonce change. This normally
   * requires a total cache reset and start from new. After this method a
   * RESET QUERY will be called and once all data are received the method
   * cacheNonceEstablished will be called! (if provided!).
   *
   * Process flow:
   *
   * Change od the NONCE field:
   *
   * (1) cacheNonceChanged
   * (2) - RESET QUERY and refill the cache
   *     - or abbandon the cache completely
   * (3) cacheNonceEstablished
   *
   *
   * @param calCacheID The id of the cache whose nonce changed.
   * @param newNonce The new cache nonce.
   */
  void (*cacheNonceChangedCallback)(uint32_t valCacheID, uint16_t newNonce);

  /**
   * This method will be cahhed after a cache nonce change and reset query
   * was executed. This method also will be called if the decission was
   * made to abbandon the cache after a cache nonce change.
   *
   * @param calCacheID The id of the cache whose nonce changed.
   * @param newNonce The new cache nonce.
   */
  void (*cacheNonceEstablishedCallback)(uint32_t valCacheID, uint16_t newNonce);


  /**
   * An error report was received. 
   *
   * @note Optional - can be \c NULL
   *
   * @param errNo Error number all except 2 are fatal
   * @param msg Error message
   * @param user User data
   * @return \c true = keep the connection, \c false = close connection
   */ 
  bool (*errorCallback)(uint16_t errNo, const char* msg, void* user);

  /**
   * The connections was lost. The client will try to reconnect.
   *
   * @note Optional - can be \c NULL
   *
   * @param user User data
   * @return \c -1 = do not reconnect, otherwise wait sec and then try to
   *      reconnect to the server
   */
  int (*connectionCallback)(void* user);
  // Server connection

  /** Set this variable to the server host name - not IP. */
  const char* serverHost;
  /** Set this variable to the server port number. */
  int         serverPort;
} RPKIRouterClientParams;

/**
 * A single client.
 *
 * @note Do not modify any of the variables!
 */
typedef struct {
  /** This id MUST be unique within the application. */
  uint32_t                 routerClientID;

  /** The Parameters of this clinet */
  RPKIRouterClientParams*  params;
  /** */
  void*                    user;

  /** The socet information. */
  ClientSocket             clSock;
  /** Indicates if the sonnection is stopped. */
  bool                     stop;
  /** The worker thread for this connection. */
  pthread_t                thread;
  /** The rwite mutex of this connection. */
  Mutex                    writeMutex;
  /** The last used serial number for this connection (in network order!). */
  uint32_t                 serial; // < Stored in network order
  /** The type of the previous send PDU. */
  RPKIRouterPDUType        lastSent;
  /** The type of the last received PDU. */
  RPKIRouterPDUType        lastRecv;

  // The following attributes are needed for cache nonce and possible changes
  // witin the nonce number.
  /** The cache nonce given by the validation cache (in network order!). */
  uint32_t                 cache_nonce; // < Stored in network order
  /** Indicates change in nonce_number. Once set the cache has to remove
   * all entries coming from this cache where the nonce numbers are different.*/
  bool                     cache_nonce_changed;
  /** Is needed to prevent twice reset at the befinning. If startup is true
   * the cache nonce values will be "ignored" but set to the receivied value.
   * See receivePDU for more info. */
  bool                     startup;
} RPKIRouterClient;

/**
 * Create a unique router client ID
 * 
 * @param self the router clinet the ID has to be generated for.
 * 
 * @return the ID;
 */
extern uint32_t createRouterClientID(RPKIRouterClient* self);

/**
 * Initializes a client and establishes a new connection to a server.
 * This also sends a reset query.
 *
 * @note \c All variables of \c param need to be set
 *
 * @param self Client variable that will be initialized.
 * @param params Pre-set parameters.
 * @param routerClientID an ID that will be assigned to this RPKIRouterClient.
 * @param user User data - will be passed to the callbacks
 * @return \c true = successful, \c false = something failed
 */
extern bool createRPKIRouterClient(RPKIRouterClient* client,
                                   RPKIRouterClientParams* params,
                                   void* user);

/**
 * Closes the connection, and releases all used resources.
 *
 * @param client Client instance
 */
extern void releaseRPKIRouterClient(RPKIRouterClient* client);

/**
 * Sends a reset query.
 *
 * @param client Client instance
 * @return \c true = PDU sent, \c false = sending failed
 */
extern bool sendResetQuery(RPKIRouterClient* client);

/**
 * Sends a serial query. This request all new and all expired prefixes.
 *
 * @param client Client instance
 * @return \c true = PDU sent, \c false = sending failed
 */
extern bool sendSerialQuery(RPKIRouterClient* client);

/**
 * Returns the type of the last sent PDU.
 *
 * @param client Client instance
 * @return Type
 */
extern RPKIRouterPDUType getLastSentPDUType(RPKIRouterClient* client);

/**
 * Returns the type of the last received PDU.
 *
 * @param client Client instance
 * @return Type
 */
extern RPKIRouterPDUType getLastReceivedPDUType(RPKIRouterClient* client);

int g_rpki_single_thread_client_fd;

extern void generalSignalProcess(void);

#endif // !__RPKI_ROUTER_CLIENT_H__
