/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
///////////////// OBd

/**
 * @file server_connection_handler.c
 * @date Created: 4/15/2010
 */

#include <stdio.h>
#include <stdint.h>

#include "util/log.h"
#include "shared/srx_identifier.h"
#include "shared/srx_packets.h"
#include "server_connection_handler.h"
#include "srx_packet_sender.h"

#define HDR  "([0x%08X] SrvConnHdlr): "

/**
 * Initialized the connection handler and creates a server-socket and
 * binds it to a specific port.
 *
 * @param self The connection handler to be initialized. It is required that the
 *             connection handler is already instantiated (memory is allocated)
 * @param updCache Reference to the update cache. the cache MUST NOT be accessed
 *                 for writing, only reading.
 * @param port The port SRx is listening on for proxy connections.
 * @param verbose if set more information is provided via output.
 *
 * @return true if the handler could be initialized properly, otherwise false
 */
bool createServerConnectionHandler(ServerConnectionHandler* self,
                                   UpdateCache* updCache, int port,
                                   bool verbose)
{
  bool cont = true;
  if (updCache == NULL)
  {
    RAISE_ERROR("The update cache is NULL!");
  }
  if (cont)
  {
    initSList(&self->clients);
    if (createServerSocket(&self->svrSock, port, verbose))
    {
      self->updateCache = updCache;
    }
    else
    {
      cont = false;
      releaseSList(&self->clients);
    }
  }
  return cont;
}

/**
 * Frees allocated resources.
 *
 * @param self The connection handler itself
 */
void releaseServerConnectionHandler(ServerConnectionHandler* self)
{
  releaseSList(&self->clients);
}

/**
 * This method processes the validation result request. This method is called by
 * the packet handler and if necessary the request will be added to the command
 * queue. This method does NOT send error packets to the proxy!
 *
 * @param self The server connection handler.
 * @param svrSock The server socket used to send a possible validation request
 *                receipt
 * @param client The client instance where the packet was received on
 * @param updateCache The instance of the update cache
 * @param hdr The validation request header
 *
 * @return false if an internal (fatal) error occurred, otherwise true.
 */
static bool processValidationRequest(ServerConnectionHandler* self,
                                    ServerSocket* svrSock, ServerClient* client,
                                    SRXRPOXY_BasicHeader_VerifyRequest* hdr)
{
  LOG(LEVEL_DEBUG, HDR "Enter processValidationRequest", pthread_self());
  bool     receipt = (hdr->flags & SRX_FLAG_REQUEST_RECEIPT) > 0;
  bool     addToQueue = true; // By default add into Queue

  bool      v4     = hdr->type == PDU_SRXPROXY_VERIFY_V4_REQUEST;
  IPPrefix* prefix;

  uint32_t bgpsecLength;
  uint8_t* bgpsecData = (uint8_t*)hdr+(v4 ? sizeof(SRXPROXY_VERIFY_V4_REQUEST)
                                          : sizeof(SRXPROXY_VERIFY_V6_REQUEST));
  uint32_t originAS;
  uint32_t updateID = 0;

  bool retVal = true;

  // Check if a receipt is requested. In this case, an update ID needs to
  // be generated, otherwise it is done in the command handler.
  if (receipt)
  {
    // Generate update ID

    prefix = malloc(sizeof(IPPrefix));
    prefix->length     = hdr->prefixLen;
    if (v4)
    {
      SRXPROXY_VERIFY_V4_REQUEST* v4Hdr = (SRXPROXY_VERIFY_V4_REQUEST*)hdr;
      prefix->ip.version = 4;
      prefix->ip.addr.v4 = v4Hdr->prefixAddress;
      originAS           = ntohl(v4Hdr->originAS);
      bgpsecLength       = ntohl(v4Hdr->bgpsecLength);
    }
    else
    {
      SRXPROXY_VERIFY_V6_REQUEST* v6Hdr = (SRXPROXY_VERIFY_V6_REQUEST*)hdr;
      prefix->ip.version = 6;
      prefix->ip.addr.v6 = v6Hdr->prefixAddress;
      originAS           = ntohl(v6Hdr->originAS);
      bgpsecLength       = ntohl(v6Hdr->bgpsecLength);
    }
    
    updateID = generateIdentifier(originAS, prefix, bgpsecLength, bgpsecData);
    free(prefix);

    // Now try to find if a result exists already
    SRxResult        srxRes;
    SRxDefaultResult defResInfo;
    srxRes.roaResult    = (uint8_t)SRx_RESULT_UNDEFINED;
    srxRes.bgpsecResult = (uint8_t)SRx_RESULT_UNDEFINED;

    // Copy the information of the default result.
    defResInfo.result.roaResult    = hdr->roaDefRes;
    defResInfo.result.bgpsecResult = hdr->bgpsecDefRes;
    defResInfo.resSourceROA        = hdr->roaResSrc;
    defResInfo.resSourceBGPSEC     = hdr->bgpsecResSrc;

    // if a current active result exist, do not add to queue otherwise do add
    // it.
    if (getUpdateResult(self->updateCache, &updateID, &srxRes, &defResInfo))
    {
      addToQueue = false;
    }

    // Now send the result;
    if (!sendVerifyNotification(svrSock, client, updateID, hdr->flags,
                                srxRes.roaResult, srxRes.bgpsecResult))
    {
      retVal = false;
    }
  }

  if (addToQueue)
  {
    if (!queueCommand(self->cmdQueue, COMMAND_TYPE_SRX_PROXY, svrSock, client, 
                      updateID, ntohl(hdr->length), (uint8_t*)hdr))
    {
      RAISE_ERROR("Could not add validation request to command queue!");
      retVal = false;
    }
  }

  LOG(LEVEL_DEBUG, HDR "Exit processValidationRequest", pthread_self());
  return retVal;
}

/**
 * This method processes the validation result request. This method is called by
 * the packet handler and if necessary the request will be added to the command
 * queue. This method does NOT send error packets to the proxy!
 *
 * @param self The server connection handler.
 * @param svrSock The server socket used to send a possible validation request
 *                receipt
 * @param client The client instance where the packet was received on
 * @param updateCache The instance of the update cache
 * @param hdr The signature request header
 *
 * @return false if an internal (fatal) error occurred, otherwise true.
 */
static bool processSignatureRequest(ServerConnectionHandler* self,
                                    ServerSocket* svrSock, ServerClient* client,
                                    SRxUpdateID updateID,
                                    SRXPROXY_SIGN_REQUEST* hdr)
{
  LOG(LEVEL_DEBUG, HDR "Enter processSignatureRequest", pthread_self());
  UpdSigResult* signResult = malloc(sizeof(UpdSigResult));
  bool complete = (hdr->blockType & SRX_PROXY_BLOCK_TYPE_LATEST_SIGNATURE) == 0;

  bool retVal = true;

  signResult = getUpdateSignature(self->updateCache, signResult,
                                  &updateID, ntohl(hdr->prependCounter),
                                  ntohl(hdr->peerAS),
                                  ntohs(hdr->algorithm), complete);

  if (signResult->containsError)
  {
    sendError(signResult->errorCode, svrSock, client);
    // If the error was that the update could not be located send a sync request
    // to synchronize the server and client.
    if (signResult->errorCode == SRXERR_UPDATE_NOT_FOUND)
    {
      retVal = sendSynchRequest(svrSock, client);
    }
  }
  //No error - Check if data is available that can be send right away
  else if (signResult->signatureBlock != NULL)
  {
    if (!sendSignatureNotification(svrSock, client, updateID,
                                   signResult->signatureLength,
                                   signResult->signatureBlock))
    {

      retVal = false;
    }
  }
  else // No data was available, add request to command handler for signing
  {
    if (!queueCommand(self->cmdQueue, COMMAND_TYPE_SRX_PROXY, svrSock, client, 
                                      updateID, ntohl(hdr->length), 
                                      (uint8_t*)hdr))
    {
      RAISE_ERROR("Could not add validation request to command queue!");
      retVal = false;
    }
  }
  free(signResult);

  LOG(LEVEL_DEBUG, HDR "Exit processSignatureRequest", pthread_self());
  return retVal;
}
                                    
/**
 * SRx receives a packet from one of the proxy clients. This method is called
 * before the command handler will see the request. This will be decided in this
 * method. In case a receipt is requested this method will generate the update
 * identifier and query the update cache, if the update has already a result.
 *
 * In case a result is found, the result is taken and will be send back to the
 * proxy together with the update ID.
 *
 * In case the update is never be seen before the default values will be
 * returned and the update validation request will be added to the command
 * queue for further processing.
 *
 * @param svrSock The server socket
 * @param client  The client instance where the packet was received on
 * @param packet  The packet itself
 * @param length  length The length of the packet received
 * @param srvConHandler The pointer to the sever connection handler.
 */
static void handlePacket(ServerSocket* svrSock, ServerClient* client,
                         void* packet, PacketLength length, 
                         void* srvConHandler)
{
  LOG(LEVEL_DEBUG, HDR "Enter handlePacket", pthread_self());
  ServerConnectionHandler* self = (ServerConnectionHandler*)srvConHandler;
  SRXPROXY_BasicHeader* bhdr;
  SRXPROXY_SIGN_REQUEST* srHdr;

  bool addToQueue = false;

  // No data?
  if (length < sizeof(SRXPROXY_BasicHeader))
  {
    RAISE_ERROR("The packet given with %u bytes does not have the minimum "
                "required length of %u bytes!", length,
                sizeof(SRXPROXY_BasicHeader));
  } 
  else
  {
    bhdr = (SRXPROXY_BasicHeader*)packet;
    switch (bhdr->type)
    {
      case PDU_SRXPROXY_VERIFY_V4_REQUEST:
      case PDU_SRXPROXY_VERIFY_V6_REQUEST:
        LOG(LEVEL_DEBUG, HDR "Received Validation Request type=%u", 
                         pthread_self(), bhdr->type);
        // the following method will add the command to the queue if
        // necessary therefore keep it false here so it won't be added twice.
        // This is done because within this process SRx calculates already the
        // UpdateID and adds it to the command item.
        if (!processValidationRequest(self, svrSock, client,
                                  (SRXRPOXY_BasicHeader_VerifyRequest*)packet))
        {
          sendError(SRXERR_INTERNAL_ERROR, svrSock, client);
          sendGoodbye(svrSock, client);
        }
        break;
      case PDU_SRXPROXY_SIGN_REQUEST:
        // the following method will add the command to the queue if
        // necessary
        srHdr = (SRXPROXY_SIGN_REQUEST*)packet;
        LOG(LEVEL_DEBUG, HDR "Received signature request fore update [0x%08X]", 
                         pthread_self(), ntohl(srHdr->updateIdentifier));
        if (!processSignatureRequest(self, svrSock, client,
                                     ntohl(srHdr->updateIdentifier), srHdr))
        {
          sendError(SRXERR_INTERNAL_ERROR, svrSock, client);
          sendGoodbye(svrSock, client);
        }
        break;
      default:
        if (bhdr->type <= PDU_SRXPROXY_UNKNOWN)
        {
          // This deals with peer changes, update deletes, etc.
          addToQueue = true;
        }
        else
        {
          RAISE_ERROR("Invalid srx pdu [\'%u\'] received", bhdr->type);
        }
    }

    // Add to command queue if a valid type and necessary
    if (addToQueue)
    {
      // Whatever SRX packet except validation and signature request. It will
      // be added to the command queue for further processing.
      queueCommand(self->cmdQueue, COMMAND_TYPE_SRX_PROXY, svrSock, client,
                   0, length, (uint8_t*)packet);
    } 
  }
  LOG(LEVEL_DEBUG, HDR "Exit handlePacket", pthread_self());
}

/**
 * This method is called as soon as a TCP connection is established to the 
 * SRx client proxy.
 * 
 * @param svrSock The server socket the connection was accepted on.
 * @param client The thread that deals with this connection or NULL id the 
 *               connection is a multi user connection. (later one is currently
 *               not supported!)
 * @param fd The file descriptor if the client socket/
 * @param connected If the session is connected.
 * @param user The server connection handler itself.
 * 
 * @return false in case an error occurred.
 */
static bool handleStatusChange(ServerSocket* svrSock, ServerClient* client, 
                               int fd, bool connected, void* serverConnHandler) 
{
  LOG(LEVEL_DEBUG, HDR "Enter handleStatusChange", pthread_self());
  ServerConnectionHandler* self = (ServerConnectionHandler*)serverConnHandler;
  CommandQueue* cmdQueue = (CommandQueue*)self->cmdQueue;
  bool retVal = true;

  // A new client
  if (connected) 
  {
    // Add the new client to the client list.
    if (!appendDataToSList(&self->clients, client))
    {
      RAISE_SYS_ERROR("Not enough memory to handle the new connection!");
      retVal = false;
    }
  }
  else
  {
    LOG(LEVEL_DEBUG, HDR "Lost a connection [0x%08X], remove it from client "
                         "list!", pthread_self(), client);
    deleteFromSList(&self->clients, client);
    signalCond(&cmdQueue->recvCond);
  }
//    ServerClient** sc = (ServerClient**)appendToSList(&self->clients, 
//                                                      sizeof(ServerClient*));
//    if (sc == NULL) 
//    {
//      RAISE_SYS_ERROR("Not enough memory to change the client list");
//      retVal = false;
//    } 
//    else 
//    {
//      *sc = client;
//    }
//
//  // We lost a client
//  } 
//  else 
//  {
//    SListNode*      cn;
//    ServerClient**  sc;
//
//    FOREACH_SLIST(&self->clients, cn) 
//    {
//      sc = getDataOfSListNode(cn);
//      if ((sc != NULL) && (*sc == client)) 
//      {
//        deleteFromSList(&self->clients, sc);
//        break;
//      }
//    }
//  }
  LOG(LEVEL_DEBUG, HDR "Exit handleStatusChange (%s)", pthread_self(), 
                   retVal ? "true" : "false");
  return retVal;
}

void startProcessingRequests(ServerConnectionHandler* self,
                             CommandQueue* cmdQueue) 
{
  LOG(LEVEL_DEBUG, HDR "Enter startProcessingRequests", pthread_self());
  self->cmdQueue = cmdQueue;
  runServerLoop(&self->svrSock, MODE_SINGLE_CLIENT, handlePacket, 
                handleStatusChange, self);
  LOG(LEVEL_DEBUG, HDR "Exit startProcessingRequests", pthread_self());
}

void stopProcessingRequests(ServerConnectionHandler* self) 
{
  LOG(LEVEL_DEBUG, HDR "Enter stopProcessingRequests", pthread_self());
  stopServerLoop(&self->svrSock);
  LOG(LEVEL_DEBUG, HDR "Exit stopProcessingRequests", pthread_self());
}

bool broadcastPacket(ServerConnectionHandler* self, 
                     void* packet, PacketLength length) 
{
  LOG(LEVEL_DEBUG, HDR "Enter broadcastPacket", pthread_self());
  bool retVal = true;
  if (sizeOfSList(&self->clients) > 0) 
  {
    SListNode*    cnode;
    ServerClient* clientPtr;

    FOREACH_SLIST(&self->clients, cnode) 
    {
      clientPtr = (ServerClient*)getDataOfSListNode(cnode);
      if (clientPtr != NULL) 
      {
        if (!sendPacketToClient(&self->svrSock, clientPtr,
                                packet, (size_t)length)) 
        {
          LOG(LEVEL_DEBUG, HDR "Could not send the packet to the client", 
                           pthread_self());
          retVal = false;
          break;
        }
      }
    }
  }
  LOG(LEVEL_DEBUG, HDR "Exit broadcastPacket", pthread_self());

  return true;
}


