/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 * @file server_connection_handler.h
 * @date Created: 4/8/2010
 *
 * Server Connection Handler.
 */

#ifndef __SERVER_CONNECTION_HANDLER_H__
#define __SERVER_CONNECTION_HANDLER_H__

#include "server/command_queue.h"
#include "server/update_cache.h"
#include "util/server_socket.h"
#include "util/slist.h"
#include "util/types.h"

/**
 * A single Server Connection.
 */
typedef struct {
  // The server socket for SRx - proxy communication
  ServerSocket    svrSock;
 
  // Contains the commands to be processed
  CommandQueue*   cmdQueue;

  // Contains the proxy connections
  SList           clients;

  // Newly added to allow lookups within the update cache
  // Use this for read only !!!
  UpdateCache*    updateCache;
} ServerConnectionHandler;

/**
 * Initialized the connection handler and creates a server-socket and
 * binds it to a specific port.
 *
 * @param self The connection handler to be initialized. It is required that the
 *             connection handler is already instantiated (memory is allocated)
 * @param updCache Reference to the update cache. the cache MUST NOT be accessed
 *                 for writing, only reading.
 * @param port The port SRx is listening on for proxy connections.
 * @param verbose if set more information is provided via output.
 *
 * @return true if the handler could be initialized properly, otherwise false
 */
bool createServerConnectionHandler(ServerConnectionHandler* self, 
                                   UpdateCache* updCache, int port,
                                   bool verbose);

/**
 * Frees allocated resources.
 *
 * @param self The connectio handler itself
 */
void releaseServerConnectionHandler(ServerConnectionHandler* self);

/**
 * Provides the server loop that waits for all client requests processes them.
 *
 * @note Blocking-call!
 *
 * @param self The connection handler instance
 * @param cmdQueue Existing Command Queue
 * @see stopProcessingRequests
 */
void startProcessingRequests(ServerConnectionHandler* self, 
                             CommandQueue* cmdQueue);

/** 
 * Stops processing client requests.
 * Closes all client connections and the server-socket.
 *
 * @param self Instance
 */
void stopProcessingRequests(ServerConnectionHandler* self);

/**
 * Sends a packet to all connected clients.
 *
 * @param self Instance
 * @param packet Data-stream
 * @param length Length of \c packet in Bytes
 * @return \c true = sent, \c false = failed
 */
bool broadcastPacket(ServerConnectionHandler* self,
                     void* packet, PacketLength length);

#endif // !__SERVER_CONNECTION_HANDLER_H__

