/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 * This file contains the functions to send srx-proxy packets.
 */

#include <stdint.h>
#include <stdbool.h>
#include "shared/srx_packets.h"
#include "util/log.h"
#include "util/mutex.h"
#include "util/server_socket.h"
#include "srx_packet_sender.h"

/**
 * Send a hello response to the client.
 *
 * @param proxyID The id of the proxy
 * @param srcSock The server socket
 * @param client The client who received the original message
 *
 * @return true if the packet could be send, otherwise false.
 */
bool sendHelloResponse(ServerSocket* srcSock, ServerClient* client,
                       uint32_t proxyID)
{
  bool retVal = true;
  uint32_t length = sizeof(SRXPROXY_HELLO_RESPONSE);
  SRXPROXY_HELLO_RESPONSE* pdu = malloc(length);
  memset(pdu, 0, length);

  pdu->type    = PDU_SRXPROXY_HELLO_RESPONSE;
  pdu->version = htons(SRX_PROTOCOL_VER);
  pdu->length  = htonl(length);
  pdu->proxyIdentifier = htonl(proxyID);
  
  if (!sendPacketToClient(srcSock, client, pdu, length))
  {
    RAISE_ERROR("Could not send Goodbye message!");
    retVal = false;
  }

  free(pdu);
  return retVal;
}

/**
 * Send a goodbye packet to the proxy. The proxy does not use the keepWindow,
 * therefore it will be 0
 *
 * @param srcSock The server socket
 * @param client The client who received the original message
 * @return
 */
bool sendGoodbye(ServerSocket* srcSock, ServerClient* client)
{
  bool retVal = true;
  uint32_t length = sizeof(SRXPROXY_GOODBYE);
  SRXPROXY_GOODBYE* pdu = malloc(length);
  memset(pdu, 0, length);

  pdu->type       = PDU_SRXPROXY_GOODBYE;
  pdu->keepWindow = 0; // Not used on the client side!
  pdu->length     = htonl(length);

  if (!sendPacketToClient(srcSock, client, pdu, length))
  {
    RAISE_ERROR("Could not send Goodbye message!");
    retVal = false;
  }

  free(pdu);
  return retVal;
}

/**
 * Send a verification notification
 *
 * @param svrSock The server socket
 * @param client The client of the communication.
 * @param updateID The id of the update.
 * @param resultType The type of results.
 * @param roaResult The ROA validation result.
 * @param bgpsecResult The BGPSEC validation result.
 *
 * @return true if the packet could be send, otherwise false.
 */
bool sendVerifyNotification(ServerSocket* svrSock, ServerClient* client,
                            SRxUpdateID updateID, uint8_t resultType,
                            uint8_t roaResult, uint8_t bgpsecResult)
{
  bool retVal = true;
  uint32_t length = sizeof(SRXPROXY_VERIFY_NOTIFICATION);
  SRXPROXY_VERIFY_NOTIFICATION* pdu = malloc(length);
  memset(pdu, 0, length);

  pdu->type          = PDU_SRXPROXY_VERI_NOTIFICATION;
  pdu->resultType    = resultType;
  pdu->roaResult     = roaResult;
  pdu->bgpsecResult  = bgpsecResult;
  pdu->length        = htonl(length);
  pdu->updateID      = htonl(updateID);

  pdu->length = htonl(length);

  if (!sendPacketToClient(svrSock, client, pdu, length))
  {
    RAISE_ERROR("Could not send the verify notification for update [0x%08X]!",
                updateID);
    retVal = false;
  }

  free(pdu);
  return retVal;
}

/**
 * Send a signature notification.
 *
 * @param svrSock The server socket
 * @param client The client of the communication.
 * @param updateID The id of the update.
 * @param bgpsecLength The length of the signature
 * @param bgpsecBlob The BGPSEC data blob
 *
 * @return true if the packet could be send, otherwise false.
 */
bool sendSignatureNotification(ServerSocket* svrSock, ServerClient* client,
                            SRxUpdateID updateID, uint32_t bgpsecLength,
                            uint8_t* bgpsecBlob)
{
  bool retVal = true;
  uint32_t length = sizeof(SRXPROXY_SIGNATURE_NOTIFICATION) + bgpsecLength;
  SRXPROXY_SIGNATURE_NOTIFICATION* pdu = malloc(length);
  memset(pdu, 0, length);

  uint8_t* blob = (uint8_t*)pdu + sizeof(SRXPROXY_SIGNATURE_NOTIFICATION);

  pdu->type = PDU_SRXPROXY_SIGN_NOTIFICATION;
  pdu->length = htonl(length);
  pdu->updateIdentifier = htonl(updateID);
  pdu->bgpsecLength = htonl(bgpsecLength);
  memcpy(blob, bgpsecBlob, bgpsecLength);

  // Send the pdu to the client
  if (!sendPacketToClient(svrSock, client, pdu, length))
  {
    RAISE_SYS_ERROR("Could not send the signature notification for update "
                    "[0x%08X]", updateID);
    retVal = false;
  }
  free(pdu);

  return retVal;
}


/**
 * Send a synchronization request to the proxy.
 *
 * @param svrSock The server socket
 * @param client The client of the communication.
 *
 * @return true if the packet could be send, otherwise false.
 */
bool sendSynchRequest(ServerSocket* svrSock, ServerClient* client)
{
  bool retVal = true;
  uint32_t length = sizeof(SRXPROXY_SYNCH_REQUEST);
  SRXPROXY_SYNCH_REQUEST* pdu = malloc(length);
  memset(pdu, 0, length);

  pdu->type      = PDU_SRXPROXY_SYNC_REQUEST;
  pdu->length    = htonl(length);

  // Send the pdu to the client
  if (!sendPacketToClient(svrSock, client, pdu, length))
  {
    RAISE_SYS_ERROR("Could not send the synchonization request");
    retVal = false;
  }
  free(pdu);

  return retVal;
}


/**
 * Send an error report to the proxy.
 *
 * @param errorCode The code of the error
 * @param svrSock The server socket
 * @param client The client of the communication.
 *
 * @return true if the packet could be send, otherwise false.
 */
bool sendError(uint16_t errorCode, ServerSocket* svrSock, ServerClient* client)
{
  bool retVal = true;
  uint32_t length = sizeof(SRXPROXY_ERROR);
  SRXPROXY_ERROR* pdu = malloc(length);
  memset(pdu, 0, length);

  pdu->type      = PDU_SRXPROXY_ERROR;
  pdu->errorCode = htons(errorCode);
  pdu->length    = htonl(length);

  // Send the pdu to the client
  if (!sendPacketToClient(svrSock, client, pdu, length))
  {
    RAISE_SYS_ERROR("Could not send the error report type [%0x04X]", errorCode);
    retVal = false;
  }
  free(pdu);

  return retVal;
}
