/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 * @file update_cache.c
 * @date Created: 4/15/2010
 */

#include <uthash.h>
#include <stdbool.h>
#include <stdint.h>
#include "update_cache.h"
#include "util/log.h"
#include "shared/srx_defs.h"
#include "shared/srx_packets.h"

/* Number of preallocated result slots */
//#define NUM_PREALLOC  20
#define NUM_PREALLOC  1

#define HDR "([0x%08X] UpdateCache): "

/**
 * A single update result.
 */
typedef struct {
  ////////////////////////////////////////////////////////////////////////////// TOUCHED(x); OK ( ); NOT YET (x)
  SRxUpdateID      updateID;
  UT_hash_handle   hh;  
  uint32_t         asn;
  IPPrefix         prefix;
  SRxResult        srxResult;      // The result generated by SRx
  SRxDefaultResult defaultResult;  // The result provided by verification 
                                   // request.
  uint32_t         roaRefCount;    // the number of ROA's that cover this update
  // Maybe add a link to the prefix cache update (and vice versa)
  
  // BGPSEC INFO - All below is subject to change once BGPSEC path validation 
  // will be implemented
  // Needed is a list of all peers, the signatures etc.
  uint32_t         blobLength;    // The length of the update blob
  uint8_t*         blob;          // The update blob itself.
} CacheEntry;

/*---------------------
 * Hash-table functions
 *
 * @note Uses R/W lock
 */
/**
 * This method searches the cache for the update with the given update id.
 * if found the result is written into the out pointer.
 * 
 * @param self The reference for the update cache
 * @param updateID The update ID to search for.
 * @param out the cache entry containing the update in case it was found.
 * 
 * @return true if the update was found, otherwise false.
 */
static bool tableFind(UpdateCache* self, SRxUpdateID updateID, CacheEntry** out) 
{
  ////////////////////////////////////////////////////////////////////////////// TOUCHED(x); OK ( ); NOT YET ( )
  acquireReadLock(&self->tableLock);
  HASH_FIND(hh, (CacheEntry*)self->table, &updateID, sizeof(SRxUpdateID), 
            (*out));
  unlockReadLock(&self->tableLock);

  return (*out != NULL);
}

/**
 * Add the update encapsulated in the cache entry element into the cache.
 * 
 * @param self the update cache.
 * @param cEntry the update to be stored.
 */
static void tableAdd(UpdateCache* self, CacheEntry* cEntry) 
{
  ////////////////////////////////////////////////////////////////////////////// TOUCHED(x); OK ( ); NOT YET ( );
  acquireWriteLock(&self->tableLock);
  HASH_ADD(hh, *((CacheEntry**)&self->table), updateID, sizeof(SRxUpdateID), 
           cEntry);
  unlockWriteLock(&self->tableLock);
}

/**
 * Remove the update entry from the update cache. This method does NOT release
 * the memory attached to the cache entry!!!
 * 
 * @param self The update cache.
 * @param cEntry the entry containing the update information.
 */
static void tableDel(UpdateCache* self, CacheEntry* cEntry) 
{
  ////////////////////////////////////////////////////////////////////////////// TOUCHED( ); OK ( ); NOT YET (x)
  acquireWriteLock(&self->tableLock);
  HASH_DEL(*((CacheEntry**)&self->table), cEntry);
  unlockWriteLock(&self->tableLock);
}  

/*--------
 * Exports
 */

/**
 * Initialized the update cache. The memory for the cache MUST be allocated
 * by the caller of this method.
 * 
 * @param self
 * @param chCallback
 * @return 
 */
bool createUpdateCache(UpdateCache* self, UpdateResultChanged chCallback) 
{
  ////////////////////////////////////////////////////////////////////////////// TOUCHED(x); OK ( ); NOT YET ( )
  if (!initMutex(&self->itemMutex)) 
  {
    RAISE_ERROR("Unable to setup the item Mutex");
    return false;
  }
  if (!createRWLock(&self->tableLock)) 
  {
    RAISE_ERROR("Unable to setup the hash table r/w lock");
    releaseMutex(&self->itemMutex);
    return false;
  }


  self->resChangedCallback = chCallback;
  self->table = NULL;
  self->itemsUsed = NUM_PREALLOC;

  initSList(&self->allItems);

  return true;
}

void releaseUpdateCache(UpdateCache* self) 
{
  RAISE_ERROR("Release Update Cache also should empty the cache first!");
  if (self != NULL) 
  {
    releaseRWLock(&self->tableLock);
    releaseMutex(&self->itemMutex);

    // Empty cache first
    emptyUpdateCache(self);
    
    releaseSList(&self->allItems);
  }
}

///**
// * Returns the result associated with this request. The result is stored in the
// * provided parameter "result" only for the requested method (ROA/BGPSEC).
// * This method is used byt eh command handler.
// *
// * @param self The Update Cache
// * @param defResult Contains the default data to be used in case no other data
// *                  is found.
// * @param result The structure where the result information will be stored
// *                  in. This structure is expected to have the update identifier
// *                  available as well as the valType set.
// */
//void requestUpdateResult(UpdateCache* self, SRxDefaultResult* defResult,
//                         SRxValidationResult* result)
//{
//  CacheEntry* cEntry    = NULL;
//  bool        found     = tableFind(self, result->updateIdentifier, &cEntry);
//  // Check if ROA processing is requested
//  if ((result->valType & SRX_FLAG_ROA) > 0)
//  {
//    result->valResult.roaResult = found ? cEntry->srxResult.roaResult
//                                        : defResult->result.roaResult;
//  }
//  if ((result->valType & SRX_FLAG_BGPSEC) > 0)
//  {
//    result->valResult.bgpsecResult = found ? cEntry->srxResult.bgpsecResult
//                                        : defResult->result.bgpsecResult;
//  }
//}

/**
 * Queries the update cache for the result associated with the update. This
 * method DOES NOT create a cache entry if no update was found. This method DOES
 * NOT change the update cache in any means. If already stored updates contain
 * SRx_RESULT_UNDEFINED as roa or bgpsec result, this result will be exchanged
 * with the default value.
 *
 * @param self The instance of the update cache
 * @param updId The update ID whose result is queried
 * @param srxRes The current result associated with the update This is an OUT
 *               parameter.
 * @param defResult The default result provided by proxy during verification
 *                   call.
 *
 * @return true if an update result was found, otherwise false, in this case the
 *              default value is used as srxRes.
 */
bool getUpdateResult(UpdateCache* self, SRxUpdateID* updateIDs, 
                     SRxResult* srxRes, SRxDefaultResult* defResult)
{
  // The cache entry also need the addition of source and predefined result.
  CacheEntry* cEntry;
  bool retVal = false;
  // This seems to be silly at this point but it might be that the id will 
  // become MD5 or even more. For this we accept a pointer to the structure
  // but store it as value only. See documentation for SRxUpdateID for more info
  SRxUpdateID updID = *updateIDs;

  // Look for the update
  if (tableFind(self, updID, &cEntry)) 
  {
    srxRes->roaResult    = cEntry->srxResult.roaResult == SRx_RESULT_UNDEFINED 
                           ? defResult->result.roaResult 
                           : cEntry->srxResult.roaResult;
    srxRes->bgpsecResult = cEntry->srxResult.bgpsecResult ==SRx_RESULT_UNDEFINED
                           ? defResult->result.bgpsecResult
                           : cEntry->srxResult.bgpsecResult;
    retVal = true;
  }
  else
  {
    srxRes->roaResult    = defResult->result.roaResult;
    srxRes->bgpsecResult = defResult->result.bgpsecResult;
  }

  return retVal;
}

/**
 * Stores an update in the update cache. this method returns false in case the 
 * update already exists in the update cache. In this case no further action is
 * taken.
 *
 * @param self The instance of the update cache.
 * @param updateID The ID of the Update.
 * @param prefix The prefix of the update.
 * @param asn    The AS number of the update.
 * @param result Result The update result.
 * @param defRes The default result info. This will only be taken when stored
 *               the very first time. In case the value is NULL for a first time
 *               storage, the internal UNDEFINED and UNKNOWN will be used.
 * @param bgpSec Contains BGPSEC data. This parameter as well as defRes is only
 *               used during initial storing of an update. The value can be 
 *               NULL.
 * 
 *
 * @return true the result stored, false indicates an internal error
 */
bool storeUpdate(UpdateCache* self, SRxUpdateID* updateID, IPPrefix* prefix, 
                 uint32_t asn, SRxResult* result, SRxDefaultResult* defRes,
                 BGPSecData* bgpSec)
{
  CacheEntry* cEntry;
  bool retVal;
  // This seems to be silly at this point but it might be that the id will 
  // become MD5 or even more. For this we accept a pointer to the structure
  // but store it as value only. See documentation for SRxUpdateID for more info
  SRxUpdateID updID = *updateID;

  LOG(LEVEL_DEBUG, HDR "Store update [ID:0x%08X] in update cache.", 
                   pthread_self(), updID);
  
  // Existing entry then only update the result values.
  if (tableFind(self, updID, &cEntry))
  {
    RAISE_SYS_ERROR("Update already exists in update cache!");
    retVal = false;
  }
  else
  {
    // Store a brand new update
    // New entry
    lockMutex(&self->itemMutex);

    if (self->itemsUsed == NUM_PREALLOC) 
    {
      self->availItems = appendToSList(&self->allItems, 
                                       sizeof(CacheEntry) * NUM_PREALLOC);
      if (self->availItems == NULL) 
      {
        return false;
      }
      self->itemsUsed = 0;
    }

    cEntry = (CacheEntry*)(self->availItems 
                          + (self->itemsUsed * sizeof(CacheEntry)));
    self->itemsUsed++; 
  //    unlockMutex(&self->itemMutex);

    // Set
    cEntry->updateID      = updID;
    cEntry->asn           = asn;
    cpyPrefix(&cEntry->prefix, prefix);    
    cEntry->srxResult.bgpsecResult = result->bgpsecResult;
    cEntry->srxResult.roaResult    = result->roaResult;
    if (defRes != NULL)
    {
      cEntry->defaultResult.result.roaResult    = defRes->result.roaResult;
      cEntry->defaultResult.result.bgpsecResult = defRes->result.bgpsecResult;
      cEntry->defaultResult.resSourceROA        = defRes->resSourceROA;
      cEntry->defaultResult.resSourceBGPSEC     = defRes->resSourceBGPSEC;
    }
    else
    {
      cEntry->defaultResult.result.roaResult    = SRx_RESULT_UNDEFINED;
      cEntry->defaultResult.result.bgpsecResult = SRx_RESULT_UNDEFINED;
      cEntry->defaultResult.resSourceROA        = SRxRS_UNKNOWN;
      cEntry->defaultResult.resSourceBGPSEC     = SRxRS_UNKNOWN;      
    }
    // Other Update relates data        
    // BGPSEC
    if (bgpSec != NULL)
    {
      cEntry->blobLength = bgpSec->length;
      cEntry->blob = malloc(bgpSec->length);
      memcpy(cEntry->blob, bgpSec->data, bgpSec->length);                       // This produces <incomplete sequence \\345>
    }
    else
    {
      cEntry->blobLength = 0;
      cEntry->blob = NULL;
    }    

    // Finally add the entry to cache.
    tableAdd(self, cEntry);

    unlockMutex(&self->itemMutex);  
  }
  return retVal;
}

/**
 * Stores a result for in the update cache for later retrieval.
 * If this overwrites an existing update result, then the registered 
 * UpdateResultChanged callback is called. The update MUST exist!
 *
 * @param self The instance of the update cache.
 * @param updateID The ID of the Update.
 * @param result the result the current update has to be updated with. In case
 *               the result differs from the stored update result, a 
 *               notification will be send to the client. to indicate which 
 *               result MUST NOT be modified use SRx_RESULT_DONOTUSE.
 *
 * @return true the result stored, false indicates an internal error such as the
 *              update does not exist.
 */
bool modifyUpdateResult(UpdateCache* self, SRxUpdateID* updateID, 
                        SRxResult* result)
{
  ////////////////////////////////////////////////////////////////////////////// TOUCHED(x); OK ( ); NOT YET (x)  
  CacheEntry* cEntry;
  bool retVal = true;
  // This seems to be silly at this point but it might be that the id will 
  // become MD5 or even more. For this we accept a pointer to the structure
  // but store it as value only. See documentation for SRxUpdateID for more info
  SRxUpdateID updID = *updateID;

//  RAISE_ERROR("Change signature by adding what has to be updates, roa, "
//              "bgpsec or both");
  
  // Existing entry then only update the result values.
  if (!tableFind(self, updID, &cEntry))
  {
    RAISE_SYS_ERROR("Does not exist in update cache, can not modify it!");
    retVal = false;    
  }
  else
  {
    lockMutex(&self->itemMutex);    
    
    SRxValidationResult valRes;
    valRes.updateID = updID;
    valRes.valType  = 0;
    valRes.valResult.roaResult    = cEntry->srxResult.roaResult;
    valRes.valResult.bgpsecResult = cEntry->srxResult.bgpsecResult;
    
    // Check if ROA results can be used.
    if (result->roaResult != SRx_RESULT_DONOTUSE)
    { // // Do ROA results differ ?
      if (result->roaResult != cEntry->srxResult.roaResult)
      {
        valRes.valType |= SRX_FLAG_ROA;        
        cEntry->srxResult.roaResult = result->roaResult;
        valRes.valResult.roaResult = result->roaResult;
      }
    }
    
    // Check if BGPSEC results can be used.
    if (result->bgpsecResult != SRx_RESULT_DONOTUSE)
    { // Check for changes in bgpsec result
      if (result->bgpsecResult != cEntry->srxResult.roaResult)
      {
        valRes.valType |= SRX_FLAG_BGPSEC;        
        cEntry->srxResult.bgpsecResult = result->bgpsecResult;
        valRes.valResult.bgpsecResult = result->bgpsecResult;
      }      
    }
    
    if (valRes.valType != 0)
    {
      if (self->resChangedCallback != NULL)
      {
        self->resChangedCallback(&valRes);     
      }
      else
      {
        RAISE_ERROR("No resChangedCallback function registered! "
                    "Cannot propagate the changes in the validation result!");
        retVal = false;
      }
    }
    
    unlockMutex(&self->itemMutex);
  }
  
  return retVal;
}

/**
 * Removes the update data from the list and releases all memory associated to 
 * it.
 * 
 * @note This method ONLY deletes the update from the update cache. It is 
 *       important to assure that other references such as the prefix_cache
 *       
 * @param self The instance of the update cache 
 * @param updateID The ID of the update that has to be removed.
 * @return true If the update could be removed, false if the update was not 
 *               found in the cache.
 */
bool _deleteUpdate(UpdateCache* self, SRxUpdateID* updateID) 
{
  ////////////////////////////////////////////////////////////////////////////// TOUCHED(x); OK (x); NOT YET ( ); Tested ( );
  CacheEntry* cEntry;
  bool retVal = true;
  // This seems to be silly at this point but it might be that the id will 
  // become MD5 or even more. For this we accept a pointer to the structure
  // but store it as value only. See documentation for SRxUpdateID for more info
  SRxUpdateID updID = *updateID;

  // Get the update cache entry from the update cache.
  if (tableFind(self, updID, &cEntry)) 
  {
    // Does not release the memory but only removes the hash table entry
    tableDel(self, cEntry);
    
    // Free the memory of the bgpsec;
    free(cEntry->blob);
    // Free the cache entry.
    free(cEntry);
  }
  else
  {
    LOG(LEVEL_INFO, "Delete aborted, update [0x%08X] not found!", updID);
    retVal = false;    
  }

  return retVal;
}

/**
 * This function returns the update signature if already existent. It will NOT
 * start the signing. If no signature exists the return value is NULL
 *
 * @param self The instance of the update cache
 * @param result The result of the function call.
 * @param updateID The id of the update
 * @param prependCount the prepend count of the host AS
 * @param peerAS The peer AS
 * @param algorithm The algorithm used
 * @param complete Indicates if the complete signature block or just the
 *                 new addition will be returned.
 * @return The signature block.
 */
UpdSigResult* getUpdateSignature(UpdateCache* self, UpdSigResult* result,
                                 SRxUpdateID* updateID, uint32_t prependCount, 
                                 uint32_t peerAS, uint16_t algorithm,
                                 bool complete)
{
  ////////////////////////////////////////////////////////////////////////////// TOUCHED( ); OK ( ); NOT YET (x)
  CacheEntry* cEntry;
  // This seems to be silly at this point but it might be that the id will 
  // become MD5 or even more. For this we accept a pointer to the structure
  // but store it as value only. See documentation for SRxUpdateID for more info
  SRxUpdateID updID = *updateID;
  
  
  // Currently we do NOT support any algorithm!
  RAISE_ERROR("UPDATE SIGNATURES ARE CURRENTLY NOT SUPPORTED!");

  // Currently we by default throw an algorithm not supported error.
  // This can be overwritten by update not found. Once signing is supported
  // this will be changed.
  result->containsError   = true;
  result->errorCode       = SRXERR_ALGO_NOT_SUPPORTED;
  result->signatureLength = 0;
  result->signatureBlock  = NULL;

  // Look for the update
  if (!tableFind(self, updID, &cEntry))
  {
    LOG(LEVEL_INFO, "Update [0x%08X] not found! Can not sign it!", updID);
    result->errorCode = SRXERR_UPDATE_NOT_FOUND;
  }

  return result;
}

/**
 * This method is not for usage within the update cache management. It is 
 * mainly to allow the server console to query for update information.
 * 
 * @param self The update cache 
 * @param statistics The statistics information to be filles. The value 
 *                   updateID MUST be set and will be used to locate the update.
 * 
 * @return true if the update was found, otherwise false
 */
bool getUpdateData(UpdateCache* self, UC_UpdateStatistics* statistics)
{ 
  CacheEntry* cEntry;
  bool retVal = false;
  
  if (statistics == NULL)
  {
    RAISE_SYS_ERROR("The given statistics block is NULL!");
  }
  else if (*statistics->updateID == 0)
  {
    RAISE_SYS_ERROR("The given updaetID is 0 (INVALID ID)!");
  }
  // Look for the update
  else if (tableFind(self, *statistics->updateID, &cEntry)) 
  {
    retVal = true;
    statistics->asn                           = cEntry->asn;
    cpyPrefix(&statistics->prefix, &cEntry->prefix);
    statistics->bgpsecResult.containsError    = false;
    statistics->bgpsecResult.errorCode        = 0;
    statistics->bgpsecResult.signatureBlock   = NULL;
    statistics->bgpsecResult.signatureLength  = 0;
    statistics->defResult.resSourceROA   =cEntry->defaultResult.resSourceROA;    
    statistics->defResult.resSourceBGPSEC=cEntry->defaultResult.resSourceBGPSEC;
    statistics->defResult.result.roaResult =
                                         cEntry->defaultResult.result.roaResult;
    statistics->defResult.result.bgpsecResult = 
                                      cEntry->defaultResult.result.bgpsecResult;
    statistics->result.roaResult    = cEntry->srxResult.roaResult;
    statistics->result.bgpsecResult = cEntry->srxResult.bgpsecResult;
    statistics->roa_count           = cEntry->roaRefCount;
  }  
  return retVal;
}


void emptyUpdateCache(UpdateCache* self) 
{
  ////////////////////////////////////////////////////////////////////////////// TOUCHED( ); OK ( ); NOT YET (x); Tested ( )
  acquireWriteLock(&self->tableLock);
  lockMutex(&self->itemMutex);
  
  emptySList(&self->allItems);
  unlockMutex(&self->itemMutex);

  self->table     = NULL;
  self->itemsUsed = NUM_PREALLOC;

  unlockWriteLock(&self->tableLock);
}
