/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 * @file update_cache.h
 * @date Created: 4/8/2010
 *
 * Update Cache.
 */

#ifndef __UPDATE_CACHE_H__
#define __UPDATE_CACHE_H__

#include "shared/srx_defs.h"
#include "shared/srx_packets.h"
#include "util/mutex.h"
#include "util/rwlock.h"
#include "util/slist.h"
#include "util/types.h"

/**
 * Function that is called in case a result changed.
 *
 * @param updateId Affected update
 * @param newRes New result
 */
typedef void (*UpdateResultChanged)(SRxValidationResult* result);

/**
 * A single Update Cache.
 */
typedef struct {
  UpdateResultChanged resChangedCallback;
  Mutex               itemMutex;
  SList               allItems;
  void*               availItems;
  int                 itemsUsed;
  RWLock              tableLock;
  void*               table;
} UpdateCache;

/** Return value for method getUpdateSignature the memory of this instance
 * is managed by the update cache.
 */
typedef struct {
  bool     containsError;
  uint16_t errorCode;
  uint32_t signatureLength;
  void*    signatureBlock;
} UpdSigResult;

/**
 * The sole purpose of this typedef is to allow gathering some statistical 
 * information for the server console. It is NOT used within the update cache.
 * The memory MUST be managed by the caller.
 */
typedef struct {
  SRxUpdateID*     updateID;
  uint32_t         asn;
  uint32_t         roa_count;
  SRxDefaultResult defResult;
  SRxResult        result;
  IPPrefix         prefix;
  UpdSigResult     bgpsecResult;
} UC_UpdateStatistics;

/**
 * Initializes all necessary resources, and registers a given callback in case
 * a result gets changed.
 *
 * @param self Variable that should be initialized
 * @param chCallback Function that should be called in case a result changes
 * @return \c true = successfully initialized, \c false = an error occurred
 */
bool createUpdateCache(UpdateCache* self, UpdateResultChanged chCallback);

/**
 * Frees all allocated resources.
 *
 * @param self Instance that should be released
 */
void releaseUpdateCache(UpdateCache* self);

/**
 * Queries the update cache for the result associated with the update. This
 * method DOES NOT create a cache entry if no update was found. This method DOES
 * NOT change the update cache in any means.
 *
 * @param self The instance of the update cache
 * @param updId The update ID whose result is queried
 * @param srxRes The current result associated with the update This is an OUT
 *               parameter.
 * @param defResult The default result provided by proxy during verification
 *                   call.
 *
 * @return true if an update result was found, otherwise false, in this case the
 *              default value is used as srxRes.
 */
bool getUpdateResult(UpdateCache* self, SRxUpdateID* updateID, 
                     SRxResult* srxRes, SRxDefaultResult* defaultRes);

/**
 * This function returns the update signature if already existent. It will NOT
 * start the signing. If no signature exists the return value is NULL
 *
 * @param self The instance of the update cache
 * @param result The result of the function call.
 * @param updateID The id of the update
 * @param prependCount the prepend count of the host AS
 * @param peerAS The peer AS
 * @param algorithm The algorithm used
 * @param complete Indicates if the complete signature block or just the
 *                 new addition will be returned.
 * @return The signature block.
 */
UpdSigResult* getUpdateSignature(UpdateCache* self, UpdSigResult* result,
                                 SRxUpdateID* updateID, uint32_t prependCount, 
                                 uint32_t peerAS, uint16_t algorithm,
                                 bool complete);

/**
 * This method is not for usage within the update cache management. It is 
 * mainly to allow the server console to query for update information.
 * 
 * @param self The update cache 
 * @param statistics The statistics information to be filles. The value 
 *                   updateID MUST be set and will be used to locate the update.
 * 
 * @return true if the update was found, otherwise false
 */
bool getUpdateData(UpdateCache* self, UC_UpdateStatistics* statistics);

/**
 * Stores an update in the update cache. this method returns false in case the 
 * update already exists in the update cache. In this case no further action is
 * taken.
 *
 * @param self The instance of the update cache.
 * @param updateID The ID of the Update.
 * @param prefix The prefix of the update.
 * @param asn    The AS number of the update.
 * @param result Result The update result.
 * @param defRes The default result info. This will only be taken when stored
 *               the very first time. In case the value is NULL for a first time
 *               storage, the internal UNDEFINED and UNKNOWN will be used.
 * @param bgpSec Contains BGPSEC data. This parameter as well as defRes is only
 *               used during initial storing of an update. The value can be 
 *               NULL.
 * 
 *
 * @return true the result stored, false indicates an internal error
 */
bool storeUpdate(UpdateCache* self, SRxUpdateID* updateID, IPPrefix* prefix, 
                 uint32_t asn, SRxResult* result, SRxDefaultResult* defRes,
                 BGPSecData* bgpSec);

/**
 * Removes the update data from the list and releases all memory associated to 
 * it.
 *
 * @param self The instance of the update cache 
 * @param updateID The ID of the update that has to be removed.
 * @return true If the update could be removed, false if the update was not 
 *               found in the cache.
 */
bool _deleteUpdate(UpdateCache* self, SRxUpdateID* updateID);

/**
 * Empties a cache and releases all memory attached to each of the elements.
 *
 * @note Primarily for development purposes or program shutdown.
 *
 * @param self Instance of the update cache.
 */
void emptyUpdateCache(UpdateCache* self);

#endif // !__UPDATE_CACHE_H__

