/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/*
 * This program allows to test the SRX server implementation.
 *
 * @version 0.2.0
 *
 *
 * Changelog:
 * -----------------------------------------------------------------------------
 *   0.2.0 - 2011/01/07 - oborchert
 *           * Changelog added with version 0.2.0 and date 2011/01/07
 *           * Version tag added
 *           * Fixed some output to be less confusing.
 *           * Trimmed all input.
 *   0.0.0 - 2010/05/11 - pgleichm
 *           * Code Created
 * -----------------------------------------------------------------------------
 *
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <errno.h>
#include <readline/readline.h>
#include <readline/history.h>
#include "client/srx_api.h"
#include "util/log.h"
#include "util/prefix.h"
#include "util/str.h"
// This inlude is only needed for some progam version defines. Not needed 
// otherwise.
#include "server/srx_server.h"

#define DEFAULT_SERVER    "localhost"
#define DEFAULT_PORT      17900
#define DEFAULT_PEERAS    50
#define HISTORY_FILENAME  ".user_srx_api.history"

#define PROMPT            ">> "

static SRxProxy* proxy   = NULL;
static LogLevel logLevel = LEVEL_DEBUG;

// Forward declaration
static bool processLine(bool log, char* line);

////////////////////////////////////////////////////////////////////////////////
// User input
////////////////////////////////////////////////////////////////////////////////

/**
 * Read data from either the argument pointer or a user input using stdandard 
 * input.
 * 
 * @param argPtr Contains arguments to be read
 * @param msg The message printed in case for command line input.
 * 
 * @return The string read from either the parameters or console
 */
static char* prompt(char** argPtr, const char* msg)
{
  #define BUFFER_SIZE 80
  static char buffer[BUFFER_SIZE];
  
  char* input;

  // Already a value
  if (*argPtr != NULL)
  {
    return strsep(argPtr, " \t");
  }

  // Get a value from the user
  do
  {
    input = readline(msg);
  } while (input == NULL);

  strncpy(buffer, input, BUFFER_SIZE);
  free(input);
  
  return buffer;
}

/**
 * Read data from either the argument pointer or a user input using standard
 * input.
 *
 * @param argPtr Contains arguments to be read
 * @param msg The message printed in case for command line input.
 *
 * @return a 32 bit value or 0
 */
static uint32_t promptU32(char** argPtr, const char* msg)
{
  const char* input;
  uint32_t    num;

  do
  {
    input = trim(prompt(argPtr, msg));
    if (*input == '\0')
    {
      num = 0;
      break;
    }
    errno = 0;
    num = strtoul(input, NULL, 0);
  } while (errno != 0);

  return num;
}

static void addToHistory(const char* fmt, ...)
{
  char* line;

  va_list ap;
  va_start(ap, fmt);
  vasprintf(&line, fmt, ap);
  //sprintf(&line, fmt, ap);
  va_end(ap);

  add_history(line);
  free(line);
}

////////////////////////////////////////////////////////////////////////////////
// Event handlers
////////////////////////////////////////////////////////////////////////////////

/**
 * @param updID Affected update identifier
 * @param res The SRx validation result.
 * @param userPtr a pointer to self
 */
void handleValidationResult(SRxValidationResult* res, void* userPtr)
{
  uint32_t updateID = res->updateID;
  bool isReceipt    = (res->valType & SRX_PROXY_RESTYPE_RECEIPT) != 0;
  bool printROA     = (res->valType & SRX_PROXY_RESTYPE_ROA) != 0;
  bool printBGPSEC  = (res->valType & SRX_PROXY_RESTYPE_BGPSEC) != 0;
  char* resultStr;

  printf("=> Validation result%s: update [ID:0x%08X] (%u): ",
             (isReceipt ? " receipt" : ""), updateID, updateID);

  if (printROA)
  {
    switch (res->valResult.roaResult)
    {
      case SRx_RESULT_UNDEFINED:
        resultStr = "NOT DEFINED";
        break;
      case SRx_RESULT_VALID:
        resultStr = "VALID";
        break;
      case SRx_RESULT_UNKNOWN:
        resultStr = "UNKNOWN";
        break;
      case SRx_RESULT_INVALID:
        resultStr = "INVALID";
        break;
      case SRx_RESULT_DONOTUSE:
        resultStr = "DONOTUSE <- SHOULD NOT BE HERE - SRx INTERNAL ONLY, "
                    "NEVER PROXY";
        break;        
      default:
        resultStr = "???????";
    }
    printf(" ROA=%s", resultStr);
  }

  if (printBGPSEC)
  {
    switch (res->valResult.bgpsecResult)
    {
      case SRx_RESULT_UNDEFINED:
        resultStr = "NOT DEFINED";
        break;
      case SRx_RESULT_VALID:
        resultStr = "VALID";
        break;
      case SRx_RESULT_UNKNOWN:
        resultStr = "UNKNOWN <- INVALID VALUE HERE!";
        break;
      case SRx_RESULT_INVALID:
        resultStr = "INVALID";
        break;
      case SRx_RESULT_DONOTUSE:
        resultStr = "DONOTUSE <- SHOULD NOT BE HERE - SRx INTERNAL ONLY, "
                    "NEVER PROXY";
        break;        
      default:
        resultStr = "???????";
    }
    printf("%sBGPSEC=%s", (printROA ? "; " : ""), resultStr);
  }
  printf("\n");
}

/**
 * Used to return the calculated signatures.
 *
 * @todo Parameters have yet to be defined
 *
 * @param updateID The update id the signature is calculated for
 * @param userPtr data the data containing the signature (structure TBD)
 */
void handleSignatures(SRxUpdateID* updateID, BGPSecData* bgpsec, void* usrPtr)
{
  int i;
  printf("Received a handle signature for update [0x%08X]\n", *updateID);
  printf("Data (%u bytes): ", bgpsec->length);
  for (i = 0; i < bgpsec->length; i++)
  {
    printf("%c", (unsigned char)bgpsec->data[i]);
  }
  printf("\n");
}

/*-------------
 * Call the API
 */
/**
 * Connect to the SRx server using the global configures proxy instance.
 * @param log indicates if the method logs the connection request.
 * @param argPtr The arguments for the SRx server, Host and Port. Can be kept
 * empty, then it will be read from standard in
 */
static void doConnect(bool log, char** argPtr)
{
  const char* host;
  uint32_t    port;

  int         noPeers      = 0;
  int         peerLocation = 0;
  int         size = sizeof(uint32_t);
  uint32_t*   peerAS  = malloc(size); // space for one peer
  uint32_t    peer = 0;
  uint32_t    proxyID = 0;

  bool        connected;

  // Get the parameters
  host = prompt(argPtr, "Host [default: 'localhost'] ? ");
  if (strlen(host) == 0)
  {
    host = DEFAULT_SERVER;
  }
  port = promptU32(argPtr, "Port [default: 17900] ? ");
  if (port == 0)
  {
    port = DEFAULT_PORT;
  }

  peer = promptU32(argPtr, "PeerAS [default: 50] ? ");
  while (peer != 0)
  {
    noPeers++;
    if (size < (noPeers * 4))
    {
      size = noPeers * 4;
      peerAS = realloc(peerAS, size);
    }

    peerAS[peerLocation] = peer;
    peerLocation++;
    peer = promptU32(argPtr, "PeerAS (enter for stop) ? ");
  }
  if ((noPeers == 0) && (peer == 0))
  {
    peerAS[0] = DEFAULT_PEERAS;
    noPeers++;
  }

  addPeers(proxy, noPeers, peerAS);

  if (log)
  {
    // Only the first peer is in history
    if (noPeers > 0)
    {
      addToHistory("connect %s %u %u", host, port, peerAS[0]);
    }
    else
    {
      addToHistory("connect %s %u %u 0");
    }
  }

  connected = connectToSRx(proxy, host, port, SRX_DEFAULT_HANDSHAKE_TIMEOUT);
  LOG(LEVEL_INFO,"Connected: %s\n", connected ? "Yes" : "No");
}

/**
 * Request to disconnect from the SRx server instance
 * @param log indicates if the disconnect will be logged.
 */
static void doDisconnect(bool log)
{
  if (log)
  {
    addToHistory("disconnect from SRx server");
  }
  if (disconnectFromSRx(proxy, SRX_DEFAULT_KEEP_WINDOW))
  {
    LOG(LEVEL_INFO, "Disconnected\n");
  }
  else
  {
    LOG(LEVEL_INFO, "Proxy is not connected, can not disconnect!\n");
  }
}

/**
 * Maintains the number of peers to the proxy. This method does not require an
 * active proxy connection.
 * 
 * @param log indicates if the method loggs the connection request.
 * @param argPtr The arguments for the SRx server, Host and Port. Can be kept
 * empty, then it will be read from standard in
 *
 * @param add it true the peer will be added, otherwise deleted.
 */
static void doMaintainPeer(bool log, char** argPtr, bool add)
{
  uint32_t peerNr = 0;
  uint32_t newPeer = 0;
  uint32_t size   = sizeof(uint32_t);
  uint32_t* peerAS = malloc(size);

  do {
    newPeer = promptU32(argPtr, "Peer AS (0=end): ");
    if (newPeer > 0)
    {
      if (size <= (peerNr * sizeof(uint32_t)))
      {
        size = peerNr * sizeof(uint32_t);
        peerAS = realloc(peerAS, size);
      }
      peerAS[peerNr] = newPeer;
      peerNr++;
    }
  } while (newPeer > 0);
  
  if (peerNr > 0)
  {
    if (add)
    {
      addPeers(proxy, peerNr, peerAS);
    }
    else
    {
      removePeers(proxy, peerNr, peerAS);
    }
  }
  free (peerAS);
}

/**
 * Indicates to execute a verification request.
 * @param log
 * @param argPtr
 */
static void doVerify(bool log, char** argPtr)
{
  uint8_t          method;
  SRxUpdateID      updId;
  uint32_t         as32;
  bool             doPrompt;
  const char*      prefixInput;
  const char*      bgpsecInput;
  IPPrefix         prefix;
  SRxDefaultResult defResult;
  BGPSecData       bgpsec;
  SRxValidationResult* valResult;
  char ipString[255];
  char* ipStringPtr = ipString;
  memset(ipString,'\0',255);
  uint32_t    receipt;
  // The client connection handler
//  ClientConnectionHandler* connHandler = 
//                                   (ClientConnectionHandler*)proxy->connHandler;

  if (!isConnected(proxy))
  {
    LOG(LEVEL_INFO, "-> Proxy is not connected, can not request "
                    "verification!\n");
    return;
  }
  
  ////////////////////////////////////////
  /// READ PARAMETERS EITHER FROM COMMAND PATAMETER OR COMMAND LINE
  ///////////////////////////////////////
  // Receipt
  receipt = (uint32_t)promptU32(argPtr, "(Verify) Request Receipt: "
                                        "[1=request 2=do not request] ? ");
  if (receipt < 1 || receipt > 2)
  {
    LOG (LEVEL_DEBUG, "Verify command provided invalid receipt request "
                      "value of %u! Only 1-request or 2=not request are "
                      "allowed!", receipt);
    return;
  }

  // Method
  method = (uint8_t)promptU32(argPtr, "(Verify) Method: "
                                   "[0=just store 1=ROA only, "
                                   "2=BGPSec only, 3=both] ? ");
  if (method > 3)
  {
    printf ("Invalid Method %u\n", method);
    return;
  }

  // Header data
  as32  = promptU32(argPtr, "(Verify) AS number ? ");

  // Prefix
  prefixInput = prompt(argPtr, "(Verify) IP Prefix [] ? ");
  if (!strToIPPrefix(prefixInput, &prefix))
  {
    printf("Error: Prefix [%s] is invalid\n", prefixInput);
    return;
  }
  sprintf(ipStringPtr, "%s", prefixInput);

  defResult.result.roaResult = (uint8_t)promptU32(argPtr,
                 "(Verify) DefROAVal: [0=VALID, 1=UNKNOWN, 2=INVALID, "
                 "3=Not Defined] ? ");
  if (defResult.result.roaResult > 3)
  {
    printf("Error: Default ROA result [%u] is invalid\n", 
           defResult.result.roaResult);
    return;
  }

  defResult.result.bgpsecResult = (uint8_t)promptU32(argPtr,
                         "(Verify) DefBGPSECVal [0=VALID, 2=INVALID, "
                         "3=Not Defined] ? ");
  if (   (defResult.result.bgpsecResult > 3)
      || (defResult.result.bgpsecResult == 1))
  {
    printf("Error: Default BGPSEC result [%u] is invalid\n",
           defResult.result.bgpsecResult);
    return;
  }

  defResult.resSourceROA    = SRxRS_UNKNOWN;
  defResult.resSourceBGPSEC = SRxRS_UNKNOWN;

  bgpsecInput = prompt(argPtr, "(Verify) BGPSEC some string ? ");

  if (bgpsecInput == '\0')
  {
    bgpsec.length = 0;
    bgpsec.data = NULL;
  }
  else
  {
    bgpsec.length = strlen(bgpsecInput)+1;
    bgpsec.data = (uint8_t*)bgpsecInput;
  }
  
  if (log)
  {
    if (bgpsec.length == 0)
    {
      addToHistory("verify %u %u %u %s %u %u %c" ,
                   receipt, method, as32, ipStringPtr,
                   defResult.result.roaResult, 
                   defResult.result.bgpsecResult, '\0');
    }
    else
    {
      addToHistory("verify %u %u %u %s %u %u %s" ,
                   receipt, method, as32, ipStringPtr,
                   defResult.result.roaResult,
                   defResult.result.bgpsecResult, bgpsecInput);
    }
  }

  //////////////////////////////////////////////////////////
  //  Now send the message
  //////////////////////////////////////////////////////////

  if (receipt == 1)
  {
    method = method | SRX_FLAG_REQUEST_RECEIPT;
  }

  valResult = malloc(sizeof(SRxValidationResult));

  // The method verifyUpdate will go into wait mode if receipt is requested.
  if (verifyUpdate(proxy, valResult, method, &defResult, &prefix, as32, &bgpsec)
      == NULL)
  {
    RAISE_ERROR("(Verify) Validation encountered an error.\n");

  }
  else
  {
    LOG(LEVEL_DEBUG, "(Verify) Transmission successful!\n");
    if (receipt && (proxy->resCallback != NULL))
    {
      proxy->resCallback(valResult, proxy);
    }
  }
  free(valResult);
}

static void doSign(bool log, char** argPtr)
{
  uint32_t    updateID       = 0;
  uint32_t    prependCounter = 0;
  uint32_t    peerAS         = DEFAULT_PEERAS;

  if (!isConnected(proxy))
  {
    LOG(LEVEL_INFO, "Proxy is not connected, can not request signatures!\n");
    return;
  }

  while (updateID == 0)
  {
    updateID = promptU32(argPtr, "Update ID 1..n (not 0) ? ");
  }

  while (prependCounter == 0)
  {
    prependCounter = promptU32(argPtr, "Prepend Counter 1..m (not 0) ? ");
  }

  peerAS = promptU32(argPtr, "Peer AS ? ");

  if (!signUpdate(proxy, &updateID, false, 0, prependCounter, peerAS))
  {
    LOG(LEVEL_INFO, "The signature request could not be send!");
  }
  else
  {
    LOG(LEVEL_INFO, "The signature request successful send!");
  }
}

/**
 * This callback function allows the SRx to send a synchronization request to
 * the proxy.
 */
static void handleSyncRequest()
{
  char* buff = malloc(46);
  memset(buff, '\0', 46);
  sprintf(buff, "verify 2 3 65500 10.10.0.0/16 0 0 MyUpdateNo1");
  processLine(false, buff);
  sprintf(buff, "verify 2 3 65501 10.20.0.0/17 0 2 MyUpdateNo2");
  processLine(false, buff);
  sprintf(buff, "verify 2 3 65502 10.30.0.0/18 1 2 MyUpdateNo3");
  processLine(false, buff);
  sprintf(buff, "verify 2 1 65503 10.40.0.0/19 2 3 MyUpdateNo4");
  processLine(false, buff);
  sprintf(buff, "verify 2 1 65504 10.50.0.0/20 3 3 MyUpdateNo5");
  processLine(false, buff);
  LOG(LEVEL_DEBUG, "This update expects back a notification because SRx assumes"
                  " that the update might be stored using the new default"
                  " value. A notification is necessary to guarantee consistency"
                  " between SRx and ALL proxies!");
  sprintf(buff, "verify 2 1 65504 10.50.0.0/20 0 0 MyUpdateNo5");
  processLine(false, buff);
  sprintf(buff, "verify 2 1 65504 10.60.0.0/21 3 3 MyUpdateNo6");
  processLine(false, buff);
  LOG(LEVEL_DEBUG, "This update expects back a notification because proxy "
                  " requested receipt!");
  sprintf(buff, "verify 2 1 65504 10.60.0.0/21 0 0 MyUpdateNo6");
  processLine(false, buff);
  free(buff);
}

static void doDelete(bool log, char** argPtr)
{
  uint32_t    updateID    = 0;
  uint16_t    keep_window = 0;

  if (!isConnected(proxy))
  {
    LOG(LEVEL_INFO, "Proxy is not connected, can not request update deletion!\n");
    return;
  }

  keep_window = promptU32(argPtr, "Keep Window in seconds ? ");

  while (updateID == 0)
  {
    updateID = promptU32(argPtr, "Update ID 1..n (decimal not 0) ? ");
  }

  deleteUpdate(proxy, keep_window, &updateID);
}

/*-----
 * Main
 */

/**
 * Display version information.
 */
static void showVersion()
{
  printf("SRX - Proxy for SRx Server Version " SRX_SERVER_VERSION "\n");
}

/**
 * Display credits information.
 */
static void showCredits()
{
  showVersion();  
  printf(SRX_CREDITS);
}

/**
 * Display the help
 */
static void showHelp()
{
  printf("Note that all arguments are optional\n"
         "------------------------------------\n"
         "help  This screen!\n"
         "credits\n"
         "      Display the program credits!\n"
         "quit, \\q, exit\n"
         "      Quit this program.\n"
         "connect <host> <port> <peerAS> [<peerAS>*]\n"
         "      Connect to this SRx server.\n"
         "disconnect\n"
         "      Disconnect from SRx.\n"
         "addPeer <peerAS> [<peerAS>*]\n"
         "      Adds a peer to the configuration.\n"
         "delPeer <peerAS> [<peerAS>*]\n"
         "      Removed a peer from the configuration.\n"
        "verify <receipt(1=true, 2=false)> <method(1=RPKI, 2=BGPSec, 3=Both)>\n"
         "      <as> <prefix> "
               "<defaultROAVal(0-Valid, 1=Unknown, 2=Invalid, 3=Not Defined)>\n"
      "      <defaultBGPSECVal(0-Valid, 1=Unknown, 2=Invalid, 3=Not Defined)>\n"
         "      [bgpsec hex]\n"
         "      Initiates a verification request. This method waits \n"
         "      for SRx to return a receipt if receipt is requested!\n"
         "sign <updateID hex>\n"
         "      Initiate a signing request!\n"
         "run <filename>\n"
         "      Execute a script with commands in it.\n\n"
         "Extended Commands:\n"
         "------------------\n"
         "LOG_LEVEL <level(0=CANCEL, 3=ERROR, 5=NOTICE, 6=INFO, 7=DEBUG)>\n"
         "      Change the log level of the application.\n"
         "RESET_PROXY <proxyID(0=cancel)>\n"
         "      Release the current proxy and instantiate a new one.\n"
         "\n\n");
}

/**
 * Change the log level of the application
 *
 * @param argPtr the possible argument.
 */
static void processLogLevel(char** argPtr)
{
  uint32_t newLevel = 0;
  LogLevel ll;
  bool levelOK = false;

  while (!levelOK)
  {
    newLevel = promptU32(argPtr, "LOG_LEVEL (0=CANCEL, 3=ERROR, 5=NOTICE, "
                                 "6=INFO, 7=DEBUG) : ");
    switch(newLevel)
    {
      case LEVEL_ERROR  :
      case LEVEL_NOTICE :
      case LEVEL_INFO   :
      case LEVEL_DEBUG  :
        setLogLevel(newLevel);
      case 0 : // cancel
        levelOK = true;
        break;
      default:
        printf("The provided level [%u] is invalid!\n", newLevel);
    }
    if (newLevel == 0)
    {
      printf("Level not changed!\n");
    }
    else
    {
      printf("Changed log level to %u!\n", newLevel);      
    }
  }
}

/**
 * Reset the proxy.
 *
 * @param argPtr Contains the proxy ID
 */
static void resetProxy(char** argPtr)
{
  uint32_t proxyID = promptU32(argPtr, "ProxyID: ");
  if (proxyID == 0)
  {
    printf ("Reset canceled!\n");
  }
  else
  {
    if (proxy != NULL)
    {
      printf("Release current proxy [ID:%u]....\n", proxy->proxyID);
      releaseSRxProxy(proxy); // Also disconnects if necessary
      printf("Create new proxy [ID:%u]...\n", proxyID);
      proxy = createSRxProxy(handleValidationResult, handleSignatures,
                             handleSyncRequest,
                             proxyID,  // ProxyID
                             50, // ProxyAS
                             NULL);
      if (proxy == NULL)
      {
        RAISE_SYS_ERROR("Proxy Could not be created!");
      }
      else
      {
        printf("Proxy [ID:%u] generated!\n", proxy->proxyID);
      }
    }
  }

}


/**
 * Load a script and executes line by line.
 * 
 * @param log it true the commands will be added to the history.
 *
 * @param argPtr contains the script filename
 */
static void runScript(bool log, char** argPtr)
{
  #define SCRIPT_BUF_SIZE 1024

  char* fname;
  FILE* fh;
  char  buf[SCRIPT_BUF_SIZE];
  int   pos;

  fname = prompt(argPtr, "Script filename ? ");
  fh = fopen(fname, "rt");
  if (fh == NULL)
  {
    printf("Error: Failed to open '%s'\n", fname);
  }

  if (log)
  {
    addToHistory("run %s", fname);
  }

  while (fgets(buf, SCRIPT_BUF_SIZE, fh))
  {
    // Strip the line ending and space
    chomp(buf);

    // An empty line or comment
    if ((pos < 0) || (*buf == '#'))
    {
      continue;
    }

    (void)processLine(false, buf);
  }

  fclose(fh);
}

/**
 * Process and execute the given line.
 *
 * @param log it ture the line will be added to the history
 * @param line The line to be processed
 *
 * @return returns if the program has to be ended.
 */
static bool processLine(bool log, char* line)
{
  char* cmd, *arg;


  if ((line == NULL) || (*line == '\0'))
  {
    return false;
    
  }
    
  // Split into command and (optional) arguments
  arg = line;
  cmd = strsep(&arg, " \t");

  // Execute the command
  #define IF_EQ_DO(LONG, CMD) \
    if (!strcmp(cmd, LONG))   \
    {                         \
      CMD;                    \
    }

  IF_EQ_DO("quit", return true)
  else IF_EQ_DO("\\q", return true)
  else IF_EQ_DO("exit", return true)
  else IF_EQ_DO("help", showHelp())
  else IF_EQ_DO("credits", showCredits())
  else IF_EQ_DO("connect", doConnect(log, &arg))
  else IF_EQ_DO("addPeer", doMaintainPeer(log, &arg, true))
  else IF_EQ_DO("delPeer", doMaintainPeer(log, &arg, false))
  else IF_EQ_DO("disconnect", doDisconnect(log))
  else IF_EQ_DO("verify", doVerify(log, &arg))
  else IF_EQ_DO("sign", doSign(log, &arg))
  else IF_EQ_DO("delete", doDelete(log, &arg))
  else IF_EQ_DO("run", runScript(log, &arg))
  else IF_EQ_DO("LOG_LEVEL", processLogLevel(&arg))
  else IF_EQ_DO("RESET_PROXY", resetProxy(&arg))
  else
  {
    printf("Unknown command!\n");
  }
  
  bool stop = proxy == NULL;

  return stop;
}

/** The main program.*/
int main(int argc, char* argv[])
{
  char* line;
  bool  stop;

  // Set the target for the logging.
  setLogMethodToFile(stderr);

  // Create instance
  proxy = createSRxProxy(handleValidationResult, handleSignatures,
                         handleSyncRequest, 
                         0,  // ProxyID
                         50, // ProxyAS
                         NULL);

  if (proxy == NULL)
  {
    RAISE_ERROR("Proxy could not be created. Abort program!");
    return -1;
  }

  printf ("SRX API Test Harness Version 0.2\n");

  // Main user-input loop
  using_history();
  read_history(HISTORY_FILENAME);
  do
  {
    line = readline(PROMPT);
    stop = processLine(true, line);
    free(line);
  } while (!stop);

  write_history(HISTORY_FILENAME);

  // Destroy the instance
  if (proxy != NULL)
  {
    printf("Release proxy [ID:%u]\n", proxy->proxyID);
    releaseSRxProxy(proxy);
  }

  printf ("Goodbye!\n\n");

  return 0;
}