/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 * @file client_socket.h
 * @date Created: 12/28/2009
 *
 * Functions to create a client-socket, to receive from and to send data
 * to a server.
 *
 * log.h is used for the handling of error messages.
 */

#ifndef __CLIENT_SOCKET_H__
#define __CLIENT_SOCKET_H__

#include <netinet/in.h>
#include "util/packet.h"
#include "util/types.h"

/**
 * A Client Socket.
 */
typedef struct {
  /** The socket Address */
  struct sockaddr_in  svrAddr;
  /** The file descriptor of the client socket. */
  int                 clientFD;
  /** indicates if a reconnect is requested in case the  connection is lost. */  
  bool                reconnect; // Might be deleted, don't like it!
} ClientSocket;

/**
 * Creates a client-socket and tries to connect to a server.
 *
 * @param self Variable that should be initialized
 * @param host Server hostname (not IP-address)
 * @param port Server port
 * @param failNoServer Fail if a connection to the server could not be 
 *      established
 * @return \c true = socket created, or \c false in case of an error
 * @see createClientSocket
 */
extern bool createClientSocket(ClientSocket* self, const char* host, int port,
                               bool failNoServer);

/**
 * Closes a client-socket.
 *
 * @param self Client-socket instance
 * 
 * @see closeClientSocket
 */
extern void closeClientSocket(ClientSocket* self);

/**
 * Tries to establish a connection to the server again.
 *
 * @note A \c delay of > 0 makes this a blocking call
 *
 * @param self Client-socket instance
 * @param delay Wait \c delay seconds before trying again
 * @param max_attempts The maximum attempts this function performs to establish
 *              a connection before it returns false
 * 
 * @return \c true = connection established, \c failed to connect
 */
extern bool reconnectToServer(ClientSocket* self, int delay, int max_attempts);

/**
 * Stops any further attempts to connect to the server.
 *
 * @param self Client-socket instance
 */
extern void stopReconnectingToServer(ClientSocket* self);

/**
 * Returns a pointer to the file-descriptor of a client-socket.
 *
 * @param self Client-socket instance
 *
 * @return int* - file descriptor
 */
extern int* getClientFDPtr(ClientSocket* self);

/**
 * Returns \c true if the client is connected to the server.
 *
 * @param self Client-socket instance
 * 
 * @return \c true = connected, \c false = not connected
 */
extern bool isConnectedToServer(ClientSocket* self);

/**
 * Sends data to the server.
 *
 * @note An integer is put in front of the data to indicate the data length
 * 
 * @param self Client-socket instance
 * @param data Data to send
 * @param length Size (in Bytes) of the data
 * @return \c true = successfully sent, \c false = an error occurred
 * @see receiveData
 */
extern bool sendData(ClientSocket* self, void* data, PacketLength length);

/**
 * Waits and receives data from the server.
 *
 * @note The first 32-bit integer is expected to be the data length
 * @note Blocking-call
 *
 * @param self Client-socket instance
 * @param buffer Buffer where to store the received data
 * @param length Length of the buffer
 * @return Number of received Bytes, or \c -1 in case of an error
 * @see sendData
 * @see sendDataIndirect
 * @see skipBytes
 */
extern size_t receiveData(ClientSocket* self, void* buffer, size_t length);

/**
 * Reads and discards a certain number of Bytes.
 *
 * @note Blocking-call
 *
 * @param self Client-socket instance
 * @param num Number of Bytes to skip
 * @return \c true = successfully skipped the Bytes, \c = failed to skip
 * @see receiveData
 */
extern bool skipBytes(ClientSocket* self, size_t num);

#endif // !__CLIENT_SOCKET_H__

