/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 *
 * Oliver Borchert
 *
 *
 *
 *
 * @file packet.c
 * @date Created: 04/09/2010 by Patrick Gleichmann
 */

#include <stdlib.h>
#include <pthread.h>
#include <semaphore.h>
#include "client/client_connection_handler.h"
#include "packet.h"
#include "util/log.h"
#include "util/socket.h"
#include "client/srx_api.h"

#include "util/mutex.h"
#include "server_socket.h"
#include "server/server_connection_handler.h"
//#include "../server/server_connection_handler.h"

#define HDR "([0x%08X] Packet): "

/**
 * This function runs in a loop to receive packets. This function is used as 
 * receiver loop on both sides, SRx server as well as SRx client. IN case the 
 * SRx server uses. this method the parameter connHandler MUST be NULL.
 *
 * @note Blocking call!
 *
 * @param fdPtr        The file descriptor of the socket
 * @param dispatcher   The dispatcher method that receives all packets and
 *                     distributes them.
 * @param pHandler     Instance of the packet handler. On SRx server side this 
 *                     will be ClientThread, on the proxy side this will be 
 *                     SRxProxy.
 * @param pHandlerType The type of handler, srx-proxy or srx-server.
 *
 * @return true if the method ended clean.
 */
bool receivePackets(int* fdPtr, SRxPacketHandler dispatcher, void* pHandler, 
                    PacketHandlerType pHandlerType)
{
  bool retVal = true;
  // 8 bytes that contain the type and length
  uint32_t              basicLength = sizeof(SRXPROXY_BasicHeader);
  // Buffer that contains the bytes received.
  uint8_t*              buffer = malloc(basicLength);
  // A helper that points into the buffer at byte 9
  uint8_t*              buffHelper = buffer;//+basicLength;
  // The size of the current buffer. Can be larger than the message itself
  uint32_t              buffSize   = basicLength;
  // Bytes left after the basic length (of 8 bytes)
  uint32_t              remainder  = buffSize - basicLength;
  // The header mask layer put on top of the buffer for easier access
  SRXPROXY_BasicHeader* hdr        = (SRXPROXY_BasicHeader*)buffer;
  // The length of the PDU within the buffer
  uint32_t              pduLength = 0;

  int i_semval;

  ClientThread* cthread ;
  ServerConnectionHandler *hSvrConnection;
  CommandQueue* cmdQueue;

  if (pHandlerType == PHT_SERVER){
      cthread = (ClientThread*)pHandler;
      hSvrConnection = (ServerConnectionHandler *)cthread->svrSock->user;
      cmdQueue = (CommandQueue *)hSvrConnection->cmdQueue;
  }

  memset(buffer, 0, buffSize);

  // used to keep the receiver going.
  bool keepGoing = true;

  // Keeps the thread rolling - Process all packets
  while (keepGoing)
  {
    hdr = (SRXPROXY_BasicHeader*)buffer;

    // Get the packet length
    memset(hdr, 0, sizeof(SRXPROXY_BasicHeader));

    if (!recvNum(fdPtr, buffer, basicLength))
    {
      if (pHandlerType == PHT_PROXY)
      {
        // Within SRx Proxy
        ClientConnectionHandler* cHandler = 
                   (ClientConnectionHandler*)((SRxProxy*)pHandler)->connHandler;
        if (!cHandler->stop)
        {
          RAISE_ERROR("Could not receive the basic %u byte header!",
                      basicLength);
        }
        else
        {
          LOG(LEVEL_DEBUG, HDR "Data delivery interrupted - End server loop!", 
		                       pthread_self());
	      }
#if 0
	do {
	    usleep(2000);
	    sem_getvalue(&cHandler->sem_transx, &i_semval);
	} while( i_semval != 0);
#endif
	//sem_wait(&cHandler->sem_transx);  // P operation : atomically decrease

      }
      else if (pHandlerType == PHT_SERVER)
      {
        // Within SRx Server (or rpki_rtr test harness
        LOG(LEVEL_DEBUG, HDR "Connection to client closed.", pthread_self());
      }
      else
      {
        RAISE_SYS_ERROR("Invalid pHandler type!!!", pHandlerType);
        retVal = false;
      }
      break;
    }

    
    pduLength = ntohl(hdr->length);
    if (buffSize < pduLength)
    {
      // The current packet is Larger than the current buffer size
      // - need to resize the buffer
      buffSize = (size_t)pduLength;
      buffer  = realloc(buffer, buffSize);

      if (buffer == NULL)
      {
        RAISE_ERROR("Not enough memory for receiving packets");
        retVal = false;
        break;
      }
      // might be same address but just in case.
      buffHelper = buffer + basicLength;
    }
    remainder = pduLength - basicLength;
    if (remainder > pduLength)
    {
      RAISE_ERROR( HDR "Received PDU is invalid!", pthread_self());
      remainder = 0;
    }
    memset(buffHelper, 0, buffSize - basicLength);

    // Receive the packet
    if (!recvNum(fdPtr, buffHelper, (size_t)remainder))
    {
      RAISE_ERROR("Could not receive the basic remaining %u bytes!",
                  remainder);
      break;
    }

    // the first byte of the buffer contains the pdu type
    LOG(LEVEL_DEBUG, HDR "Received data and call dispatcher.", pthread_self());
    dispatcher((SRXPROXY_BasicHeader*)buffer, pHandler);
  }

  // Release the buffer and parameter structure
  if (buffer != NULL)
  {
    buffHelper = NULL;
    free(buffer);
  }

  return retVal;
}
