/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 * @file server_socket.h
 * @date Created: 12/23/2009
 *
 * Function to create a server-socket and to start/stop a server runloop.
 *
 * The runloop acts depending on the ClientMode:
 *
 * <table>
 * <tr>
 *   <th>ClientMode</th>
 *   <th>Description</th>
 *   <th>runServerLoop \c callback type</th>
 *   <th>Handles packets</th>
 * </tr>
 * <tr>
 *   <td>MODE_SINGLE_CLIENT</td>
 *   <td>1 client, 1 connection</td>
 *   <td>ServerPacketReceived</td>
 *   <td>yes</td>
 * </tr>
 * <tr>
 *   <td>MODE_MULTIPLE_CLIENTS</td>
 *   <td>N clients, 1 connection</td>
 *   <td>ServerPacketReceived</td>
 *   <td>yes</td>
 * </tr>
 * <tr>
 *   <td>MODE_CUSTOM_CALLBACK</td>
 *   <td>Handling is up to user's callback</td>
 *   <td>ClientConnectionAccepted</td>
 *   <td>no</td>
 * </tr>
 * </table>
 *
 * log.h is used for the handling of error messages.
 */

#ifndef __SERVER_SOCKET_H__
#define __SERVER_SOCKET_H__

#include "util/packet.h"
#include "util/slist.h"

/** Maximum number of clients waiting to be accepted for connection. */
#define MAX_PENDING_CONNECTIONS 5

/**
 * The Client-modes.
 * For more details see the detailed description.
 */
typedef enum
{
  MODE_SINGLE_CLIENT = 0, // 1 client  : 1 connection, ServerPacketReceived
  MODE_MULTIPLE_CLIENTS, // N clients : 1 connection, ServerPacketReceived
  MODE_CUSTOM_CALLBACK, // Custom, ClientConnectionAccepted

  NUM_CLIENT_MODES ///< Number of different modes (needs to be the last item)
} ClientMode;

/**
 * A client.
 *
 * @note Currently everything is hidden from the user.
 */
typedef void ServerClient;

/* Forward declaration */
struct _ServerSocket;

/**
 * A server-socket.
 *
 * @note Do not access any of the member variables.
 *
 * @note The typedef'd version is called ServerSocket. A direct definition is 
 *       not possible because a forward declaration is necessary.
 */
typedef struct _ServerSocket ServerSocket;

/**
 * Once a complete packet has been received by the server, this user-defined 
 * function is called.
 * To send the result, use \c sendPacketToClient - for example:
 *
 * @code
 * static my_receiptCallback(...) {
 *   uint32_t result;
 *   :
 *   sendPacketToClient(svrSock, client, &result, sizeof(uint32_t));
 * }
 * @endcode
 *
 * ClientMode: MODE_SINGLE_CLIENT, MODE_MULTIPLE_CLIENTS
 *
 * @param svrSock Server-socket instance
 * @param client Client that sent the packet
 * @param packet Data
 * @param length Number of Bytes of \c packet
 * @param user User-defined data
 */
typedef void (*ServerPacketReceived)(ServerSocket* svrSock,
                                     ServerClient* client,
                                     void* packet, PacketLength length,
                                     void* user);

/**
 * Each time a client connects to the server, this function is called. 
 * Note that the runloop takes care of creating the necessary threads.
 *
 * ClientMode: MODE_CUSTOM_CALLBACK
 *
 * @param svrSock Server-socket instance
 * @param sock Socket of the new client
 * @param user User-defined data
 */
typedef void (*ClientConnectionAccepted)(ServerSocket* svrSock, int sock,
                                         void* user);

/** 
 * Called upon the establishment or loss of a connection to a client. 
 * false as return-value denies the client, true accepts the client.
 *
 * @param svrSock Server-socket instance
 * @param client Client in MODE_SINGLE_CLIENT, otherwise \c NULL
 * @param fd New/lost file descriptor (= socket)
 * @param connected true = new client, false = client lost
 * @param user User-defined data
 * 
 * @return returns true if the client is accepted, otherwise false.
 */
typedef bool (*ClientStatusChanged)(ServerSocket* svrSock, ServerClient* client,
                                    int fd, bool connected, void* user);

/* The actual ServerSocket definition */
struct _ServerSocket
{
  // Arguments
  ClientMode mode;
  void (*modeCallback)();
  ClientStatusChanged statusCallback;
  void* user;

  // Internal variables
  int serverFD;
  int stopping;
  SList cthreads;
  bool verbose;
} ;

#if 1
/**
 * A single client thread.
 */
typedef struct
{
  pthread_t thread;

  // Arguments
  bool active;
  /** The file descriptor of the client socket. */
  int clientFD;
  /** Used as proxyID for SRx-Proxy connections to allow assigning updates to 
   *  the client. */
  uint32_t clientID;
  Mutex writeMutex;
  ServerSocket* svrSock;
} ClientThread;

#endif



/**
 * Creates a server-socket and binds it to a specific port.
 *
 * @param self The resulting server-socket
 * @param port Port that the server should listen on
 * @param verbose \c true = enable verbose mode
 * @return \c true = successfully created, \c false = an error occured
 */
extern bool createServerSocket(ServerSocket* self, int port, bool verbose);

/**
 * Starts the runloop which processes all client connections, and depending 
 * on the mode even the receipt of the packets.
 *
 * @param self Existing server-socket instance
 * @param clMode The client-mode
 * @param modeCallback Callback depending on \c clMode
 * @param statusCallback Function to invoke in case a client dis/connects. 
 *      May be \c NULL.
 * @param user User-defined data - will be passed to \c modeCallback and 
 *      \c statusCallback
 *
 * @see stopServerLoop
 */
extern void runServerLoop(ServerSocket* self, ClientMode clMode,
                          void (*modeCallback)(),
                          ClientStatusChanged statusCallback,
                          void* user);

/**
 * Stops the runloop, and closes all client connections.
 *
 * @param self Server-socket instance
 * @see runServerLoop
 */
extern void stopServerLoop(ServerSocket* self);


/**
 * Sends a packet to a clients.
 *
 * @note For MODE_MULTIPLE_CLIENTS this function must be called from
 *       within ServerPacketReceived.
 *
 * @param self Server-socket instance
 * @param client Client
 * @param data Data (w/o length) that should be send
 * @param size Size in Bytes of \c data
 * @return \c true = sent, \c false = an error occurred (e.g. inactive client)
 */
extern bool sendPacketToClient(ServerSocket* self, ServerClient* client,
                               void* data, size_t size);

/**
 * Closes the connection associated with the given client.
 * 
 * @param self The server socket whose client has to be handled,
 * @param client The client connection object to be closed.
 * 
 * @return true if the socket could be closed.
 */
extern int closeClientConnection(ServerSocket* self, ServerClient* client);

/**
 * Assign the given id as the clients ID. In case the ID is actively in use the
 * return value is 0 and no ID is assigned. In case the provided ID is zero the
 * next available ID will be used.
 * 
 * @param self The Server socket whose client has to be ID'ed
 * @param client The client to be ID'ed
 * @param clientID The ID to use for this client.
 * 
 * @return the id used for the client or 0 for an error
 */
int setServerClientID(ServerSocket* self, ServerClient* client, 
                      uint32_t proposeID);

/**
 * Return the ID of the client for this server.
 * 
 * @param client The client whose ID has to be determined.
 * @return The ID or 0 if none is assigned yet.
 */
extern uint32_t getServerClientID(ServerClient* client);

int g_single_thread_client_fd;

#endif // !__SERVER_SOCKET_H__

