/**
 * This software was developed at the National Institute of Standards and
 * Technology by employees of the Federal Government in the course of
 * their official duties. Pursuant to title 17 Section 105 of the United
 * States Code this software is not subject to copyright protection and
 * is in the public domain.
 * 
 * NIST assumes no responsibility whatsoever for its use by other parties,
 * and makes no guarantees, expressed or implied, about its quality,
 * reliability, or any other characteristic.
 * 
 * We would appreciate acknowledgement if the software is used.
 * 
 * NIST ALLOWS FREE USE OF THIS SOFTWARE IN ITS "AS IS" CONDITION AND
 * DISCLAIM ANY LIABILITY OF ANY KIND FOR ANY DAMAGES WHATSOEVER RESULTING
 * FROM THE USE OF THIS SOFTWARE.
 * 
 * 
 * This software might use libraries that are under GNU public license or
 * other licenses. Please refer to the licenses of all libraries required 
 * by thsi software.
 *
 */
/**
 * \file socket.c
 * \date Created: 12/29/2009 by Paterick Gleichmann
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <arpa/inet.h>
#include "socket.h"
#include "util/log.h"

#include <fcntl.h>
#include <pthread.h>

#define HDR "([0x%08X] Socket {%u}): "

/**
 * This method receives bytes and writes them into the given buffer.
 *
 * @param fd The file descriptor of the socket.
 * @param buffer The buffer to be send.
 * @param num The number of bytes to be send.
 *
 * @return true if data could be received and written in the buffer,
 * otherwise false.
 */
bool recvNum(int* fd, void* buffer, size_t num)
{
  ssize_t     rbytes;
    
  if (*fd == -1)
  {
    return false;
  }

  // Loop until all data is received.
  while (num > 0)
  {
    LOG(LEVEL_DEBUG, HDR "Wait to read (%u) bytes from Socket (status:%u).", 
                     pthread_self(), *fd, num, errno);
    rbytes = recv(*fd, buffer, num, MSG_WAITALL);
    LOG(LEVEL_DEBUG, HDR "Read %u of %u bytes from Socket (status: %u).", 
                     pthread_self(), *fd, rbytes, num, errno);
    // An error occurred
    if (rbytes == -1)
    {
      if (errno != EAGAIN)
      {
        // Print an error message only if not intentional
        if ((errno != EBADF) && (errno != ECONNRESET))
        {
          RAISE_SYS_ERROR("Socket error 0x%X (%u) while receiving data!", 
                          errno, errno);
        }
        else
        {
          LOG(LEVEL_DEBUG, HDR "Socket error 0x%X (%u) while receiving data - "
                           "Close socket!", pthread_self(), errno, errno);
          close(*fd);
        }
        *fd = -1;
        return false;
      }
    }

    // Connection lost
    if (rbytes == 0)
    {
      LOG(LEVEL_INFO, "Connection reset by peer.");
      close(*fd); // No CLOSE_WAIT
      *fd = -1;
      return false;
    }


    buffer += rbytes;
    num -= rbytes;
  };
  return true;
}

/**
 * Send the data stored in the buffer. this methid closes the socket in case of
 * an error.
 *
 * @param fd the file descripter of the socket to write into.
 * @param buffer The buffer to be written.
 * @param num the number of bytes to be written.
 *
 * @return true if the data could be send, otherwise false.
 */
bool sendNum(int* fd, void* buffer, size_t num)
{
  ssize_t sbytes;

  if (*fd == -1)
  {
    LOG(LEVEL_DEBUG, FILE_LINE_INFO " File descriptor is invalid!");
    return false;
  }

  while (num > 0)
  {
    sbytes = send(*fd, buffer, num, 0);

    // Any Error occurred
    if (sbytes == -1)
    {
      RAISE_SYS_ERROR("Socket error 0x%X (%u) while sending data!", 
                       errno, errno);
      if ((errno != EBADF) && (errno != ECONNRESET))
      {        
        close(*fd);
      }
      *fd = -1;
      return false;
    }

    buffer += sbytes;
    num -= sbytes;
  }

  return true;
}

/**
 * Generate the address string of the socket.
 * 
 * @param sa The socket address plus port number
 * @param dest The destination string pointer
 * @param size the size of the destination.
 * 
 * @return pointer to the destination.
 */
const char* sockAddrToStr(const struct sockaddr* sa, char* dest, size_t size)
{
  const char* ret;
  in_port_t port;
  size_t len;

  if (sa == NULL)
  {
    return strncpy(dest, "(null)", size);
  }

  // IPv4 
  if (sa->sa_family == AF_INET)
  {
    ret = inet_ntop(AF_INET, &((struct sockaddr_in*)sa)->sin_addr,
                    dest, size);
    port = ((struct sockaddr_in*)sa)->sin_port;

    // IPv6
  }
  else if (sa->sa_family == AF_INET6)
  {
    ret = inet_ntop(AF_INET6, &((struct sockaddr_in6*)sa)->sin6_addr,
                    dest, size);
    port = ((struct sockaddr_in6*)sa)->sin6_port;

    // Unknown address family
  }
  else
  {
    return strncpy(dest, "(Unknown type)", size);
  }

  if (ret == NULL)
  {
    strncpy(dest, "(Invalid IP)", size);
  }

  // Append the port
  len = strlen(dest);
  snprintf(dest + len, size - len, ":%d", ntohs(port));
  return dest;
}

/**
 * Returns the address of the given socket as string.
 * 
 * @param fd The file descriptor.
 * @param isPeer
 * @param dest
 * @param size
 * 
 * @return the address of the given socket as string.
 */
const char* socketToStr(int fd, bool isPeer, char* dest, size_t size)
{
  struct sockaddr sa;
  socklen_t slen = sizeof (struct sockaddr);

  if (   (isPeer && getpeername(fd, &sa, &slen) == -1)
      || (!isPeer && getsockname(fd, &sa, &slen) == -1))
  {
    return strncpy(dest, "(Invalid socket)", size);
  }
  return sockAddrToStr(&sa, dest, size);
}
