/* vim: tabstop=2 expandtab
 *
 * packet-rpkirtr.c
 * Routines for RPKI/Router Protocol dissection
 * http://tools.ietf.org/html/draft-ymbk-rpki-rtr-protocol-04
 * By Patrick H. Gleichmann <pgleichm@nist.gov>
 * http://tools.ietf.org/html/draft-ietf-sidr-rpki-rtr-12
 * By Oliver Borchert <borchert@nist.gov>
 *
 * Current: draft-ietf-sidr-rpki-rtr-12
 *
 * $Id$
 *
 * Wireshark - Network traffic analyzer
 * By Gerald Combs <gerald@wireshark.org>
 * Copyright 1998 Gerald Combs
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include <gmodule.h>
#include <epan/packet.h>
#include <epan/prefs.h>

/* Define version if we are not building Wireshark statically */
#ifndef ENABLE_STATIC
// rpki-rtr-protocol draft version 12
G_MODULE_EXPORT const gchar version[] = "0.12.0";
#endif

/* Useful constans */
#define PDU_HEADER_SIZE 8
#define IP_V4_ADDR_LEN  4
#define IP_V6_ADDR_LEN  16

/* PDU types */
#define PDU_TYPE_SERIAL_NOTIFY  0
#define PDU_TYPE_SERIAL_QUERY   1
#define PDU_TYPE_RESET_QUERY    2
#define PDU_TYPE_CACHE_RESPONSE 3
#define PDU_TYPE_IP_V4_PREFIX   4
#define PDU_TYPE_IP_V6_PREFIX   6
#define PDU_TYPE_END_OF_DATA    7
#define PDU_TYPE_CACHE_RESET    8
#define PDU_TYPE_ERROR_REPORT   10
#define MAX_PDU_TYPE            10

static const value_string pdu_type_vals[] = {
  { PDU_TYPE_SERIAL_NOTIFY, "Serial Notify"},
  { PDU_TYPE_SERIAL_QUERY, "Serial Query"},
  { PDU_TYPE_RESET_QUERY, "Reset Query"},
  { PDU_TYPE_CACHE_RESPONSE, "Cache Response"},
  { PDU_TYPE_IP_V4_PREFIX, "IPv4 Prefix"},
  { PDU_TYPE_IP_V6_PREFIX, "IPv6 Prefix"},
  { PDU_TYPE_END_OF_DATA, "End of Data"},
  { PDU_TYPE_CACHE_RESET, "Cache Reset"},
  { PDU_TYPE_ERROR_REPORT, "Error Report"},
  { 0, NULL}
};

/* Fixed offsets (in Bytes) */
#define OFFSET_VERSION    0
#define OFFSET_PDU_TYPE   1
#define OFFSET_RESERVED   2
#define OFFSET_LENGTH     4
#define OFFSET_DATA       8

/* Structures */
struct rpkirtr_prefix_header_t
{
  guint8 flags;
  guint8 prefix_len;
  guint8 max_len;
  guint8 field_zero;
};

/* Flags */
#define PREFIX_FLAG_ANNOUNCEMENT  0x01

/* Forward declarations */
void proto_reg_handoff_rpkirtr(void);
void proto_register_rpkirtr();
static int dissect_rpkirtr(tvbuff_t*, packet_info*, proto_tree*);


/* Preference variables */
/* Disabled by default, because the draft does not specify a default port */
static guint pref_tcp_port = 0;
static gboolean pref_decode_pdu = FALSE;

static int proto_rpkirtr = -1;
static int hf_rpkirtr_version = -1;
static int hf_rpkirtr_pdu_type = -1;
static int hf_rpkirtr_reserved = -1;
static int hf_rpkirtr_cache_nonce = -1;
static int hf_rpkirtr_error_code = -1;
static int hf_rpkirtr_length = -1;
static int hf_rpkirtr_serial_no = -1;
static int hf_rpkirtr_prefix4 = -1;
static int hf_rpkirtr_prefix6 = -1;
static int hf_rpkirtr_error_msg = -1;

static gint ett_rpkirtr = -1;
static gint ett_rpkirtr_pdu = -1;

#ifndef ENABLE_STATIC
G_MODULE_EXPORT void plugin_register(void)
{
   /* register the new protocol, protocol fields, and subtrees */
   if (proto_rpkirtr == -1) 
   { /* execute protocol initialization only once */
      proto_register_rpkirtr();
   }
}

G_MODULE_EXPORT void plugin_reg_handoff(void)
{
   proto_reg_handoff_rpkirtr();
}
#endif


static void dissect_rpkirtr_prefix(tvbuff_t* tvb, proto_tree* tree,
                                   guint32 offset, gboolean ipv6)
{
  struct rpkirtr_prefix_header_t phdr;
  guint8 valid_len;
  emem_strbuf_t* flags_strbuf;
  int ofs;

#define PHDR_OFFSET(MEMBER) \
    (offset + ((size_t)(&((struct rpkirtr_prefix_header_t*)0)->MEMBER)))
//    (OFFSET_DATA + ((size_t)(&((struct rpkirtr_prefix_header_t*)0)->MEMBER)))

  /* Copy into a struct for more convenient access */
  tvb_memcpy(tvb, &phdr, offset, sizeof (struct rpkirtr_prefix_header_t));

  /* Flags */
  flags_strbuf = ep_strbuf_new_label("");
  ep_strbuf_append(flags_strbuf, (phdr.flags & PREFIX_FLAG_ANNOUNCEMENT) ?
                   "Announcement" : " Withdrawal");
  proto_tree_add_text(tree, tvb, offset, 1,
                      "Flags: 0x%x (%s)", phdr.flags, flags_strbuf->str);

  /* Validate both lengths and add them to the tree */
  valid_len = ipv6 ? 128 : 32;
  if (phdr.prefix_len > valid_len)
  {
    proto_tree_add_text(tree, tvb, PHDR_OFFSET(prefix_len), 1,
                        "Invalid Prefix Length: %u (0..%u)",
                        phdr.prefix_len, valid_len);
  }
  else
  {
    proto_tree_add_text(tree, tvb, PHDR_OFFSET(prefix_len), 1,
                        "Prefix Length: %u", phdr.prefix_len);
  }

  if (phdr.max_len > valid_len)
  {
    proto_tree_add_text(tree, tvb, PHDR_OFFSET(max_len), 1,
                        "Invalid Max Length: %u (0..%u)",
                        phdr.max_len, valid_len);
  }
  else
  {
    proto_tree_add_text(tree, tvb, PHDR_OFFSET(max_len), 1,
                        "Prefix Max: %u", phdr.max_len);
  }

  /* Zero field */
  proto_tree_add_text(tree, tvb, PHDR_OFFSET(field_zero), 1,
                      "zero: 0x%x", phdr.field_zero);

  /* Prefix (v4 or v6) */
  ofs = offset + sizeof (struct rpkirtr_prefix_header_t);
  if (ipv6)
  {
    proto_tree_add_ipv6(tree, hf_rpkirtr_prefix6, tvb, ofs, IP_V6_ADDR_LEN,
                        tvb_get_ptr(tvb, ofs, IP_V6_ADDR_LEN));
    ofs += IP_V6_ADDR_LEN;
  }
  else
  {
    proto_tree_add_ipv4(tree, hf_rpkirtr_prefix4, tvb, ofs, IP_V4_ADDR_LEN,
                        tvb_get_ipv4(tvb, ofs));
    ofs += IP_V4_ADDR_LEN;
  }

  /* AS */
  proto_tree_add_text(tree, tvb, ofs, 4, "AS Number: %u",
                      tvb_get_ntohl(tvb, ofs));
}

static void dissect_rpkirtr_error_report(tvbuff_t* tvb, proto_tree* tree,
                                         guint32 ofs)
{
  guint32 epdu_len;
  guint32 emsg_len;

  /* Encapsulated (= errorneous) PDU */
  epdu_len = tvb_get_ntohl(tvb, ofs);
  proto_tree_add_text(tree, tvb, ofs, 4, "pdu_length: %u", epdu_len);
  ofs += 4;
  if (epdu_len == 0)
  {
    proto_tree_add_text(tree, tvb, ofs, 0, "pdu: na");
  }
  else
  {
    /* Decode the PDU */
    if (pref_decode_pdu)
    {
      proto_item* ti;
      proto_tree* pdu_tree;
      tvbuff_t* epdu_buf;

      ti = proto_tree_add_text(tree, tvb, ofs, epdu_len,
                               "Erroneous PDU");
      pdu_tree = proto_item_add_subtree(ti, ett_rpkirtr_pdu);
      epdu_buf = tvb_new_subset(tvb, ofs, epdu_len, epdu_len);
      // DISECT ERRONEOUS PDU
      dissect_rpkirtr(epdu_buf, NULL, pdu_tree);
    }
    else
    {
      proto_tree_add_text(tree, tvb, ofs, epdu_len, "PDU: (see bytestream)");
    }
    ofs += epdu_len;
  }

  /* Error message (4 = field width) */
  emsg_len = tvb_get_ntohl(tvb, ofs);
  proto_tree_add_text(tree, tvb, ofs, 4, "length msg: %u", emsg_len);
  ofs += 4;
  if (emsg_len == 0)
  {
    proto_tree_add_text(tree, tvb, ofs, 0, "message: na");
  }
  else
  {
    proto_tree_add_text(tree, tvb, ofs, emsg_len, "message: (see bytestream)");
  }
}

static int dissect_rpkirtr(tvbuff_t *tvb, packet_info *pinfo, proto_tree *tree)
{
  guint8 version;
  guint8 type;
  // total length of disected data
  guint32 bytes_dissected;
  // Length of current PDU to dissect
  guint32 pdu_len;
  /* the length of the copmlete data */
  guint32 data_len;
  // How much data is still undissected
  guint32 remainder;

  proto_item* ti;
  proto_tree* rpkirtr_tree;
  emem_strbuf_t* pdu_strbuf;

  bytes_dissected = 0;
  pdu_len         = 0;
  data_len        = tvb_length(tvb);
  remainder       = data_len;
  pdu_strbuf      = ep_strbuf_new_label("");

  if (pinfo != NULL)
  {
    if (check_col(pinfo->cinfo, COL_PROTOCOL))
    {
      col_set_str(pinfo->cinfo, COL_PROTOCOL, "RPKI/RTR");
    }
  }

  // Repeat as long as data is available.
  while (remainder > 0)
  {
    /* Test if enough data is available */
    if (remainder < PDU_HEADER_SIZE)
    {
      ep_strbuf_append(pdu_strbuf, "Not enough data to read PDU "
                                       "length field!");
      break; // exit while loop
    }
    /* Get the PDU length */
    pdu_len = tvb_get_ntohl(tvb, OFFSET_LENGTH + bytes_dissected);

    // Check if enough data is available to dissect PDU
    if (remainder < pdu_len)
    {
      ep_strbuf_append(pdu_strbuf, "Not enough data to dissect PDU");
      break; // exit while loop
    }

    /* Test if Version is correct! */
    version = tvb_get_guint8(tvb, (bytes_dissected + OFFSET_VERSION));
    if (version != 0)
    {
      ep_strbuf_append(pdu_strbuf, "Incorrect Version");
      break;
      // exit while loop
    }

    /* Check the type */
    type = tvb_get_guint8(tvb, (bytes_dissected + OFFSET_PDU_TYPE));
    if (type == 5 || type == 9 || type > MAX_PDU_TYPE)
    {
      ep_strbuf_append(pdu_strbuf, val_to_str(type, pdu_type_vals,
                                                "Unknown PDU type %d"));
      break; // exit while loop
    }

    if (pdu_len > remainder)
    {
      ep_strbuf_append(pdu_strbuf, "Not enough Data to proceed!");
      break; // exit while loop
    }

    /* Add the protocol and info column to the summary display */
    ep_strbuf_append(pdu_strbuf,
                     val_to_str(type, pdu_type_vals, "Unknown PDU type %d"));
    if (pdu_len < remainder)
    {
      ep_strbuf_append(pdu_strbuf, ", ");
    }

    if (tree)
    {
      /* The main-node covers the whole PDU of this loop run  */
      ti = proto_tree_add_item(tree, proto_rpkirtr, tvb, bytes_dissected,
                               pdu_len, FALSE);
      rpkirtr_tree = proto_item_add_subtree(ti, ett_rpkirtr);

      /* Add the nodes for the header (FALSE = big-endian) */
      proto_tree_add_item(rpkirtr_tree, hf_rpkirtr_version, tvb,
                          bytes_dissected + OFFSET_VERSION, 1, FALSE);
      proto_tree_add_item(rpkirtr_tree, hf_rpkirtr_pdu_type, tvb,
                          bytes_dissected + OFFSET_PDU_TYPE, 1, FALSE);
      switch(type)
      {
        case PDU_TYPE_SERIAL_NOTIFY:
        case PDU_TYPE_SERIAL_QUERY:
        case PDU_TYPE_CACHE_RESPONSE:
        case PDU_TYPE_END_OF_DATA:
          proto_tree_add_item(rpkirtr_tree, hf_rpkirtr_cache_nonce, tvb,
                              bytes_dissected + OFFSET_RESERVED, 2, FALSE);
          break;
        case PDU_TYPE_RESET_QUERY:
        case PDU_TYPE_IP_V4_PREFIX:
        case PDU_TYPE_IP_V6_PREFIX:
        case PDU_TYPE_CACHE_RESET:
          proto_tree_add_item(rpkirtr_tree, hf_rpkirtr_reserved, tvb,
                              bytes_dissected + OFFSET_RESERVED, 2, FALSE);
          break;
        case PDU_TYPE_ERROR_REPORT:
          proto_tree_add_item(rpkirtr_tree, hf_rpkirtr_error_code, tvb,
                              bytes_dissected + OFFSET_RESERVED, 2, FALSE);
          break;
        default:
          break;
      }
      proto_tree_add_item(rpkirtr_tree, hf_rpkirtr_length, tvb,
                          bytes_dissected + OFFSET_LENGTH, 4, FALSE);

      /* Add nodes depending on the type */
      switch (type)
      {
          /* Serial */
        case PDU_TYPE_SERIAL_NOTIFY:
        case PDU_TYPE_SERIAL_QUERY:
        case PDU_TYPE_END_OF_DATA:
          proto_tree_add_item(rpkirtr_tree, hf_rpkirtr_serial_no, tvb,
                              bytes_dissected + OFFSET_DATA, 4, FALSE);
          break;
        case PDU_TYPE_IP_V4_PREFIX:
          dissect_rpkirtr_prefix(tvb, rpkirtr_tree, bytes_dissected+OFFSET_DATA,
                                 FALSE);
          break;
        case PDU_TYPE_IP_V6_PREFIX:
          dissect_rpkirtr_prefix(tvb, rpkirtr_tree, bytes_dissected+OFFSET_DATA,
                                 TRUE);
          break;
        case PDU_TYPE_ERROR_REPORT:
          dissect_rpkirtr_error_report(tvb, rpkirtr_tree,
                                       bytes_dissected+OFFSET_DATA);
          break;
        default:
          break;
      }
    }
    // Generate the data for the next loop run
    bytes_dissected += pdu_len;
    remainder = tvb_reported_length_remaining(tvb, bytes_dissected);
  }

  if (pinfo != NULL)
  {
    if (check_col(pinfo->cinfo, COL_INFO))
    {
      col_add_str(pinfo->cinfo, COL_INFO, pdu_strbuf->str);
    }
  }

  return bytes_dissected;
}

void proto_register_rpkirtr(void)
{
  /* Header fields */
  static hf_register_info hf[] = {
    {
      &hf_rpkirtr_version,
      {"Protocol Version", "rpkirtr.ver",
       FT_UINT8, BASE_DEC, NULL, 0x0, "", HFILL}
    },
    {
      &hf_rpkirtr_pdu_type,
      {"PDU Type", "rpkirtr.type",
       FT_UINT8, BASE_DEC, NULL, 0x0, "", HFILL}
    },
    {
      &hf_rpkirtr_reserved,
      {"Reserved", "rpkirtr.reserved",
       FT_UINT16, BASE_DEC_HEX, NULL, 0x0, "", HFILL}
    },
    {
      &hf_rpkirtr_cache_nonce,
      {"Cache Nonce", "rpkirtr.cnonce",
       FT_UINT16, BASE_DEC_HEX, NULL, 0x0, "", HFILL}
    },
    {
      &hf_rpkirtr_error_code,
      {"Error Code", "rpkirtr.errcode",
       FT_UINT16, BASE_DEC_HEX, NULL, 0x0, "", HFILL}
    },
    {
      &hf_rpkirtr_length,
      {"Length", "rpkirtr.len",
       FT_UINT32, BASE_DEC, NULL, 0x0, "", HFILL}
    },                                  /* For formatting */
    {
      &hf_rpkirtr_serial_no,
      {"Serial Number", "rpkirtr.sno",
       FT_UINT32, BASE_DEC, NULL, 0x0, "", HFILL}
    },
    {
      &hf_rpkirtr_prefix4,
      {"IPv4 Prefix", "rpkirtr.prefix4",
       FT_IPv4, BASE_NONE, NULL, 0x0, "", HFILL}
    },
    {
      &hf_rpkirtr_prefix6,
      { "IPv6 Prefix", "rpkirtr.prefix6",
       FT_IPv6, BASE_NONE, NULL, 0x0, "", HFILL}
    },
    {
      &hf_rpkirtr_error_msg,
      { "Error Message", "rpkirtr.errmsg",
       FT_UINT_STRING, BASE_NONE, NULL, 0x0, "", HFILL}
    }
  };

  static gint * ett[] = {
                         &ett_rpkirtr,
                         &ett_rpkirtr_pdu
  };

  module_t* rpkirtr_module;

  /* Register the protocol */
  proto_rpkirtr = proto_register_protocol(
                                          "RPKI/Router protocol",
                                          "RPKI/RTR",
                                          "rpkirtr"
                                          );

  /* Register the header fields and tree */
  proto_register_field_array(proto_rpkirtr, hf, array_length(hf));
  proto_register_subtree_array(ett, array_length(ett));

  rpkirtr_module = prefs_register_protocol(proto_rpkirtr,
                                           proto_reg_handoff_rpkirtr);

  /* Preferences */
  prefs_register_uint_preference(rpkirtr_module,
                                 "tcp.port",
                                 "TCP port for RPKI/Router Protocol",
                                 "Set the TCP port",
                                 10,
                                 &pref_tcp_port);
  prefs_register_bool_preference(rpkirtr_module,
                                 "decode_pdu",
                                 "Decode erroneous PDU",
                                 "Try to decode the erroneous PDU within an "
                                   "Error Report",
                                 &pref_decode_pdu);
}

void proto_reg_handoff_rpkirtr(void)
{
  static gboolean initialized = FALSE;
  static guint old_tcp_port = 0;
  static dissector_handle_t rpkirtr_handle;

  /* Initialize the dissector - unless this has been done before */
  if (!initialized)
  {
    rpkirtr_handle = new_create_dissector_handle(dissect_rpkirtr,
                                                 proto_rpkirtr);
    initialized = TRUE;

    /* Remove a previously registered dissector */
  }
  else if (old_tcp_port != 0)
  {
    dissector_delete("tcp.port", old_tcp_port, rpkirtr_handle);
  }

  /* Register the dissector */
  if (pref_tcp_port != 0)
  {
    dissector_add("tcp.port", pref_tcp_port, rpkirtr_handle);
    old_tcp_port = pref_tcp_port;
  }
}