/* vim: tabstop=2 expandtab
 *
 * packet-srxproxy.c
 * Routines for RouterSRx/Proxy Protocol dissection
 * draft-srx-proxy-protocol-08
 * By Oliver Borchert <borchert@nist.gov>
 *
 * Current: draft-srx-proxy-protocol-08
 *
 * $Id$
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <glib.h>
#include <gmodule.h>
#include <epan/packet.h>
#include <epan/prefs.h>


/* Define version if we are not building Wireshark statically */
#ifndef ENABLE_STATIC
// srx proxy protocol version 0.6
G_MODULE_EXPORT const gchar version[] = "0.8";
#endif

/* Useful constans */
#define SRX_MINPDU_HEADER_SIZE  8
#define IP_V4_ADDR_LEN          4
#define IP_V6_ADDR_LEN         16

/* PDU types */
#define PDU_PROXSRX_HELLO              0
#define PDU_PROXSRX_HELLO_RESPONSE     1
#define PDU_PROXSRX_GOODBYE            2
#define PDU_PROXSRX_VERIFY_V4_REQUEST  3
#define PDU_PROXSRX_VERIFY_V6_REQUEST  4
#define PDU_PROXSRX_SIGN_REQUEST       5
#define PDU_PROXSRX_VERI_NOTIFICATION  6
#define PDU_PROXSRX_SIGN_NOTIFICATION  7
#define PDU_PROXSRX_DELTE_UPDATE       8
#define PDU_PROXSRX_PEER_CHANGE        9
#define PDU_PROXSRX_SYNC_REQUEST      10
#define PDU_PROXSRX_ERROR             11
#define PDU_PROXSRX_UNKNOWN           12    

static const value_string pdu_type_vals[] = {
  { PDU_PROXSRX_HELLO,             "Hello"},
  { PDU_PROXSRX_HELLO_RESPONSE,    "HelloResponse"},
  { PDU_PROXSRX_GOODBYE,           "Goodbye"},
  { PDU_PROXSRX_VERIFY_V4_REQUEST, "VerifyIPv4"},
  { PDU_PROXSRX_VERIFY_V6_REQUEST, "VerifyIPv6"},
  { PDU_PROXSRX_SIGN_REQUEST,      "SignRequest"},
  { PDU_PROXSRX_VERI_NOTIFICATION, "VerifyNotification"},
  { PDU_PROXSRX_SIGN_NOTIFICATION, "SignNotification"},
  { PDU_PROXSRX_DELTE_UPDATE,      "DeleteUpdate"},
  { PDU_PROXSRX_PEER_CHANGE,       "PeerChange"},
  { PDU_PROXSRX_SYNC_REQUEST,      "SyncRequest"},
  { PDU_PROXSRX_ERROR,             "Error"},
  { PDU_PROXSRX_UNKNOWN,           "Unknown"}
};

/* Fixed offsets (in Bytes) */
#define SRX_OFFSET_PDU_TYPE   0
#define SRX_OFFSET_LENGTH     4

#define SRX_OFFSET_LINE_01  0
#define SRX_OFFSET_LINE_02  4
#define SRX_OFFSET_LINE_03  8
#define SRX_OFFSET_LINE_04 12
#define SRX_OFFSET_LINE_05 16
#define SRX_OFFSET_LINE_06 20
#define SRX_OFFSET_LINE_07 24
#define SRX_OFFSET_LINE_08 28
#define SRX_OFFSET_LINE_09 32
#define SRX_OFFSET_LINE_10 36

/* Flags */
#define SRX_FLAG_VERIFY_PREFIX_ORIGIN 0x01
#define SRX_FLAG_VERIFY_PATH          0x02

/** Peer Change Type */
#define SRX_PEER_CHANGE_DEL 0
#define SRX_PEER_CHANGE_ADD 1

static const value_string peer_change_vals[] = {
  { SRX_PEER_CHANGE_DEL, "DELETE"},
  { SRX_PEER_CHANGE_ADD, "ADD"}
};

/* ORIGIN / BGPSEC Result Source */
#define SRX_RESULT_SRC_SRX     0
#define SRX_RESULT_SRC_ROUTER  1
#define SRX_RESULT_SRC_IGP     2
#define SRX_RESULT_SRC_UNKNOWN 3

static const value_string result_src_vals[] = {
  { SRX_RESULT_SRC_SRX,     "SRX"},
  { SRX_RESULT_SRC_ROUTER,  "ROUTER"},
  { SRX_RESULT_SRC_IGP,     "IGP"},
  { SRX_RESULT_SRC_UNKNOWN, "UNKNOWN"}
};

/* Result Flags */
#define SRX_FLAG_RESULT_TYPE_ORIGIN    0x01
#define SRX_FLAG_RESULT_TYPE_BGPSEC 0x02
#define SRX_FLAG_RECEIPT_REQUEST    0x80

/** General ORIGIN and BGPSEC result types */
#define SRX_RESULT_VALID     0
#define SRX_RESULT_UNKNOWN   1
#define SRX_RESULT_INVALID   2
#define SRX_RESULT_UNDEFINED 3

static const value_string result_type_vals[] = {
  { SRX_RESULT_VALID,     "VALID"},
  { SRX_RESULT_UNKNOWN,   "UNKNOWN"},
  { SRX_RESULT_INVALID,   "INVALID"},
  { SRX_RESULT_UNDEFINED, "UNDEFINED"}
};

/** Block Type */
#define SRX_BLOCKTYPE_ALL_SIGNATURIES       0
#define SRX_BLOCKTYPE_LATEST_SIGNATURE_ONLY 1

static const value_string block_type_vals[] = {
  { SRX_BLOCKTYPE_ALL_SIGNATURIES,       "ALL SIGNATURES"},
  { SRX_BLOCKTYPE_LATEST_SIGNATURE_ONLY, "LATEST ONLY"}
};

/** Error Codes */
#define SRX_ERROR_WRONG_VERSION      0
#define SRX_ERROR_DUPLICATE_PROXY_ID 1
#define SRX_ERROR_INVALID_PACKET     2
#define SRX_ERROR_INTERNAL_ERROR     3
#define SRX_ERROR_ALGO_NOT_SUPPORTED 4
#define SRX_ERROR_UPDATE_NOT_FOUND   5

static const value_string error_code_vals[] = {
  { SRX_ERROR_WRONG_VERSION,      "Wrong Protocol Version (fatal)"},
  { SRX_ERROR_DUPLICATE_PROXY_ID, "Duplicate Proxy Identifier (fatal)"},
  { SRX_ERROR_INVALID_PACKET,     "Invalid Packet (fatal)"},
  { SRX_ERROR_INTERNAL_ERROR,     "Internal Error (fatal)"},
  { SRX_ERROR_ALGO_NOT_SUPPORTED, "Algorithm Not Supported"},
  { SRX_ERROR_UPDATE_NOT_FOUND,   "Update with the given ID not found"}
};

/* Forward declarations */
void proto_reg_handoff_srxproxy(void);
void proto_register_srxproxy(void);
static int dissect_srxproxy(tvbuff_t*, packet_info*, proto_tree*);

static int dissect_srxproxy_hello(tvbuff_t*, proto_tree*, guint32 ofs);
static int dissect_srxproxy_hello_response(tvbuff_t*, proto_tree*, guint32 ofs);
static int dissect_srxproxy_goodbye(tvbuff_t*, proto_tree*, guint32 ofs);
static int dissect_srxproxy_veri_request(tvbuff_t*, proto_tree*, gboolean, 
                                         guint32 ofs);
static int dissect_srxproxy_sign_request(tvbuff_t*, proto_tree*, guint32 ofs);
static int dissect_srxproxy_veri_notification(tvbuff_t*, proto_tree*, 
                                              guint32 ofs);
static int dissect_srxproxy_sign_notification(tvbuff_t*, proto_tree*, 
                                              guint32 offs);
static int dissect_srxproxy_delete_update(tvbuff_t*, proto_tree*, guint32 ofs);
static int dissect_srxproxy_peer_change(tvbuff_t*, proto_tree*, guint32 ofs);
static int dissect_srxproxy_sync_request(tvbuff_t*, proto_tree*, guint32 ofs);
static int dissect_srxproxy_error(tvbuff_t*, proto_tree*, guint32 ofs);

/* Preference variables */
/* Disabled by default, because the draft does not specify a default port */
static guint    pref_tcp_port   = 0;

static int proto_srxproxy = -1;
static int hf_srxproxy_pdu_type = -1;
static int hf_srxproxy_prefix4 = -1;
static int hf_srxproxy_prefix6 = -1;
static gint ett_srxproxy = -1;
static gint ett_srxproxy_pdu = -1;


#ifndef ENABLE_STATIC
G_MODULE_EXPORT void plugin_register(void)
{
   /* register the new protocol, protocol fields, and subtrees */
   if (proto_srxproxy == -1) 
   { /* execute protocol initialization only once */
      proto_register_srxproxy();
   }
}

G_MODULE_EXPORT void plugin_reg_handoff(void)
{
   proto_reg_handoff_srxproxy();
}
#endif

void proto_reg_handoff_srxproxy(void)
{
  static gboolean initialized = FALSE;
  static guint    old_tcp_port = 0;
  static dissector_handle_t srxproxy_handle;

  /* Initialize the dissector - unless this has been done before */
  if (!initialized)
  {
    srxproxy_handle = new_create_dissector_handle(dissect_srxproxy,
                                                  proto_srxproxy);
    initialized = TRUE;

    /* Remove a previously registered dissector */
  }
  else if (old_tcp_port != 0)
  {
    dissector_delete("tcp.port", old_tcp_port, srxproxy_handle);
  }

  /* Register the dissector */
  if (pref_tcp_port != 0)
  {
    dissector_add("tcp.port", pref_tcp_port, srxproxy_handle);
    old_tcp_port = pref_tcp_port;
  }
}

void proto_register_srxproxy(void)
{
  /* Header fields */
  static hf_register_info hf[] = {
    // Bytes 0-3
    {
      &hf_srxproxy_pdu_type,
      {"PDU Type", "srxproxy.type",
       FT_UINT8, BASE_DEC, NULL, 0x0, "", HFILL}
    },
    // Bytes 12-15
    {
      &hf_srxproxy_prefix4,
      {"IPv4 Prefix", "srxproxy.prefix4",
       FT_IPv4, BASE_NONE, NULL, 0x0, "", HFILL}
    },
    // Bytes 12-27
    {
      &hf_srxproxy_prefix6,
      {"IPv6 Prefix", "srxproxy.prefix6",
       FT_IPv6, BASE_NONE, NULL, 0x0, "", HFILL}
    }
  };

  static gint * ett[] = {
     &ett_srxproxy,
     &ett_srxproxy_pdu
  };

  module_t* srxproxy_module;

  /* Register the protocol */
  proto_srxproxy = proto_register_protocol(
                                          "SRx/Proxy protocol",
                                          "SRx/Proxy",
                                          "srxproxy"
                                          );

  /* Register the header fields and tree */
  proto_register_field_array(proto_srxproxy, hf, array_length(hf));
  proto_register_subtree_array(ett, array_length(ett));

  srxproxy_module = prefs_register_protocol(proto_srxproxy,
                                            proto_reg_handoff_srxproxy);

  /* Preferences */
  prefs_register_uint_preference(srxproxy_module,
                                 "tcp.port",
                                 "TCP port for SRx/Proxy Protocol",
                                 "Set the TCP port",
                                 10,
                                 &pref_tcp_port);
}

////////////////////////////////////////////////////////////////////////////////
//  DISECTORS
////////////////////////////////////////////////////////////////////////////////

/**
 * Main Dissector that delegates according to the PDU type
 */
static int dissect_srxproxy(tvbuff_t* tvb, packet_info* pinfo, proto_tree* tree)
{
  guint8  pdu_type;  
  // total length of dissected data
  guint32 bytes_dissected;
  // Length of current PDU to dissect
  guint32 pdu_len;
  /* the length of the complete data */
  guint32 data_len;
  // How much data is still un-dissected
  guint32 remainder;
    
  proto_item* ti;
  proto_tree* srxproxy_tree;
  emem_strbuf_t* pdu_strbuf;

  bytes_dissected = 0;
  pdu_len         = 0;
  data_len        = tvb_length(tvb);
  remainder       = data_len;
  pdu_strbuf      = ep_strbuf_new_label("");
  
  if (pinfo != NULL)
  {
    if (check_col(pinfo->cinfo, COL_PROTOCOL))
    {
      col_set_str(pinfo->cinfo, COL_PROTOCOL, "SRX/RPOXY");
    }
  }

  while (remainder > 0)
  {
    /* Test if enough data is available */
    if (remainder < SRX_MINPDU_HEADER_SIZE)
    {
      ep_strbuf_append(pdu_strbuf, "Not enough data to read PDU "
                                       "length field!");
      break; // exit while loop
    }
    /* Check the type */
    pdu_type = tvb_get_guint8(tvb, bytes_dissected + SRX_OFFSET_PDU_TYPE);
    if (pdu_type > PDU_PROXSRX_UNKNOWN)
    {
      if (check_col(pinfo->cinfo, COL_INFO))
      {
        col_set_str(pinfo->cinfo, COL_INFO,
                    val_to_str(pdu_type, pdu_type_vals, "Unknown PDU type %d"));
      }
      break;
    }    
    
    /* Get the PDU length */
    pdu_len = tvb_get_ntohl(tvb, bytes_dissected + SRX_OFFSET_LENGTH);

    // Check if enough data is available to dissect PDU
    if (remainder < pdu_len)
    {
      ep_strbuf_append_printf(pdu_strbuf, "Not enough data to dissect PDU[%u]", 
                              pdu_type);
      break; // exit while loop
    }

    if (pdu_len > remainder)
    {
      ep_strbuf_append(pdu_strbuf, "Not enough Data to proceed!");
      break; // exit while loop
    }
    
    /* Add the protocol and info column to the summary display */
    ep_strbuf_append(pdu_strbuf,
                    val_to_str(pdu_type, pdu_type_vals, "Unknown PDU type %d"));
    if (pdu_len < remainder)
    {
      ep_strbuf_append(pdu_strbuf, ", ");
    }
    
    if (tree)
    {
      /* The main-node covers the whole PDU (= len) */
      ti = proto_tree_add_item(tree, proto_srxproxy, tvb, bytes_dissected,
                               pdu_len, FALSE);
      srxproxy_tree = proto_item_add_subtree(ti, ett_srxproxy);

      /* Add the nodes for the header (FALSE = big-endian) */
      proto_tree_add_item(srxproxy_tree, hf_srxproxy_pdu_type, tvb,
                          SRX_OFFSET_PDU_TYPE, 1, FALSE);

      switch(pdu_type)
      {
        case PDU_PROXSRX_HELLO:
          pdu_len = dissect_srxproxy_hello(tvb, srxproxy_tree, bytes_dissected);
          break;
        case PDU_PROXSRX_HELLO_RESPONSE:
          pdu_len = dissect_srxproxy_hello_response(tvb, srxproxy_tree, 
                                                    bytes_dissected);
          break;
        case PDU_PROXSRX_GOODBYE:
          pdu_len = dissect_srxproxy_goodbye(tvb, srxproxy_tree, 
                                             bytes_dissected);
          break;
        case PDU_PROXSRX_VERIFY_V4_REQUEST:
          pdu_len = dissect_srxproxy_veri_request(tvb, srxproxy_tree, FALSE, 
                                                  bytes_dissected);
          break;
        case PDU_PROXSRX_VERIFY_V6_REQUEST:
          pdu_len = dissect_srxproxy_veri_request(tvb, srxproxy_tree, TRUE, 
                                                  bytes_dissected);
          break;
        case PDU_PROXSRX_SIGN_REQUEST:
          pdu_len = dissect_srxproxy_sign_request(tvb, srxproxy_tree, 
                                                  bytes_dissected);
          break;
        case PDU_PROXSRX_VERI_NOTIFICATION:
          pdu_len = dissect_srxproxy_veri_notification(tvb, srxproxy_tree, 
                                                       bytes_dissected);
          break;
        case PDU_PROXSRX_SIGN_NOTIFICATION:
          pdu_len = dissect_srxproxy_sign_notification(tvb, srxproxy_tree, 
                                                       bytes_dissected);
          break;
        case PDU_PROXSRX_DELTE_UPDATE:
          pdu_len = dissect_srxproxy_delete_update(tvb, srxproxy_tree, 
                                                   bytes_dissected);
          break;
        case PDU_PROXSRX_PEER_CHANGE:
          pdu_len = dissect_srxproxy_peer_change(tvb, srxproxy_tree, 
                                                 bytes_dissected);
          break;
        case PDU_PROXSRX_SYNC_REQUEST:
          pdu_len = dissect_srxproxy_sync_request(tvb, srxproxy_tree, 
                                                  bytes_dissected);
          break;
        case PDU_PROXSRX_ERROR:
          pdu_len = dissect_srxproxy_error(tvb, srxproxy_tree, bytes_dissected);
          break;
        case PDU_PROXSRX_UNKNOWN:
          pdu_len = 0;
          break;
      }
    }
    // Generate the data for the next loop run
    bytes_dissected += pdu_len;
    remainder = tvb_reported_length_remaining(tvb, bytes_dissected);    
  }

  if (pinfo != NULL)
  {
    if (check_col(pinfo->cinfo, COL_INFO))
    {
      col_add_str(pinfo->cinfo, COL_INFO, pdu_strbuf->str);
    }
  }

  return bytes_dissected > 0 ? bytes_dissected 
                             : (data_len - SRX_MINPDU_HEADER_SIZE);  
}

static int dissect_srxproxy_hello(tvbuff_t* tvb, proto_tree* tree, guint32 ofs)
{
  int retVal = 0;
  guint16 version;
  guint8  zero;
  guint32 length;
  guint32 proxyID;
  guint32 asn;
  guint32 noPeers;
  guint32 peerAS;
  int offset;
  int peerNr;

  // Fill values
  version = tvb_get_ntohs  (tvb, ofs + SRX_OFFSET_LINE_01 + 1);
  zero    = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 3);
  length  = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_02);
  proxyID = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_03);
  asn     = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_04);
  noPeers = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_05);

  if (length >= 20)
  {
    retVal = length;

    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 1), 2,
                        "version: %u", version);
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 3), 1,
                        "zero: 0x%02X", zero);
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_02, 4,
                        "length: %u", length);
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_03, 4,
                        "proxy id: %u", proxyID);
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_04, 4,
                        "AS number: %u", asn);
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_05, 4,
                        "number of peers: %u", noPeers);
    offset = ofs + SRX_OFFSET_LINE_06;
    peerNr = 1;
    while (noPeers > 0)
    {
      peerAS = tvb_get_ntohl (tvb, offset);
      proto_tree_add_text(tree, tvb, offset, 4, 
                          "Peers[%u]: AS%u", peerNr, peerAS);
      // Prepare for next round in loop
      noPeers--;
      peerNr++;
      offset += 4;
    }
  }
  return retVal;
}

static int dissect_srxproxy_hello_response(tvbuff_t* tvb, proto_tree* tree, 
                                           guint32 ofs)
{
  int retVal = 0;
  guint16 version;
  guint8  zero;
  guint32 length;
  guint32 proxyID;

  // Fill values
  version = tvb_get_ntohs  (tvb, ofs + SRX_OFFSET_LINE_01 + 1);
  zero    = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 3);
  length  = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_02);
  proxyID = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_03);

  if (length == 12)
  {
    retVal = length;

    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 1), 2,
                        "version: %u", version);
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 3), 1,
                        "zero: 0x%02X", zero);
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_02), 4,
                        "length: %u", length);
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_03), 4,
                        "proxy id: [0x%04X] %u", proxyID, proxyID);
  }
  return retVal;
}

static int dissect_srxproxy_goodbye(tvbuff_t* tvb, proto_tree* tree, 
                                    guint32 ofs)
{
  int retVal = 0;
  guint16 keep_window;
  guint8  zero;
  guint32 length;

  // Fill values
  keep_window = tvb_get_ntohs  (tvb, ofs + SRX_OFFSET_LINE_01 + 1);
  zero        = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 3);
  length      = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_02);

  if (length == 8)
  {
    retVal = length;

    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 1), 2,
                        "keep_window: %u", keep_window);
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 3), 1,
                        "zero: 0x%02X", zero);
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_02), 4,
                        "length: %u", length);
  }
  return retVal;
}

static int dissect_srxproxy_veri_request(tvbuff_t* tvb, proto_tree* tree,
                                          gboolean ipv6, guint32 ofs)
{
  int retVal = 0;
  guint8  flags;
  guint8  originResSrc;
  guint8  bgpsecResSrc;
  guint32 length;
  guint8  originDefRes;
  guint8  bgpsecDefRes;
  guint8  zero;
  guint8  prefixLength;
  guint32 requestToken;
  // IPv4 will be added right away
  int offset = ipv6 ? 16 : 4;
  guint32 originAS;
  guint32 bgpsecLength;
  // Byte stream will be added as hex data
  emem_strbuf_t* strbuf;

  // Fill values
  flags        = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 1);
  originResSrc    = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 2);
  bgpsecResSrc = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 3);
  length       = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_02);
  originDefRes = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_03);
  bgpsecDefRes = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_03 + 1);
  zero         = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_03 + 2);
  prefixLength = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_03 + 3);
  requestToken = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_04)
  // IPv4/6 will be added right away
  originAS     = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_05 + offset);
  bgpsecLength = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_06 + offset);

  if (length >= 28)
  {
    retVal = length;

    // FLAGS
    strbuf = ep_strbuf_new_label(((flags & SRX_FLAG_RECEIPT_REQUEST) != 0)
                                 ? "RECEIPT, " : "");
    {

    }
    if ((flags & SRX_FLAG_VERIFY_PREFIX_ORIGIN) != 0)
    {
      ep_strbuf_append(strbuf, "ORIGIN");
      if ((flags & SRX_FLAG_VERIFY_PATH) != 0)
      {
        ep_strbuf_append(strbuf, " & BGPSEC");
      }
    }
    else if ((flags & SRX_FLAG_VERIFY_PATH) != 0)
    {
      strbuf = ep_strbuf_new_label("BGPSEC");
    }
    else
    {
      strbuf = ep_strbuf_new_label("Store Only"); // Store Only
    }
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_01 + 1, 1,
                        "flags: 0x%02X (%s)", flags, strbuf->str);
    // ORIGIN ResSrc
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 2), 1,
                        "src origin result: 0x%02X (%s)", originResSrc,
                        val_to_str(originResSrc, result_src_vals,
                                   "unknown result source type %d"));
    // BGPSEC ResSrc
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 3), 1,
                        "Src BGPSEC result: 0x%02X (%s)", bgpsecResSrc,
                        val_to_str(bgpsecResSrc, result_src_vals,
                                   "Unknown Result Source Type %d"));
    // Length
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_02, 4,
                        "length: %u", length);
    // ORIGIN Default Result
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_03, 1,
                        "origin default result: 0x%02X (%s)",
                        originDefRes, val_to_str(originDefRes, result_type_vals,
                                              "unknown result type %d"));
    // BGPSEC Default Result
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_03 + 1), 1,
                        "BGPSEC default result: 0x%02X (%s)",
                        bgpsecDefRes, val_to_str(bgpsecDefRes, result_type_vals,
                                                 "Unknown Result Type %d"));
    // Zero
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_03 + 2), 1,
                        "zero: 0x%02X", zero);

    // Prefix Length
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_03 + 3), 1,
                        "prefix length: %u", prefixLength);

    // Request Token
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_04), 1,
                        "request token: 0x%08X", requestToken);
    
    // IP Prefix
    int ofs1;
    ofs1 = ofs + SRX_OFFSET_LINE_05;
    if (ipv6)
    {
      proto_tree_add_ipv6(tree, hf_srxproxy_prefix6, tvb, ofs1, IP_V6_ADDR_LEN,
                          tvb_get_ptr(tvb, ofs1, IP_V6_ADDR_LEN));
    }
    else
    {
      proto_tree_add_ipv4(tree, hf_srxproxy_prefix4, tvb, ofs1, IP_V4_ADDR_LEN,
                          tvb_get_ipv4(tvb, ofs1));
    }

    // Origin AS
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_04+offset, 4,
                        "origin AS: %u", originAS);
    offset += 4;

    // Length BGPSEC
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_04+offset, 4,
                        "length BGPSEC: %u", bgpsecLength);
    offset += 4;

    if (bgpsecLength > 0)
    {
      proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_04+offset, 
                          bgpsecLength, "data (...)");
    }
  }
  return retVal;
}

static int dissect_srxproxy_sign_request(tvbuff_t* tvb, proto_tree* tree, 
                                         guint32 ofs)
{
  int retVal = 0;
  guint16 algorithm;
  guint8  blockType;
  guint32 length;
  guint32 updateId;
  guint32 prependCounter;
  guint32 peerAS;

  // Fill values
  algorithm      = tvb_get_ntohs  (tvb, ofs + SRX_OFFSET_LINE_01 + 1);
  blockType      = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 3);
  length         = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_02);
  updateId       = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_03);
  prependCounter = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_04);
  peerAS         = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_05);

  if (length == 20)
  {
    retVal = length;

    // Algorithm
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 1), 2,
                        "algorithm: 0x%04X", algorithm);
    // Block Type
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 3), 1,
                        "block type: 0x%02X (%s)", blockType,
                        val_to_str(blockType, block_type_vals,
                                   "Unknown Block Type %d"));
    // Length
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_02, 4, "Length: %u", 
                        length);

    // Update Id
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_03, 4,
                        "update id: [0x%08X] (%u)", updateId, updateId);
    // Prepend Counter
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_04, 4,
                        "prepend counter: %u", prependCounter);
    // Peer AS
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_05, 4,
                        "peer AS: %u", peerAS);
  }

  return retVal;
}

static int dissect_srxproxy_veri_notification(tvbuff_t* tvb, proto_tree* tree, 
                                              guint32 ofs)
{
  int retVal = 0;
  guint8  resultType;
  guint8  originResult;
  guint8  bgpsecResult;
  guint32 length;
  guint32 requestToken;  
  guint32 updateId;

  emem_strbuf_t* strbuf;

  // Fill values
  resultType     = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 1);
  originResult   = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 2);
  bgpsecResult   = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 3);
  length         = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_02);
  requestToken   = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_03);
  updateId       = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_04);

  if (length == 12)
  {
    retVal = length;

    // Result Type
    if ((resultType & SRX_FLAG_RESULT_TYPE_ORIGIN) != 0)
    {
      if ((resultType & SRX_FLAG_RESULT_TYPE_BGPSEC) != 0)
      {
        strbuf = ep_strbuf_new_label("ORIGIN & PATH");
      }
      else
      {
        strbuf = ep_strbuf_new_label("ORIGIN ONLY");
      }
    }
    else if ((resultType & SRX_FLAG_RESULT_TYPE_BGPSEC) != 0)
    {
      strbuf = ep_strbuf_new_label("PATH ONLY");
    }
    else
    {
      strbuf = ep_strbuf_new_label("??? UNDEFINED ???");
    }

    // Receipt flag and what does the data contain
    if ((resultType & SRX_FLAG_RECEIPT_REQUEST) != 0)
    {
      proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 1), 1,
                          "result type: 0x%02X (RECEIPT, %s)", resultType, 
                          strbuf->str);
    }
    else
    {
      proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 1), 1,
                          "result type: 0x%02X (%s)", resultType, strbuf->str);
    }

    // ORIGIN Result
    strbuf = ep_strbuf_new_label("");
    if (   ((resultType & SRX_FLAG_RESULT_TYPE_ORIGIN) == 0)
        && (originResult != SRX_RESULT_UNDEFINED))
    {
      strbuf = ep_strbuf_new_label(" <- NOT REQUESTED, SHOULD BE UNDEFINED "
                                   "(0x03)");
    }
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 2), 1,
                        "origin result: 0x%02X (%s) %s", originResult,
                        val_to_str(originResult, result_type_vals,
                                   "Unknown Result Type %d"),
                        strbuf->str);

    // BGPSEC Result
    strbuf = ep_strbuf_new_label("");
    if (   ((resultType & SRX_FLAG_RESULT_TYPE_BGPSEC) == 0)
        && (bgpsecResult != SRX_RESULT_UNDEFINED))
    {
      strbuf = ep_strbuf_new_label(" <- NOT REQUESTED, SHOULD BE UNDEFINED "
                                   "(0x03)");
    }
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 3), 1,
                        "BGPSEC result: 0x%02X (%s) %s", bgpsecResult,
                        val_to_str(bgpsecResult, result_type_vals,
                                   "Unknown Result Type %d"),
                        strbuf->str);

    // Length
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_02, 4, "length: %u", 
                        length);

    // Update Id
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_03, 4,
                        "update id: [0x%08X] (%u)", updateId, updateId);
  }

  return retVal;
}

static int dissect_srxproxy_sign_notification(tvbuff_t* tvb, proto_tree* tree, 
                                              guint32 ofs)
{
  int retVal = 0;
  gint16 reserved;
  gint8  zero;
  gint32 length;
  gint32 updateId;
  gint32 lengthBGPSec;

  // Fill values
  reserved       = tvb_get_ntohs  (tvb, ofs + SRX_OFFSET_LINE_01 + 1);
  zero           = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 3);
  length         = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_02);
  updateId       = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_03);
  lengthBGPSec   = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_04);

  if (length >= 16)
  {
    retVal = length;

    // Reserved
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 1), 2,
                        "reserved: 0x%04X", reserved);
    // zero
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 3), 1,
                        "zero: 0x%02X", zero);
    // Length
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_02, 4, "length: %u", 
                        length);

    // Update Id
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_03, 4,
                        "update id: [0x%08X] %u", updateId, updateId);
    // Length BGPSEC
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_04, 4,
                        "length BGPSEC: %u", lengthBGPSec);

    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_05, lengthBGPSec,
                        (lengthBGPSec > 0) ? "BGPSEC data: (...)"
                                           : "BGPSEC data: none");
  }

  return retVal;
}

static int dissect_srxproxy_delete_update(tvbuff_t* tvb, proto_tree* tree, 
                                          guint32 ofs)
{
  int retVal = 0;
  gint16 keepWindow;
  gint8  zero;
  gint32 length;
  gint32 updateId;

  // Fill values
  keepWindow     = tvb_get_ntohs  (tvb, ofs + SRX_OFFSET_LINE_01 + 1);
  zero           = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 3);
  length         = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_02);
  updateId       = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_03);

  if (length == 12)
  {
    retVal = length;

    // Keep Window
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 1), 2,
                        "keep window: %u", keepWindow);
    // zero
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 3), 1,
                        "zero: 0x%02X", zero);
    // Length
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_02, 4, "length: %u", 
                        length);

    // Update Id
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_03, 4,
                        "update id: [0x%08X] %u", updateId, updateId);
  }

  return retVal;
}

static int dissect_srxproxy_peer_change(tvbuff_t* tvb, proto_tree* tree, 
                                        guint32 ofs)
{
  int retVal = 0;
  gint16 reserved;
  gint8  changeType;
  gint32 length;
  gint32 peerAS;

  // Fill values
  reserved   = tvb_get_ntohs  (tvb, ofs + SRX_OFFSET_LINE_01 + 1);
  changeType = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 3);
  length     = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_02);
  peerAS     = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_03);

  if (length == 12)
  {
    retVal = length;

    // Reserved
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 1), 2,
                        "reserved: 0x%04X", reserved);
    // Change Type
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 3), 1,
                        "change type: %u (%s)", changeType,
                        val_to_str(changeType, peer_change_vals,
                                   "Undefined Change Type"));
    // Length
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_02, 4, "length: %u", 
                        length);

    // Peer AS
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_03, 4,
                        "peer AS: [0x%08X] %u", peerAS, peerAS);
  }

  return retVal;
}

static int dissect_srxproxy_sync_request(tvbuff_t* tvb, proto_tree* tree, 
                                         guint32 ofs)
{
  int retVal = 0;
  gint16 reserved;
  gint8  zero;
  gint32 length;

  // Fill values
  reserved = tvb_get_ntohs  (tvb, ofs + SRX_OFFSET_LINE_01 + 1);
  zero     = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 3);
  length   = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_02);

  if (length == 8)
  {
    retVal = length;

    // Reserved
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 1), 2,
                        "reserved: 0x%04X", reserved);
    // Zero
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 3), 1,
                        "zero: 0x%02X", zero);
    // Length
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_02, 4, "length: %u", 
                        length);
  }

  return retVal;
}

static int dissect_srxproxy_error(tvbuff_t* tvb, proto_tree* tree, guint32 ofs)
{
  int retVal = 0;
  gint16 errorCode;
  gint8  zero;
  gint32 length;

  // Fill values
  errorCode = tvb_get_ntohs  (tvb, ofs + SRX_OFFSET_LINE_01 + 1);
  zero      = tvb_get_guint8 (tvb, ofs + SRX_OFFSET_LINE_01 + 3);
  length    = tvb_get_ntohl  (tvb, ofs + SRX_OFFSET_LINE_02);

  if (length == 8)
  {
    retVal = length;

    // Reserved
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 1), 2,
                        "error code: %u (%s)", errorCode,
                        val_to_str(errorCode, error_code_vals,
                        "Undefined Error Code"));
    // Zero
    proto_tree_add_text(tree, tvb, (ofs + SRX_OFFSET_LINE_01 + 3), 1,
                        "zero: 0x%02X", zero);
    // Length
    proto_tree_add_text(tree, tvb, ofs + SRX_OFFSET_LINE_02, 4, "length: %u", 
                        length);
  }

  return retVal;
}